from bisect import bisect
import numpy
import time
from operator import itemgetter
from .          import __version__
from .fieldlist import FieldList
from .utils     import equals, RTOL, ATOL, CfDict
from .flags     import Flags

_global_attributes = ('comment',
                      'Conventions',
                      'history',
                      'institution',
                      'references',
                      'source',
                      'title',
                      )

# ====================================================================
#
# AggregateError class
#
# ====================================================================

class AggregateError(Exception):
    '''

An exception that indicates that two fields can not be aggregated.
'''
    pass
#--- End: class


def aggregate(fields, rtol=None, atol=None,
              verbose=False, messages=False, strict_units=False):
    '''
    
Aggregates fields in a field list in place.

Applies the CF aggregation rules.

:Parameters:

    fields : FieldList
        The fields to be aggregated into as few fields as possible.

    messages : bool, optional
        Print a reason for why a field can't be aggregated.

    verbose : bool, optional
        Print a summary of the aggregation.

    strict_units : bool, optional
        If true then assume that variables with the same standard name
        but missing units are not aggregatable. By default assume that
        variables with the same standard name but missing units all
        have equal units

:Returns: 

    None

**Examples**

>>> f
[<CF Field: eastward_wind(110, 106)>,
 <CF Field: eastward_wind(110, 106)>,
 <CF Field: eastward_wind(110, 106)>,
 <CF Field: air_temperature(12, 73, 96)>,
 <CF Field: eastward_wind(110, 106)>,
 <CF Field: eastward_wind(110, 106)>,
 <CF Field: eastward_wind(1, 1, 110, 106)>,
 <CF Field: air_temperature(96, 73)>]
>>> cf.aggregate(f)
>>> f
[<CF Field: eastward_wind(3, 2, 110, 106)>,
 <CF Field: air_temperature(13, 73, 96)>]
'''

    if atol is None:
        atol = ATOL()
    if rtol is None:
        rtol = RTOL()

    equivalent_field_units  = {}
    equivalent_coord_units  = {}
    equivalent_cm_units     = {}
    unique_cell_methods     = []
    unique_transform        = []

    aggregated_fields = FieldList()
        
    aggregating_group = {}

    history = '%s Aggregated by cf-python v%s' % \
        (time.strftime('%Y-%m-%d %H:%M:%S UTC', time.gmtime()),
         __version__)

    for f in fields:

        do_not_aggregate = 0

        # Raise an exception for missing functionality
        for attr in ('ancillary_variables',):
            if hasattr(f, attr):
                if messages:
                    print \
"Can't aggregate %s: Code 1, %s" % (repr(f), attr)
                do_not_aggregate = 1
                break
        #--- End: for

        if do_not_aggregate:
            aggregated_fields.append(f)
            continue

        # Still here? Then find this field's identity
        identity = f.getattr('standard_name', getattr(f, 'id', None))

        if identity is None:
            # This field has no aggregating name, so it can not be
            # aggregated. Put it straight into the output list.
            if messages:
                print \
"Can't aggregate %s: Field has no standard/aggregating name" % repr(f)
            aggregated_fields.append(f)
            continue

        # Still here? Then try to make a hashable representation of
        # the field's space
        meta = CfDict()

        all_standard_names     = []
        meta.nd_standard_names = []
        meta.simple_attrs      = {}   # dch - do we need this - don't seem to use it.
        meta.key               = {}
        meta.cell_measures     = [] # List of space keys which are cell measures

        for key, coord in f.space.iteritems():

            # Check for cell measures
            if key[0:2] == 'cm':
                meta.cell_measures.append(key)
                continue
             
            # Still here? Then check that there's a standard_name
            try:
                standard_name = coord.standard_name
            except AttributeError:                
                if messages:
                    print \
"Can't aggregate %s: Coordinate has no standard name" % repr(f)
                do_not_aggregate = 1
                break

            # Still here?
            all_standard_names.append(standard_name)        

            meta.key[standard_name] = key

            try:
                if coord.ndim <= 1:
                    dim = f.space.dimensions[key][0]                    
                else:
                    meta.simple_attrs[key] = sorted(coord._simple_attributes().iteritems())
                    meta.nd_standard_names.append(standard_name)
                    continue
            except AttributeError:
                # A space component has no data
                if messages:
                    print \
"Can't aggregate %s: Coordinate has no data" % repr(f)
                do_not_aggregate = 1
                break

            # Still here?
            meta.simple_attrs[key] = sorted(coord._simple_attributes().iteritems())
            if dim in meta:
                meta[dim]['1d_standard_names'].append(standard_name)
            else:
                # Initialize meta[dim]['1d_standard_names']
                meta[dim] = {'1d_standard_names': [standard_name]}
        #--- End: for

        if do_not_aggregate:
            aggregated_fields.append(f)
            continue

        if len(all_standard_names) > len(meta.key):
            # Non-unique standard names
            aggregated_fields.append(f)
            continue

        # Each dimension in each space has at least one dimension
        # coordinate or 1-d auxiliary coordinate associated with it.
        for dim in f.space.dimension_sizes:
            if dim not in meta:
                # A dimension has no scalar or 1-d coordinates
                if messages:
                    print \
"Can't aggregate %s: Dimension has no 1-d coordinates" %repr(f)
                do_not_aggregate = 1
                break
        #--- End: for

        if do_not_aggregate:
            # A dimension has no scalar or 1-d coordinates
            aggregated_fields.append(f)
            continue

        # Sort each dimension's 1-d coordinates' standard names
        for value in meta.values():
            value['1d_standard_names'].sort()
        meta.nd_standard_names.sort()

        f._space_metadata = meta

        # ------------------------------------------------------------
        # Create a key entry for the field's Units
        # ------------------------------------------------------------
        field_units = _units(f, identity,
                             equivalent_field_units, 
                             strict_units=strict_units)

        # ------------------------------------------------------------
        # Create a key entry for the field's Flags
        # ------------------------------------------------------------
        field_flags = getattr(f, 'Flags', Flags()).hash()  

        # ------------------------------------------------------------
        # Create a key entry for the field's cell methods. (Do this
        # after ascertaining that every coordinate has a standard name
        # so we can be sure that the 'name' keys are comparable.)
        # ------------------------------------------------------------
        field_cell_methods = getattr(f, 'cell_methods', None)
        if field_cell_methods is not None:
            # This field has cell methods. If we've seen this
            # equivalent cell_methods before then use the existing one
            # in the key            
            new_cell_methods = 1
            for cell_methods1 in unique_cell_methods:
                if len(field_cell_methods) != len(cell_methods1):
                    continue          
                for cm0, cm1 in zip(field_cell_methods, cell_methods1):
                    keys0 = set(cm0) - set(('dim', 'no_coords'))
                    keys1 = set(cm1) - set(('dim', 'no_coords'))
                    if keys0 != keys1:
                        continue                    
                    for key in keys0:
                        if not equals(cm0[key], cm1[key], rtol=rtol, atol=atol):
                            continue
                new_cell_methods = 0
                field_cell_methods = cell_methods1
                break
            #--- End: for
            if new_cell_methods:
                unique_cell_methods.insert(0, field_cell_methods)
        #--- End: if

        # ------------------------------------------------------------
        # Create a key entry for the field's public attributes
        # ------------------------------------------------------------
        field_simple_attrs = f._simple_attributes()

        for attr in _global_attributes:
            if attr in field_simple_attrs:
                del field_simple_attrs[attr]
        #--- End: for
        field_simple_attrs = tuple(sorted(field_simple_attrs.iteritems()))

        # ------------------------------------------------------------
        # Create a hashable description for each coordinate
        # construct. This comprises public attributes, transforms,
        # units and whether or not it has bounds (and whether or not
        # they are climatological.
        # ------------------------------------------------------------
        coordinate_attrs = []

        # Find any transforms that we haven't seen before
        for transform in f.space.transforms.values():
            if transform.is_formula_terms:
                # Can't do anything with formula_terms, yet
                if messages:
                    print \
"Can't aggregate %s: Coordinate has formula_terms" % repr(f)
                do_not_aggregate = 1
                break

            if transform not in unique_transform:
                unique_transform.append(transform)
        #--- End: for
        if do_not_aggregate:
            aggregated_fields.append(f)
            continue
        #--- End: if
           
        try:
            unique_transform.extend(t 
                                    for t in f.space.transforms.values()
                                    if t not in unique_transform)
        except AttributeError:
            pass

        for standard_name, key in sorted(meta.key.iteritems()):

            # Find any units that we haven't seen before
            coord_units = _units(f.space[key], standard_name,
                                 equivalent_coord_units,
                                 strict_units=strict_units)

            # Transform
            if hasattr(f.space[key], 'transform'):
                transform = f.space.transforms[f.space[key].transform]
                transform = tuple(sorted(transform.iteritems()))
            else:
                transform = 0

            # Are there bounds? If so, are they climatological?
            bounds      = getattr(f.space[key], 'bounds', 0)
            climatology = 0
            if bounds is not 0:
                bounds = 1                
                climatology = getattr(f.space[key], 'climatology', 0)
                if climatology is not 0:
                    climatology = 1
            #--- End: if

            coordinate_attrs.append(
                (tuple(sorted(f.space[key]._simple_attributes().iteritems())), 
                 bounds,
                 climatology,
                 transform,
                 coord_units)
                )
        #--- End: for
        coordinate_attrs = tuple(coordinate_attrs)
            
        # ------------------------------------------------------------
        # Create a hashable description for each cell measure
        # construct. This comprises units and simple attributes.
        # ------------------------------------------------------------
        cell_measure_attrs = []
        for key in meta.cell_measures:
            # Find any units that we haven't seen before
            cm_units = _units(f.space[key], None, equivalent_cm_units,
                              strict_units=strict_units)
            
            cell_measure_attrs.append(
                (cm_units,
                 tuple(sorted(f.space[key]._simple_attributes().iteritems())))
                )
        #--- End: for
        cell_measure_attrs =  tuple(sorted(cell_measure_attrs, 
                                           key=itemgetter(0,1))
                                    )
                                    
        # ------------------------------------------------------------
        # Create the final dictionary key
        # ------------------------------------------------------------    
        key = (identity,
               len(f.space.dimension_sizes),
               field_simple_attrs,
               field_cell_methods,
               field_units,
               field_flags,
               coordinate_attrs,
               cell_measure_attrs,
               )

        # ------------------------------------------------------------
        # Append this field to the group of similar fields
        # ------------------------------------------------------------
        if key in aggregating_group:
            # Append this field to the group of fields with this
            # aggregating name
            aggregating_group[key].append(f)
        else:
            # Initialize the group of fields with this aggregating
            # name
            aggregating_group[key] = FieldList(f)
    #--- End: for    

    for k, g in aggregating_group.iteritems():

        name = g[0].getattr('standard_name', getattr(g[0], 'id', None))

        original_len_g = len(g)

        f = g

        M = 0
        while 1:
            len_f = len(f)
            i = 0
            while i < len(f):
                j = i + 1
                while j < len(f):
                    try:
                        # Attempt to aggregate two fields
#                        print i,j, len(f)
                        f[i] = _aggregate_2_fields(f[i], f[j],
                                                   rtol=rtol, atol=atol)
                        # Successfully aggregated fields f[i] and f[j]
                        # into f[i], so get rid of f[j]
                        f.pop(j)
                    except AggregateError as error:
                        # Could not aggregate fields f[i] and f[j]
                        if messages:
                            print \
"Can't aggregate %s: %s" % (repr(f[j]), error)
                        j += 1
                #--- End: while 
                i += 1
            #--- End: while 
            if len(f) == len_f:           
                break
            M += 1
        #--- End: while

        # Delete the space metadata
        del f._space_metadata

        aggregated_fields.extend(f)

        if verbose:
            print "Aggregated %d '%s' field%s to %d field%s" % \
                (original_len_g, name,
                 ('s' if original_len_g>1 else ''),                 
                 len_f, ('s' if len_f>1 else ''))
        #--- End: if

        # Set the history attribute
        if original_len_g > 1:
            f.history = history
    #--- End: for

    # Update input fields in-place
    fields[:] = aggregated_fields 
#--- End: def

def _units(var, identity, units, strict_units=False):
    '''

Return an equivalent Units object that has been encountered before, if
one exists. Otherwise return this variable's Units object and update
the `units` dictionary. If strict_units is False then assume that
variables with the same standard name but missing units all have equal
units.
'''
    var_units = var.Units

    if identity in units:
        if var_units:
            for u in units[identity]:
                if var_units.equivalent(u):
                    return u
            #--- End: for

            # Still here?
            units[identity].append(var_units)
        elif not strict_units or var.dtype.kind == 'S':
            return None
    else:
        if var_units:
            units[identity] = [var_units]
        elif not strict_units or var.dtype.kind == 'S':
            return None
    #--- End: if

    # Still here?
    return var_units
#--- End: def   

def _dimensions_are_paired(dimensions0, dimensions1, dim1_name_map):
    '''
'''
    if len(dimensions0) != len(dimensions1):
        # Components have different ranks
        return 
    
    # Still here?
    for dim1 in dimensions1:
        if dim1_name_map[dim1] not in dimensions0:
            # Component's dimensions are not matching pairs
            return 
    #--- End: for

    # Still here?
    return 1
#--- End: def

def _aggregate_2_fields(field0, field1,
                        rtol=None, atol=None):

#- Both fields have the same standard_name, etc.

#- Each dimension in each space has at least one scalar or 1-d
#  coordinate (dimension or auxiliary) and each such coordinate's
#  standard_name is unique within its space.

    if field1.size > field0.size or field1.ndim > field0.ndim:
        field0, field1 = field1, field0
        
    space0 = field0.space
    space1 = field1.space

    meta0 = field0._space_metadata
    meta1 = field1._space_metadata

    # ================================================================
    # For each dimension in one of the spaces, there is a 'matching
    # dimension' in the other space with exactly the same list of
    # scalar or 1-d coordinate (dimension or auxiliary)
    # standard_names, although a standard_name may belong to different
    # types of coordinate in the two spaces. Call these coordinates a
    # pair of 'matching coordinates' for the matching dimensions
    # ================================================================
    matching_dimensions = {}
    dim1_name_map       = {'bounds': 'bounds'}
    dims1 = meta1.keys()
    
    for dim0 in meta0:        
        
        found_matching_dimension = 0
        standard_names0 = meta0[dim0]['1d_standard_names']        
        for dim1 in dims1:
            if standard_names0 == meta1[dim1]['1d_standard_names']:
                found_matching_dimension = 1
                dims1.remove(dim1)
                break
        #--- End: for
        if not found_matching_dimension:
            raise AggregateError, \
"Not all dimensions can be matched"

        # Add to the dictionaries which, given a dimension from one
        # field, easily return its matching dimension from the other
        # field. This is used to reorder variables from field1 when
        # checking that they have equivalent values and for inserting
        # one variable into another
        dim1_name_map[dim1] = dim0

        # Create a list of pairs of matching coordinates for this pair
        # of matching dimensions
        matching_dimensions[(dim0, dim1)] = [(meta0.key[sn], meta1.key[sn])
                                             for sn in standard_names0]
    #--- End: for

    # Initialize the list of pairs of matching N-d constructs. This
    # comprises all space components except 1-d coordinates. I.e. all
    # cell measures and all N-d (N>1) auxiliary coordinates.
    matching_nd_constructs = []

    # Find matching pairs of N-d auxiliary coordinate constructs
    for sn in meta0.nd_standard_names:
    
        # Check that potentially matching N-d auxiliary coordinate
        # constructs span equivalent dimensions
        key0 = meta0.key[sn]
        key1 = meta1.key[sn]

        if not _dimensions_are_paired(space0.dimensions[key0],
                                      space1.dimensions[key1],
                                      dim1_name_map):
            raise AggregateError, \
"N-d construct has no match"
                   
        # Add this pair to the list of matching N-d constructs
        matching_nd_constructs.append((key0, key1))
    #--- End: for

    # Find matching pairs of cell measure constructs
    cell_measures1 = meta1.cell_measures[:]
    for key0 in meta0.cell_measures:
        found_pair_of_cell_measures = 0
        for key1 in cell_measures1[:]:
            if _dimensions_are_paired(space0.dimensions[key0],
                                      space1.dimensions[key1],
                                      dim1_name_map):
                if found_pair_of_cell_measures:
                    raise AggregateError, \
"cm match oof"
                # check units/pubartrs here. ???
                found_pair_of_cell_measures = 1
                cell_measures1.remove(key1)
                continue
        #--- End: for
        if not found_pair_of_cell_measures:
            raise AggregateError, \
"cm match oof2"
            
        # Add this pair to the list of matching 
        matching_nd_constructs.append((key0, key1))
    #--- End: for

    # ================================================================
    # There is EXACTLY ONE pair of matching dimensions for which at
    # least one pair of matching 1-d coordinates do NOT share the same
    # values for their arrays and bounds. Call this dimension the
    # Aggregating Dimension (capitalized) in each field.
    # ================================================================
    aggregating_dims = None

    for (dim0, dim1), matching_1d_coords in matching_dimensions.iteritems():

        if aggregating_dims and (space0.dimension_sizes[dim0] != 
                                 space1.dimension_sizes[dim1]):
            raise AggregateError, \
"More than one pair of matching dimensions have different sizes"

        # Still here? Then this pair of matching dimensions has at
        # least one dimension coordinate associated with it. Therefore
        # we need to check for equivalent values.
        for key0, key1 in matching_1d_coords:
            if space0[key0]._equivalent_data(space1[key1], dim1_name_map,
                                             rtol=rtol, atol=atol,
                                             _check_units=False):
                # These two coordinates have equivalent data
                continue

            # Still here? Then this pair of matching coordinates have
            # non-equivalent values.
            if aggregating_dims:
                raise AggregateError, \
"More than one pair of matching dimensions have 1-d coordinates with different values"
            else:
                # The aggregating_dims has not yet been set, so this
                # pair of matching dimensions could comprise the
                # aggregating dimensions, provided we don't find
                # another pair of matching dimensions with unequal
                # data.
                aggregating_dims = (dim0, dim1)
                break
        #--- End: for
    #--- End: for

    # ----------------------------------------------------------------
    # Find which pairs of matching coordinates/cell measures (1-d and
    # Nd) span their aggregating dimension and which do not. Also
    # find, for each field, the key of its aggregating dimension.
    # ----------------------------------------------------------------
    if not aggregating_dims:
        # The name 'aggregating' is not defined because each pair of
        # matching coordinates has the same data values
        raise AggregateError, \
"No pair of aggregating dimensions"

    spanning_1d_coordinates = matching_dimensions[aggregating_dims]

    # ----------------------------------------------------------------
    # Check that neither or both aggregating dimensions have a
    # dimension coordinate.
    # ----------------------------------------------------------------
    dim_coords = None
    for key0, key1 in spanning_1d_coordinates:
        if key0[0] == 'd' and key1[0] == 'd':
            # Found a pair of dimension coordinates
            dim_coords = (key0, key1)
            break
        if key0[0] == 'd' or key1[0] == 'd':
            raise AggregateError, \
"Can not join a dimension and an auxiliary coordinate"
    #--- End: for

    # ----------------------------------------------------------------
    # Find, for each field, the identity of its aggregating dimension.
    # ----------------------------------------------------------------
    adim_key0, adim_key1 = aggregating_dims

    for i, (key0, key1) in enumerate(spanning_1d_coordinates):
        spanning_1d_coordinates[i] = (space0[key0], space1[key1])

    spanning_nd_constructs     = []
    non_spanning_nd_constructs = []
    for key0, key1 in matching_nd_constructs:
        if adim_key0 in space0.dimensions[key0]:            
            spanning_nd_constructs.append((space0[key0], space1[key1]))
        else:
            non_spanning_nd_constructs.append((space0[key0], space1[key1]))
    #--- End: for

    # ================================================================
    # If either or both aggregating dimensions have a dimension
    # coordinate then there are no values for it and its matching
    # coordinate which appear in both spaces (if they run in different
    # senses or their ranges overlap, the aggregated dimension will
    # need to be sorted). [All other coords are concatenated.]
    # ================================================================

    # ----------------------------------------------------------------
    # Find the insertion indices
    # ----------------------------------------------------------------
    if not dim_coords:
        # ------------------------------------------------------------
        # Neither field's aggregating dimension has a dimension
        # coordinate (because name 'dim_coord' is not
        # defined). Therefore the fields will be simply concatenated
        # along the aggregating dimension
        # ------------------------------------------------------------
        indices = None
    else:        
        # ------------------------------------------------------------
        # The aggregating dimensions of both fields have a dimension
        # coordinate. Therefore we need to find the insertion index in
        # field0 for each data element in field1.
        # ------------------------------------------------------------
        dim_coord1 = space1[dim_coords[1]]
        data0      = space0[dim_coords[0]].Data
        data1      = dim_coord1.Data

        order          = data0.order
        units0         = data0.Units
        direction      = data0.direction
        save_to_disk   = data0.save_to_disk
        revert_to_file = True

        if hasattr(dim_coord1, 'bounds'):
            has_bounds = True
            bounds1              = dim_coord1.bounds.Data
            bounds1_order        = bounds1.order
            bounds1_direction    = bounds1.direction
            bounds1              = bounds1.partitions.ravel()
        else:
            has_bounds = False
      
        mins0 = []
        maxs0 = []

        for partition in data0.partitions.flat():

            partition.conform(order, direction, units0,
                              save_to_disk, revert_to_file) 

            data = partition.data
            size = data.size
            if size == 1:
                first = data.item()
                mins0.append(first)
                maxs0.append(first)
            else:
                first_last = data[slice(0, size, size-1)]
                mins0.append(min(first_last))
                maxs0.append(max(first_last))
            #--- End: if

            # Put the partition's data on to disk if it's too big
            partition.close()
        #--- End: for

        # Set the reversed_ flag to signify whether or not we had to
        # reverse the mins0 and maxs0 lists in order to get them into
        # ascending sequences.
        if data0.is_scalar:
            reversed_ = not data0.direction
        else:
            reversed_ = not data0.direction.values()[0]

        if reversed_:
            mins0.reverse()
            maxs0.reverse()

        order        = data1.order
        units0       = data0.Units  # Yes, data0
        direction    = data1.direction
        save         = False #data1.save_to_disk
        revert_to_file = False
        indices = []

        for i, partition in enumerate(data1.partitions.flat()):

            partition.conform(order, direction, units0, save, revert_to_file) 

            data = partition.data
            size = data.size
            if size == 1:
                min1 = max1 = data.item()
            else:
                first_last = data[slice(0, size, size-1)]
                min1       = min(first_last)
                max1       = max(first_last)
            #--- End: if

            index = bisect(mins0, min1)

            if reversed_:
                indices.append(len(mins0)-index)
            else:
                indices.append(index)

            # Update mins0 and maxs0
            mins0.insert(index, min1)
            maxs0.insert(index, max1)

            # Check for overlapping dimension coordinates (not allowed
            # for now). This check also serves to check for any shared
            # dimension coordinate values.
            if index == len(maxs0)-1:
                if min1 <= maxs0[index-1]:
                    raise AggregateError, \
"Overlapping or equal '%s' dimension coordinate values" % dim_coord1.standard_name
            elif index == 0:
                if max1 >= mins0[index+1]:
                    raise AggregateError, \
"Overlapping or equal '%s' dimension coordinate values" % dim_coord1.standard_name
            else:
                 if min1 <= maxs0[index-1] or max1 >= mins0[index+1]:
                     raise AggregateError, \
"Overlapping or equal '%s' dimension coordinate values" % dim_coord1.standard_name
            #--- End: if

            # Check for overlapping dimension coordinate bounds. This
            # is a bit strict for now
            if has_bounds:
                
                bounds = bounds1[i]

                bounds = bounds.conform(bounds1_order,
                                        bounds1_direction,
                                        units0,
                                        False, True)

                if index == len(maxs0)-1:
                    if numpy.amin(bounds) <= maxs0[index-1]:
                        raise AggregateError, \
"Overlapping (ish) '%s' dimension coordinate bounds" % dim_coord1.standard_name
                elif index == 0:
                    if numpy.amax(bounds) >= mins0[index+1]:
                        raise AggregateError, \
"Overlapping (ish) '%s' dimension coordinate bounds" % dim_coord1.standard_name
                else:  
                    if (numpy.amin(bounds) <= maxs0[index-1] or
                        numpy.amax(bounds) >= mins0[index+1]):
                        raise AggregateError, \
"Overlapping (ish) '%s' dimension coordinate bounds" % dim_coord1.standard_name
                #--- End: if
            #--- End: if

            # Put the partition's data on to disk if it's too big
            partition.close()
        #--- End: for
    #--- End: if

    # ================================================================
    # Check that N-d constructs which don't span the aggregating
    # dimension have the same data values
    # ================================================================
    for variable0, variable1 in non_spanning_nd_constructs:
        if not variable0._equivalent_data(variable1, dim1_name_map,
                                          rtol=rtol, atol=atol, 
                                          _check_units=False):
            # These two variables do not have equivalent data
            raise AggregateError, \
"Non spanning N-d constructs have different values"
    #--- End: for

    # ----------------------------------------------------------------
    if space0[adim_key0].Data.is_scalar:
        adim_direction = space0[adim_key0].Data.direction
    else:
        adim_direction = space0[adim_key0].Data.direction[space1.dimensions[adim_key0][0]]

    # 
    for var0, var1 in spanning_1d_coordinates + spanning_nd_constructs:
        var0._insert_data(var1, indices,
                          adim_key0, adim_direction, dim1_name_map)
    #--- End: for

    # Insert data
    field0._insert_data(field1, indices,
                        adim_key0, adim_direction, dim1_name_map)

    # Update the data dimensions 
    space0.dimensions['data'] = field0.Data.order[:]

    # Update the Aggregating Dimension size
    space0.dimension_sizes[adim_key0] += space1.dimension_sizes[adim_key1]

    # ----------------------------------------------------------------
    # Return the aggregated field
    # ----------------------------------------------------------------
    return field0
#--- End: def
