import re
from .utils import CfList, CfDict
from .data  import Data
from .utils import equals as utils_equals
# ====================================================================
#
# CellMethods object
#
# ====================================================================

class CellMethods(CfList):
    '''

A CF cell methods object to describe the characteristic of a field
that is represented by cell values.

Each cell method is stored in a dictionary and these dictionaries are
stored in a list-like object. Similarly to a CF cell_methods string,
the order of cell methods in the list is important.

The dictionary representing each cell method recognizes the following
keys (where all keys are optional; the referenced sections are from
the NetCDF Climate and Forecast (CF) Metadata Conventions and words in
double quotes ("...") refer to CF cell_methods components in the
referenced sections):

=========  ========================================================
Key        Value
=========  ========================================================
method     A "method" component. Refer to section 7.3.
           
name       A list of all of the "name" components involved with the
           "method" component. Each element is either:
           
               * The standard_name of the cell method's dimension
               * The string 'area'          
               * A netCDF variable name or netCDF dimension name
           
           The last option only occurs if the cell method's dimension
           either has no dimension coordinate or has a dimension
           coordinate with no standard_name.
           
           This is a deviation from the "name" components as
           described in sections 7.3, 7.3.1 and 7.3.4, in that
           standard_names are stored whenever possible, rather than
           only in the special cases described in section
           7.3.4. Inspection of the 'no_coords' list retains the
           information required to fully interpret the cell method.
           
dim        A list of space dimension names corresponding to the
           'name' list. A 'name' list value of 'area' always
           corresponds to a 'dim' list value of None.
           
no_coords  A list of booleans, corresponding to the 'name' list,
           indicating whether or not a particular cell method is
           relevant to the data in a way which may not be precisely
           defined by the corresponding dimension or dimensions. Refer
           to section 7.3.4.
           
interval   A list of Data objects giving "interval" component values,
           with their associated units, corresponding to the 'name'
           list. Refer to section 7.3.2.
           
comment    A "comment" or non-standardised information,
           component. Refer to sections 7.3.2 and 7.4
           
within     A "within" climatology component. Refer to section 7.4.
           
over       An "over" cell portion or "over" climatology component.
           Refer to sections 7.3.3 and 7.4.
           
where      A "where" component. Refer to section 7.3.3.
=========  ========================================================

Note that the above table assumes that the cell methods have been
constructed in the context of a field, the only way in which the 'dim'
and 'no_coords' keys may take sensible values. If this is not the
case, then elements of the 'dim' key default to False and elements of
the 'no_coords' key default to None.


**Examples**

>>> c = cf.CellMethods('time: minimum within years time: mean over years (ENSO years)')
>>> print c
Cell methods    : time: minimum within years
                  time: mean over years (ENSO years)
>>> list(c)
[
 {'method'   : 'minimum',
  'name'     : ['time'],     
  'dim'      : [False],
  'no_coords': [None], 
  'within'   : 'years'
 },
 {'method'   : 'mean'
  'name'     : ['time'],
  'dim'      : [False], 
  'no_coords': [None],     
  'comment'  : 'ENSO years', 
  'over'     : 'years'
 }
]

>>> c = c.CellMethods('lat: lon: standard_deviation')
>>> list(c)
[
 {'dim'      : [False, False],
  'method'   : 'standard_deviation'
  'name'     : ['lat', 'lon'],
  'no_coords': [None, None],
 }
]
 
'''

    def __init__(self, string=None):
        '''

**Initialization**

:Parameters:

    string : str, optional
        Initialize new instance from a CF-netCDF-like cell methods
        string. See the `parse` method for details. By default an
        empty cell methods is created.

**Examples**

>>> c = CellMethods()
>>> c = CellMethods('time: mean')

'''
        super(CellMethods, self).__init__()

        self.parse(string)
    #--- End: def

    def __repr__(self):
        '''
x.__repr__() <==> repr(x)

'''
        return '<CF %s: %s>' % (self.__class__.__name__, 
                                ' '.join(self.strings()))
    #--- End: def

    def __str__(self):
        '''
x.__str__() <==> str(x)

'''        
        strings = self.strings()

        if strings:
            string = ['Cell methods    : %s' % strings[0]]
            for s in strings[1:]:
                string.append('                  %s' % s)            
        else:
            return  'Cell methods    : '

        return '\n'.join(string)
    #--- End: def

    def dump(self, id=None):
         '''

Return a string containing a full description of the instance.

If a cell methods 'name' is followed by a '*' then that cell method is
relevant to the data in a way which may not be precisely defined its
corresponding dimension or dimensions.

:Parameters:

    id : str, optional
       Set the common prefix of component names. By default the
       instance's class name is used.

:Returns:

    out : str
        A string containing the description.

**Examples**

>>> x = c.dump()
>>> print c.dump()
>>> print c.dump(id='cellmethods1')

'''
         if id is None:
             id = self.__class__.__name__
                              
         string = [] #'cell methods\n------------']
         
         for i, s in enumerate(self.strings()):
             string.append('%s[%d] -> %s' % (id, i, s))
             
         return '\n'.join(string)
    #--- End: def

    def equals(self, other, rtol=None, atol=None, traceback=False):
        '''

True if two cell methods are logically equal, False otherwise.

Keys 'dim' and 'no_coords' are ignored in the comparison.

:Parameters:

    other : 
        The object to compare for equality.

    atol : float, optional
        The absolute tolerance for all numerical comparisons, By
        default the value returned by the `ATOL` function is used.

    rtol : float, optional
        The relative tolerance for all numerical comparisons, By
        default the value returned by the `RTOL` function is used.

    traceback : bool, optional
        If True then print a traceback highlighting where the two
        instances differ.

:Returns: 

    out : bool
        Whether or not the two instances are equal.

**Examples**

'''
        if self is other:
            return True

        # Check that each instance is the same type
        if self.__class__ != other.__class__:
            if traceback:
                print \
"%s: Different types: %s, %s" % (self.__class__.__name__,
                                 self.__class__.__name__,
                                 other.__class__.__name__)
            return False
        #--- End: if

        if len(self) != len(other):
            if traceback:
                print \
"%s: Different numbers of methods: %d, %d" % (self.__class__.__name__,
                                              len(self), len(other))
            return False
        #--- End: if

        omit_keys = set(('dim', 'no_coords'))

        for cm0, cm1 in zip(self, other):

            keys0 = set(cm0) - omit_keys
            keys1 = set(cm1) - omit_keys
            if keys0 != keys1: 
                if traceback:
                    print \
"%s: Different keys: %s" % (self.__class__.__name__,
                            keys0.symmetric_difference(keys1))
                return False
            #--- End: if

            for key in keys0:
                x = cm0[key]
                y = cm1[key]
                if not utils_equals(x, y, rtol=rtol, atol=atol,
                                    traceback=traceback):
                    if traceback:
                        print \
"%s: Different '%s' key values: %s, %s" % (self.__class__.__name__, key,
                                           repr(x), repr(y))
                    return False
        #--- End: for

        return True
    #--- End: def

    def has_cellmethods(self, other):
        '''

Return True if and only if this cell methods is a super set of another.

:Parameters:
    other : CellMethods
        The other cell methods for comparison.

:Returns:
    out : bool
        Whether or not this cell methods is a super set of the other.

**Examples**

>>> c = cf.CellMethods('time: mean height: mean area: mean')
>>> d = cf.CellMethods('area: mean time: mean')
>>> c.has_cellmethods(d)
True

'''
        for cm in self:
            match = False
            for i in other:
                for key in ('name', 'method', 'over', 'within'):
                    if key in i:
                        if key not in cm or i[key] != cm[key]:
                            match = False
                            break
                    #--- End: for
                    match = True
                #--- End: for
            #--- End: for
            if match:
                return True
        #--- End: for

        return
    #--- End: def

    def parse(self, string=None):
        '''

Parse a CF cell_methods string into this CellMethods instance in
place.

:Parameters:

    string : str, optional
        The CF cell_methods string to be parsed into the CellMethods
        object. By default the cell methods will be empty.

:Returns:

    None

**Examples**

>>> c = cf.CellMethods()
>>> c = c.parse('time: minimum within years time: mean over years (ENSO years)')    
>>> print c
Cell methods    : time: minimum within years
                  time: mean over years (ENSO years)

'''
        if not string:
            self._list[:] = []
            return

        # Split the cell_methods string into a list of strings ready
        # for parsing into the result list. E.g.
        #   'lat: mean (interval: 1 hour)'
        # maps to 
        #   ['lat:', 'mean', '(', 'interval:', '1', 'hour', ')']
        cell_methods = re.sub('\((?=[^\s])' , '( ', string)
        cell_methods = re.sub('(?<=[^\s])\)', ' )', cell_methods).split()

        while cell_methods:

            # Create a new element for this 
            self.append(CfDict())            

            # List of names
            self[-1]['name']      = CfList()
            self[-1]['dim']       = CfList()
            self[-1]['no_coords'] = CfList()

            while cell_methods:
                if not cell_methods[0].endswith(':'):
                    break
                self[-1]['name'].append(cell_methods.pop(0)[:-1])            
                self[-1]['dim'].append(False)
                self[-1]['no_coords'].append(None)
            #--- End: while

            if not cell_methods:
                break

            # Method
            self[-1]['method'] = cell_methods.pop(0)

            if not cell_methods:
                break

            # Climatological statistics and statistics which apply to
            # portions of cells
            while cell_methods[0] in ('within', 'where', 'over'):
                term = cell_methods.pop(0)
                self[-1][term] = cell_methods.pop(0)
                if not cell_methods:
                    break
            #--- End: while

            if not cell_methods: 
                break

            # interval and comment
            if cell_methods[0].endswith('('):
                cell_methods.pop(0)

                if not (re.search('^(interval|comment):$', cell_methods[0])):
                    cell_methods.insert(0, 'comment:')
                           
                while not re.search('^\)$', cell_methods[0]):
                    term = cell_methods.pop(0)[:-1]

                    if term == 'interval':
                        if 'interval' not in self[-1]:
                            self[-1]['interval'] = CfList()
                        interval = cell_methods.pop(0)
                        try:
                            interval = float(interval)
                        except ValueError:
                            pass

                        if cell_methods[0] != ')':
                            units = cell_methods.pop(0)
                        else:
                            units = None

                        try:
                            self[-1]['interval'].append(Data(interval, units))
                        except ValueError:
                            del self[-1]['interval']                            
                            print \
"WARNING: Ignoring unknown interval in %s initialization: %s %s" % \
(self.__class__.__name__, interval, units)
                            
                        continue
                    #--- End: if

                    if term == 'comment':
                        comment = CfList()
                        while cell_methods:
                            if cell_methods[0].endswith(')'):
                                break
                            if cell_methods[0].endswith(':'):
                                break
                            comment.append(cell_methods.pop(0))
                        #--- End: while
                        self[-1]['comment'] = ' '.join(comment)
                    #--- End: if

                #--- End: while 

                if cell_methods[0].endswith(')'):
                    cell_methods.pop(0)
            #--- End: if
        #--- End: while

    #--- End: def

    def strings(self):
        '''

Return a list of a CF-netCDF-like string of each cell method.

Note that if the intention is to concatenate the output list into a
string for creating a netCDF(CF) cell_methods attribute, then the cell
methods "name" components may need to be modified, where appropriate,
to reflect netCDF variable names.

:Returns:

    out : list
        A string for each cell method.

**Examples**

>>> c = cf.CellMethods('time: minimum within years time: mean over years (ENSO years)')
>>> c.strings()
['time: minimum within years',
 'time: mean over years (ENSO years)']

'''
        strings = []

        for cm in self:

            string = []

            for j in xrange(len(cm['name'])):
                string.append('%s: ' % cm['name'][j])

            string.append(cm['method'])

            x = []
            try:
                x.extend(('within', cm['within'])) 
            except KeyError:
                try:
                    x.extend(('where', cm['where'])) 
                except KeyError:
                    pass
            try:
                x.extend(('over', cm['over'])) 
            except KeyError:
                pass
            if x:
                string.append(' %s' % ' '.join(x))

            if 'interval' in cm:
                string.append(' (')
                for j in xrange(len(cm['interval'])):
                    interval = cm['interval'][j]
                    string.append('interval: %s' % interval.array.item())
                    if interval.Units:
                        string.append(' %s' % interval.Units)
                if 'comment' in cm:
                    string.append(' comment: %s' % cm['comment'])
                string.append(')')

            elif 'comment' in cm:
                string.append(' (%s)' % cm['comment'])
            
            strings.append(''.join(string))
        #--- End: for

        return strings
    #--- End: def

    def netCDF_translation(self, field):
        '''

Translate netCDF variable names.

:Parameters:

    field : Field

        The field which provides the translation.

:Returns:

    out : CellMethods
        A new cell methods instance with translated names.

**Examples**

>>> c = CellMethods('time: mean lon: mean')
>>> d = c.netCDF_translation(f)

'''
        new_cell_methods = self.copy()

        # Change each 'name' value to a standard_name (or space
        # coordinate key) and create the 'dim' key
            
        # From the CF conventions (1.5): In the specification of this
        # attribute, name can be a dimension of the variable, a scalar
        # coordinate variable, a valid standard name, or the word
        # 'area'.
        for cell_method in new_cell_methods:
            for i in xrange(len(cell_method['name'])):
    
                # Since input scalar coordinate variables have now
                # been converted into 1d dimension coordinates, name
                # being "a dimension of the variable, a scalar
                # coordinate variable" may be checked by testing name
                # against field.space.nc_dimensions.values().
                found_ncdimension = False
                for dim, ncdim in field.space.nc_dimensions.iteritems():
                    if cell_method['name'][i] == ncdim:
                        cell_method['no_coords'][i] = False
                        cell_method['dim'][i]       = dim
                        try:                            
                            cell_method['name'][i] = getattr(field.space[dim],
                                                             'standard_name',
                                                             ncdim)
                        except KeyError:
                            cell_method['name'][i] = ncdim
                        found_ncdimension = True
                        break
                #--- End: for
                if found_ncdimension:
                    continue
    
                # Still here? If the 'name' is the netCDF variable
                # name of a coordinate, then set 'dim' and, if
                # possible, replace 'name' with the coordinate's
                # standard name.
                found_ncvar = 0
                for key in field.space.get_keys('dim|aux'):
                    ncvar = field.space[key].ncvar
                    if cell_method['name'][i] == ncvar:
                        cell_method['name'][i] = getattr(field.space[key],
                                                         'standard_name', ncvar)
                        cell_method['dim'][i]  = key
                        found_ncvar = 1
                        break
                #--- End: for
                if found_ncvar:
                    continue
    
                # Still here? If the 'name' is the standard name of a
                # coordinate (or the string "area"), then set 'dim'
                cell_method['no_coords'][i] = True
                if cell_method['name'][i] == 'area':
                    cell_method['dim'][i] = None
                else:
                    for key in field.space.get_keys('dim|aux'):
                        try:
                            sn = field.space[key].standard_name
                        except AttributeError:
                            pass
                        else:
                            if cell_method['name'][i] == sn:
                                dim = field.space.coord(sn, key=True, exact=True)
                                cell_method['dim'][i] = dim
                                break
                    #--- End: for
                #--- End: if
    
            #--- End: for
        #--- End: for
    
        return new_cell_methods

    #--- End: def

#--- End: class
