from copy import deepcopy
from .data import Data

# ====================================================================
#
# Comparison object
#
# ====================================================================

class Comparison(object):
    '''

Create an object for storing a comparison expression.

A comparison expression comprises a relation (such as 'less than') and
a value (any object capable of being related to with the standard
comparison operators). These are stored in the `relation` and `value`
attributes respectively.

The comparison is evaluated for an arbitrary object with the `evaluate`
method.

Valid relations are:

=========  =============================================
Relation   Description
=========  =============================================
'lt'       Strictly less than
'le'       Less than or equal
'gt'       Strictly greater than
'ge'       Greater than or equal
'eq'       Equals (within a tolerance)
'ne'       Not equal to (within a tolerance)
'inside'   Inside a given range (range bounds included)
'outside'  Outside a given range (range bounds excluded)
=========  =============================================

As a convenience, for each relation in the above list there is an
identically named function which returns the appropriate Comparison
object.


**Examples**

>>> c = cf.Comparison('le', 5)
>>> c.evaluate(4)
True
>>> c.evaluate(5)
True

>>> c = lt(5)
>>> c.evaluate(4)
True

>>> c = cf.Comparison('inside', (1,2))
>>> a = numpy.arange(4)
>>> c.evaluate(a)
array([False,  True,  True, False], dtype=bool)
>>> (a >= 1) & (a <= 2)
array([False,  True,  True, False], dtype=bool)

'''

    def __init__(self, relation, value, units=None):
        '''

**Initialization**

:Parameters:

    relation : str
        The comparison operator type.

    value : object
        The value on the right hand side of the comparison operation.

    units : str or Units, optional
        The units of `value`. By default, the same units as the left
        hand side of the comparison operation are assumed.

'''
        if units:
            self.relation  = relation
# Overriding Data units ????
            self.value     = Data(value, units=units)
        else:
            self.relation  = relation
            self.value     = value
    #--- End: def

    def __eq__(self, x):
        '''
'''
        return self.evaluate(x)
    #--- End: def

    def __ne__(self, x):
        '''
'''
        return not self.evaluate(x)
    #--- End: def

    def __repr__(self):
        '''
x.__repr__() <==> repr(x)

'''
        return '<CF Comparison: %s %s>' % (self.relation,
                                           repr(self.value))
    #--- End: def

    def __str__(self):
        '''
x.__str__() <==> str(x)

''' 
        return '%s %s' % (self.relation, repr(self.value))
    #        return 'Comparison: x '+self.relation+' '+repr(self.value)+' ''+str(self.value.Units)
    #--- End: def

    def copy(self):
        '''

Return a deep copy.

Equivalent to ``copy.deepcopy(c)``.

:Returns:

    out :
        The deep copy.

**Examples**

>>> c.copy()

'''
        return deepcopy(self)
    #--- End: def

    def dump(self, id=None):
        '''
        
Return a string containing a full description of the instance.

:Parameters:

    id : str, optional
       Set the common prefix of component names. By default the
       instance's class name is used.

:Returns:

    out : str
        A string containing the description.

**Examples**

>>> x = c.dump()
>>> print c.dump()
>>> print c.dump(id='comparison1')

'''
        return str(self)
    #--- End: def

    def evaluate(self, x):
        '''

Evaluate the comparison expression for a given object.

:Parameters:

    x : object
        The object for the left hand side of the comparison
        expression.

:Returns:

    out : 
        The output of the comparison given by the `relation` attribute
        with `x` on the left hand side and the `value` attribute on
        the right hand side.
    
**Examples**

>>> c = cf.Comparison('lt', 5.5)
>>> c.evaluate(6)
False

>>> c = cf.Comparison('inside', (1,2))
>>> array = numpy.arange(4)
>>> array
array([0, 1, 2, 3])
>>> c.evaluate(array)
array([False,  True,  True, False], dtype=bool)

'''
        if self.relation == 'lt':
            return x <  self.value
        elif self.relation == 'le':
            return x <= self.value
        elif self.relation == 'gt':
            return x >  self.value
        elif self.relation == 'ge':
            return x >= self.value
        elif self.relation == 'eq':
            return x == self.value
        elif self.relation == 'ne':
            return x != self.value
        elif self.relation == 'inside':
            if self.value[0] < self.value[1]:
                return (x >= self.value[0]) & (x <= self.value[1])
            else:
                return (x >= self.value[1]) & (x <= self.value[0])    
        elif self.relation == 'outside':
            if self.value[0] < self.value[1]:
                return (x <  self.value[0]) | (x >  self.value[1])
            else:
                return (x <  self.value[1]) | (x <  self.value[0])    
    #--- End: def

#--- End: class

def lt(value, units=None):
    '''
    
Return an object for testing whether a variable is strictly less than
a given value.

:Parameters:

    value : object
        The value which a variable is to be compared with.

:Returns: 

    out : Comparison
        A Comparison object which will evaluate whether or not the
        comparison evaluates to True.

**Examples**

>>> c = cf.lt(5)
>>> c
<CF Comparison: x lt 5>
>>> c.evaluate(4)
True
>>> c.evaluate(5)
False

'''
    return Comparison('lt', value, units=units)
#--- End: def
    
def le(value, units=None):
    '''
    
Return an object for testing whether a variable is less than or equal
to the given value.

:Parameters:

    value : object
        The value which a variable is to be compared with.

:Returns: 

    out : Comparison
        A Comparison object which will evaluate whether or not the
        comparison evaluates to True.

**Examples**

>>> c = cf.le(5)
>>> c
<CF Comparison: x le 5>
>>> c.evaluate(5)
True
>>> c.evaluate(6)
False

'''
    return Comparison('le', value, units=units)
#--- End: def
    
def gt(value, units=None):
    '''
      
Return an object for testing whether a variable is greater than the
given value.

:Parameters:

    value : object
        The value which a variable is to be compared with.

:Returns: 

    out : Comparison
        A Comparison object which will evaluate whether or not the
        comparison evaluates to True.

**Examples**

>>> c = cf.gt(5)
>>> c
<CF Comparison: x gt 5>
>>> c.evaluate(6)
True
>>> c.evaluate(5)
False

'''
    return Comparison('gt', value, units=units)
#--- End: def
    
def ge(value, units=None):
    '''
     
Return an object for testing whether a variable is greater than or equal
to the given value.

:Parameters:

    value 
        The value which a variable is to be compared with.

:Returns:

    out : Comparison
        A Comparison object which will evaluate whether or not the comparison
        evaluates to True.

**Examples**

>>> c = cf.ge(5)
>>> c
<CF Comparison: x ge 5>
>>> c.evaluate(5)
True
>>> c.evaluate(4)
False

'''
    return Comparison('ge', value, units=units)
#--- End: def

# C-x f
    
def eq(value, units=None):
    '''
    
Return an object for testing whether a variable is equal to the given
value.

:Parameters:

    value 
        The value which a variable is to be compared with.

:Returns:

    out : Comparison
        A Comparison object which will evaluate whether or not the comparison
        evaluates to True.

**Examples**

>>> c = cf.eq(5)
>>> c
<CF Comparison: x eq 5>
>>> c.evaluate(5)
True
>>> c.evaluate(4)
False

'''
    return Comparison('e', value, units=units)
#--- End: def
    
def ne(value, units=None):
    '''
    
Return an object for testing whether a variable is not equal to the
given value.

:Parameters:

    value : object
        The value which a variable is to be compared with.

:Returns: 

    out : Comparison
        A Comparison object which will evaluate whether or not the
        comparison evaluates to True.

**Examples**

>>> c = cf.ne(5)
>>> c
<CF Comparison: x ne 5>
>>> c.evaluate(4)
True
>>> c.evaluate(5)
False

'''
    return Comparison('ne', value, units=units)
#--- End: def
    
def inside(value0, value1, units=None):
    '''
    
Return an object for testing whether a variable is is inside the given
range.

:Parameters:

    value0 : object
         The lower bound of the range which a variable is to be
         compared with.

    value1 : object
         The upper bound of the range which a variable is to be
         compared with.

:Returns: 

    out : Comparison
        A Comparison object which will evaluate whether or not the
        comparison evaluates to True.

**Examples**

>>> c = cf.inside(5,7)
>>> c
<CF Comparison: x inside (5, 7)>
>>> c.evaluate(6)
True
>>> c.evaluate(4)
False

'''
    return Comparison('inside', (value0, value1), units=units)
#--- End: def

def outside(value0, value1, units=None):
    '''
    
Return an object for testing whether a variable is is outside the
given range.

:Parameters:

    value0 : object
         The lower bound of the range which a variable is to be
         compared with.

    value1 : object
         The upper bound of the range which a variable is to be
         compared with.

:Returns: 

    out : Comparison
        A Comparison object which will evaluate whether or not the
        comparison evaluates to True.

**Examples**

>>> c = cf.outside(5)
>>> c
<CF Comparison: x outside (5, 7)>
>>> c.evaluate(4)
True
>>> c.evaluate(6)
False

'''
    return Comparison('outside', (value0, value1), units=units)
#--- End: def
