import numpy
from .utils    import parse_indices
from .variable import Variable, SubsetVariable

# ====================================================================
#
# Coordinate object
#
# ====================================================================

class Coordinate(Variable):
    '''

A CF dimension or auxiliary coordinate construct.

The coordinate array's bounds (if present) are stored in the `bounds`
attribute. The `climatology` attribute indicates if the bounds are
intervals of climatological time.

If the coordinate is connected to a transform construct then a pointer
to the transformation is stored in the `transform` attribute.

'''

    _special_attributes = Variable._special_attributes.union(set(('bounds',
                                                                  'climatology',
                                                                  'transform',
                                                                  )))

    def _binary_operation(self, other, method):
        '''
'''
        self_bounds = getattr(self, 'bounds', None)

        if isinstance(other, self.__class__):
            other_bounds = getattr(other, 'bounds', None)
            if other_bounds:
                if not self_bounds:
                    raise TypeError("asdfsdfdfds 0000")
            elif self_bounds:
                raise TypeError("asdfsdfdfds 00001")

        elif isinstance(other, (float, int, long)):
            other_bounds = other

        elif isinstance(other, numpy.ndarray):
            if self_bounds:
                if other.size > 1:
                    raise TypeError("4444444")
                other_bounds = other
        #-- End: if

        new = super(Coordinate, self)._binary_operation(other, method)

        if self_bounds:
            bounds = self_bounds._binary_operation(other_bounds, method)

        inplace = method[2] == 'i'

        if not inplace:
            new.bounds = bounds
            return new
        else: 
            self.bounds = bounds
            return self
    #--- End: def

    def _get_direction(self):
        '''
    
Return True if a coordinate is increasing, otherwise return False.

A coordinate is considered to be increasing if its *raw* data array
values are increasing in index space or if it has no data. If the
direction can not be inferred from the coordinate's data, then the
`positive` attribute is used, if present, or the coordinate's units.

The direction is inferred from the coordinate's data array values or
its from coordinates. It is not taken directly from its Data object.

:Returns:

    out : bool
        Whether or not the coordinate is increasing.
        
**Examples**

>>> c.array
array([  0  30  60])
>>> c._get_direction()
True
>>> c.array
array([15])
>>> c.bounds.array
array([  30  0])
>>> c._get_direction()
False

'''
        if self.ndim > 1:
            raise ValueError(
                "Can't find direction of a multidimensional coordinate")
        
        if hasattr(self, 'Data'):
            c = self.Data
        else:
            return True
        
        # Infer the direction from the sign of the difference between
        # its first two elements
        if c.size > 1:
            c = c[:2].varray
            return c[0] < c[1]
        #--- End: if

        # Still here? Then infer the direction from the dimension
        # coordinate's bounds, if it has any
        if hasattr(self, 'bounds'):
            b = self.bounds.Data
            indices = (0,)*(b.ndim-1) + (slice(None),)
            b = b[indices].varray[indices]
            return b[0] < b[1]
        #--- End: if

        # Still here? Then infer the direction from the dimension
        # coordinate's positive attribute, if it has one.
        if hasattr(self, 'positive'):
            if self.positive.lower().startswith('d'): # d for down
                return False
        #--- End: if

        # Still here? Then infer the direction from the units
        return not self.Units.is_pressure
    #--- End: def

    def _equivalent_data(self, other, dim_name_map,
                         rtol=None, atol=None,
                         _check_units=True):
        '''

'''
        # Compare the data arrays
        if not super(Coordinate, self)._equivalent_data(other, dim_name_map,
                                                        rtol=rtol, atol=atol,
                                                        _check_units=_check_units):
            return False

        # Compare the bounds' data arrays
        bounds       = getattr(self,  'bounds', None)
        other_bounds = getattr(other, 'bounds', None)
        if bounds is not None:
            if other_bounds is None:
                return False

            return bounds._equivalent_data(other_bounds, dim_name_map,
                                           rtol=rtol, atol=atol,
                                           _check_units=_check_units)
        elif other_bounds is not None:
            return False

        # Still here? Then the data are equivalent.
        return True
    #--- End: def

    def _set_Data_attributes(self, dimensions, directions):
        '''
        Set the metadata.    

:Parameters:

    dimensions : list

    directions : dict

:Returns:

    None

'''
        super(Coordinate, self)._set_Data_attributes(dimensions, directions)

        # ------------------------------------------------------------
        # Set the metadata for its bounds' data
        # ------------------------------------------------------------
        if hasattr(self, 'bounds'):

            c = self.Data
            b = self.bounds.Data
    
            # Units. Give them the coordinate's units. Note that we
            # don't want to change any units set on the bounds
            b.Units = c.Units.copy()
            if not b.Units:
                b.partitions[0].Units = b.Units.copy()  # dch check
                
            # Order
            dim_name_map = dict((dim, new_dim) 
                                for dim, new_dim in zip(b.order[:-1], dimensions))
            dim_name_map[b.order[-1]] = 'bounds'
            b.change_dimension_names(dim_name_map)
     
            # Direction
            if c.is_scalar:
                b.direction = {'bounds': c.direction}

            else:
                b.direction = c.direction.copy()
                if c.ndim == 1:
                    # 1-d coordinate
                    b.direction['bounds'] = c.direction[c.order[0]]
                else:
                    # N-d coordinate
                    b.direction['bounds'] = True
            #--- End: if

            for partition in b.partitions.flat():
                partition.direction = b.direction.copy()
        #--- End: if            

    #--- End: def
        
    @property
    def subset(self):
        '''

Return a new coordinate whose data and bounds are subsetted in a
consistent manner.

This attribute may be indexed to select a subset from dimension index
values.

**Subsetting by indexing**

Subsetting by dimension indices uses an extended Python slicing
syntax, which is similar numpy array indexing. There are two
extensions to the numpy indexing functionality:

* Size 1 dimensions are never removed.

  An integer index i takes the i-th element but does not reduce the
  rank of the output array by one.

* When advanced indexing is used on more than one dimension, the
  advanced indices work independently.

  When more than one dimension's slice is a 1-d boolean array or 1-d
  sequence of integers, then these indices work independently along
  each dimension (similar to the way vector subscripts work in
  Fortran), rather than by their elements.

**Examples**

'''
        return SubsetCoordinate(self)
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: Units
    # ----------------------------------------------------------------
    @property
    def Units(self):
        '''
The Units object containing the units of the data array.

'''
        return Variable.Units.fget(self)
    #--- End: def

    @Units.setter
    def Units(self, value):
        Variable.Units.fset(self, value)

        # Set the Units object on the bounds
        if hasattr(self, 'bounds') and hasattr(self.bounds, 'Units'):
            self.bounds.Units = value.copy()
    #--- End: def
    
    @Units.deleter
    def Units(self):
        Variable.Units.fdel(self)
        
        # Delete the Units object from the bounds
        if hasattr(self, 'bounds') and hasattr(self.bounds, 'Units'):
            del self.bounds.Units
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: units (a special attribute)
    # ----------------------------------------------------------------
    # DCH possible inconsistency when setting self.Units.units
    @property
    def units(self):
        '''

The units CF attribute.

This attribute is a mirror of the units stored in the `Units`
attribute.

**Examples**

>>> c.units = 'degrees_east'
>>> c.units
'degree_east'
>>> del c.units

'''
        return Variable.units.fget(self)
    #--- End: def

    @units.setter
    def units(self, value):
        Variable.units.fset(self, value)

        # Set the units on the bounds        
        if hasattr(self, 'bounds') and hasattr(self.bounds, 'Units'):
            self.bounds.units = value 
    #--- End: def
    
    @units.deleter
    def units(self):
        Variable.units.fdel(self)
        
        # Delete the units from the bounds
        if hasattr(self, 'bounds') and hasattr(self.bounds, 'Units'):
            del self.bounds.units
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: axis
    # ----------------------------------------------------------------
    @property
    def axis(self):
        '''

The axis CF attribute.

**Examples**

>>> c.axis = 'Y'
>>> c.axis
'Y'
>>> del c.axis

'''
        return self.getattr('axis')
    #--- End: def
    @axis.setter
    def axis(self, value): self.setattr('axis', value)
    @axis.deleter
    def axis(self):        self.delattr('axis')

    # ----------------------------------------------------------------
    # Property attribute: positive (a simple attribute)
    # ----------------------------------------------------------------
    @property
    def positive(self):
        '''

The positive CF attribute.

**Examples**

>>> c.positive = 'up'
>>> c.positive
'up'
>>> del c.positive

'''
        return self.getattr('positive')
    #--- End: def

    @positive.setter
    def positive(self, value): self.setattr('positive', value)
    @positive.deleter
    def positive(self):        self.delattr('positive')

    # ----------------------------------------------------------------
    # Property attribute: calendar (a special attribute)
    # ----------------------------------------------------------------
    @property
    def calendar(self):
        '''
The calendar CF attribute.

This attribute is a mirror of the calendar stored in the `Units`
attribute.

**Examples**

>>> c.calendar = 'noleap'
>>> c.calendar
'noleap'
>>> del c.calendar

'''
        return Variable.calendar.fget(self)
    #--- End: def

    @calendar.setter
    def calendar(self, value):
        Variable.calendar.fset(self, value)
        
        # Set the calendar on the bounds
        if hasattr(self, 'bounds') and hasattr(self.bounds, 'Units'):
            self.bounds.calendar = value
    #--- End: def

    @calendar.deleter
    def calendar(self):
        Variable.calendar.fdel(self)
        
        # Delete the calendar from the bounds
        if hasattr(self, 'bounds' and hasattr(self.bounds, 'Units')):
            del self.bounds.calendar
    #--- End: def
   
    # ----------------------------------------------------------------
    # Property attribute: bounds (a special attribute)
    # ----------------------------------------------------------------
    @property
    def bounds(self):
        '''

The `CoordinateBounds` object containing the coordinate array's cell
bounds.

**Examples**

>>> c.bounds
<CF CoordinateBounds: >

'''
        return self._get_special_attr('bounds')
    #--- End: def
    @bounds.setter
    def bounds(self, value): self._set_special_attr('bounds', value)
    @bounds.deleter
    def bounds(self):        self._del_special_attr('bounds')

    # ----------------------------------------------------------------
    # Property attribute: climatology (a special attribute)
    # ----------------------------------------------------------------
    @property
    def climatology(self):
        '''

Indicator for the coordinate array's bounds representing
climatological time intervals.

If not set then bounds are assumed to not represent climatological
time intervals.

**Examples**

>>> c.climatology = True
>>> c.climatology
True
>>> del c.climatology

'''
        return self._get_special_attr('climatology')
    #--- End: def
    @climatology.setter
    def climatology(self, value): self._set_special_attr('climatology', value)
    @climatology.deleter
    def climatology(self):        self._del_special_attr('climatology')
   
    # ----------------------------------------------------------------
    # Property attribute: transform
    # ----------------------------------------------------------------
    @property
    def transform(self):
        '''

Pointer to coordinate's transform.

'''
        return self._getter('transform')
    #--- End: for
    @transform.setter
    def transform(self, value): self._setter('transform', value)
    @transform.deleter
    def transform(self):        self._deleter('transform')

    @property
    def varray(self):
        '''

Return a numpy view of the data.

Making changes to elements of the returned view changes the underlying
data. Refer to :obj:`numpy.ndarray.view`.
    
**Examples**

>>> a = c.varray
>>> type(a)
<type 'numpy.ndarray'>
>>> a
array([0, 1, 2, 3, 4])
>>> a[0] = 999
>>> c.varray
array([999, 1, 2, 3, 4])

'''
        if hasattr(self, 'bounds'):
            self.bounds.varray
       
        return super(Coordinate, self).varray
    #--- End: if
  
    def chunk(self, chunksize=None, extra_boundaries=None, adim=None):
        '''
'''         
        extra_boundaries, adim = super(Coordinate, self).chunk(chunksize, 
                                                               extra_boundaries,
                                                               adim)
        if not extra_boundaries:
            return None, None
        
        if hasattr(self, 'bounds'):
            self.bounds.chunk(None, extra_boundaries, adim)

        return extra_boundaries, adim
    #--- End: def

    def copy(self):
        '''

Return a deep copy.

Equivalent to ``copy.deepcopy(c)``

:Returns:

    out : 
        The deep copy.

**Examples**

>>> d = c.copy()

'''
        new = super(Coordinate, self).copy(_omit_special=('bounds',))

        if hasattr(self, 'bounds'):
            new.bounds = self.bounds.copy()

        return new
    #--- End: def

    def dump(self, id=None, omit=()): 
        '''

Return a string containing a full description of the coordinate.

:Parameters:

    id: str, optional
       Set the common prefix of variable component names. By default
       the instance's class name is used.

    omit : sequence of strs
        Omit the given attributes from the description.

:Returns:

    out : str
        A string containing the description.

**Examples**

>>> x = c.dump()
>>> print c.dump()
>>> print c.dump(id='time_coord')
>>> print c.dump(omit=('long_name',))

'''
        if id is None:
            id = self.__class__.__name__       

        default = getattr(self, 'ncvar', '')

        string = ['%s coordinate' % self.name(default=default, long_name=True)]
        string.append(''.ljust(len(string[0]), '-'))
        
        string.append(super(Coordinate, self).dump(id=id, omit=omit+('bounds',
                                                                     'transform',
                                                                     )))

        if hasattr(self, 'bounds'):
           string.append(self.bounds.dump(id=id+'.bounds', omit=omit+('bounds',
                                                                      'Units')))

        if hasattr(self, 'transform'):
            string.append('%s.transform = %s\n' % (id,
                                                   repr(self.transform)))
            
        return '\n'.join(string)
    #--- End: def

    def _insert_data(self, other, indices,
                     ADim, ADim_direction, dim_name_map):
        '''
'''
        # Insert the coordinate's data
        super(Coordinate, self)._insert_data(other, indices, 
                                             ADim, ADim_direction,
                                             dim_name_map)
        
        # Insert the coordinate's bounds, if it has any
        if hasattr(self, 'bounds'):
            self.bounds._insert_data(other.bounds, indices,
                                     ADim, ADim_direction, dim_name_map)
        #--- End: if

    #--- End: def

    def expand_dims(self, axis, dim, direction):
        '''
axis is an integer
dim is a string
direction is a boolean

'''
        # Expand the requested dimensions in the coordinate's data
        super(Coordinate, self).expand_dims(axis, dim, direction)

        # Expand the requested dimensions in the coordinate's bounds,
        # if it has any
        if hasattr(self, 'bounds'):
            if axis < 0:
                axis += self.ndims

            self.bounds.expand_dims(axis, dim, direction)
    #--- End: def

    def reverse_dims(self, axes=None):
        '''
'''  
        # Reverse the requested dimensions in the coordinate's data
        axes = super(Coordinate, self).reverse_dims(axes)

        # ------------------------------------------------------------
        # Reverse the requested dimensions in the coordinate's bounds,
        # if it has any.
        #
        # As per section 7.1 in the CF conventions: i) if the
        # coordinate is 0 or 1 dimensional then reverse all dimensions
        # (including the the trailing size 2 dimension); ii) if the
        # coordinate has 2 or more dimensions then do not reverse the
        # trailing size 2 dimension.
        # ------------------------------------------------------------
        if hasattr(self, 'bounds'):
            if self.bounds.ndim <= 2:
                axes = None   # None means reverse all dimensions
            self.bounds.reverse_dims(axes)
    #--- End: def

    def squeeze(self, axes=None):
        '''

Remove size 1 dimensions from the coordinate's data array and bounds
in place.

:Parameters:

    axes : int or sequence of ints, optional
        The size 1 axes to remove. By default, all size 1 axes are
        removed. Size 1 axes for removal may be identified by the
        integer positions of dimensions in the data array.

:Returns:

    None

**Examples**

>>> c.squeeze()
>>> c.squeeze(1)
>>> c.squeeze([1, 2])

'''  
        # Squeeze the requested dimensions in the coordinate's data
        super(Coordinate, self).squeeze(axes)

        # Squeeze the requested dimensions in the coordinate's bounds,
        # if it has any
        if hasattr(self, 'bounds'):
            self.bounds.squeeze(axes)
    #--- End: def
          
    def transpose(self, axes=None):
        '''

Permute the dimensions of the coordinate's data array and bounds in
place.

:Parameters:

    axes : sequence of ints, optional
        The new order of the data array. By default, reverse the
        dimensions' order, otherwise the axes are permuted according
        to the values given. The values of the sequence comprise the
        integer positions of the dimensions in the data array in the
        desired order.

:Returns:

    None

**Examples**

>>> c.transpose()
>>> c.ndim
3
>>> c.transpose([1, 2, 0])

'''  
        # Transpose the coordinate's data
        super(Coordinate, self).transpose(axes=axes)

        # Tran-pose the coordinate's bounds, if it has any.
        if hasattr(self, 'bounds'):
            axes = list(axes)
            axes.append(self.ndim)
            self.bounds.transpose(axes=axes)
    #--- End: def

#--- End: class


# ====================================================================
#
# SubsetCoordinate object
#
# ====================================================================

class SubsetCoordinate(SubsetVariable):

    __slots__ = []

    def __getitem__(self, indices):
        '''
x.__getitem__(index) <==> x[index]

'''
        coord = self.variable

        # Copy the coordinate
        new = coord.copy()

        # Parse the index (so that it's ok for appending the bounds
        # index if required)
        indices = tuple(parse_indices(coord, indices))
    
        new.Data = coord.Data[indices]

        # Slice the bounds, if there are any
        if hasattr(coord, 'bounds'):
            new.bounds = coord.bounds.subset[indices+(slice(None),)]
      
        # Return the new coordinate
        return new
    #--- End: def

#--- End: class
