import numpy
from copy     import copy
from operator import mul
from .units          import Units
from .utils          import parse_indices, CHUNKSIZE, RTOL, ATOL, array_allclose
from .partition      import Partition
from .partitionarray import PartitionArray

# the direction values in self are arbitrary, but athe directions in
# each partition are relative to self.

# ====================================================================
#
# Data object
#
# ====================================================================

class Data(object):
    '''

An N-dimensional data array with units and masked values.

* Contains an N-dimensional, indexable and broadcastable array with
  many similarities to a numpy array.

* Contains the units of the array elements.

* Supports masked arrays, regardless of whether or not it was
  initialized with a masked array.

* Uses Large Amounts of Massive Arrays (LAMA) functionality to store
  and operate on arrays which are larger then the available memory.


**Indexing**

A data array is indexable in a similar way to numpy array indexing but
for two important differences:

* Size 1 dimensions are never removed.

  An integer index i takes the i-th element but does not reduce the
  rank of the output array by one.

* When advanced indexing is used on more than one dimension, the
  advanced indices work independently.

  When more than one dimension's slice is a 1-d boolean array or 1-d
  sequence of integers, then these indices work independently along
  each dimension (similar to the way vector subscripts work in
  Fortran), rather than by their elements.

**Examples**

>>> d.shape
[12, 19, 73, 96]
>>> d[0, :, [0,1], [0,1,2]].shape
[1, 19, 2, 3]


**Conversion to a numpy array**

The data array may be converted to either a numpy array view or an
independent numpy array of the underlying data with the `varray` and
`array` attributes respectively. Changing a numpy array view in place
will also change the data array. Note that the numpy array created
with the `array` or `varray` attribute forces all of the data to be
read into memory at the same time, which may not be possible for very
large arrays.

'''
    __slots__ = ('partitions',)
    
    def __init__(self, data=None, units=None, _FillValue=None, chunk=True):
        '''

**Initialization**

:Parameters:
    data : array-like, optional
        The data for the array.

    units : str or Units, optional
        The units of the data.

    _FillValue : object, optional 
        The fill value of the data. By default the numpy fill value
        appropriate to the data type will be used.

    chunk : bool, optional
        If True then the data array will be partitioned if it is
        larger than the chunk size.

**Examples**

>>> d = cf.Data(5)
>>> d = cf.Data([1,2,3], units='K')
>>> import numpy   
>>> d = cf.Data(numpy.arange(10).reshape(2,5), units=cf.Units('m/s'), _FillValue=-999)
>>> d = cf.Data(('f', 'l', 'y'))

'''
        # Set units if they're not already a Units object
        if units is None:            
            units = Units()

        if isinstance(units, basestring):
            try:
                units = Units(units)
            except TypeError:
                raise ValueError(
"Unsupported unit in %s initialization: '%s'" %
(self.__class__.__name__, units))
        #--- End: if

        if data is None:
            return

        if isinstance(data, (list, float, tuple, int, long, basestring)):
            data = numpy.array(data)
        
        shape = list(data.shape)
        ndim  = data.ndim
        
        order = [str(i) for i in xrange(ndim)]
        
        if not ndim:
            # Scalar data
            direction = True
        else:
            # 1 or more dimensional data
            direction = dict(((dim, True) for dim in order))

        self.partitions = PartitionArray(
            [Partition(location  = [(0,n) for n in shape],
                       shape     = shape[:],
                       order     = order[:],
                       direction = copy(direction),
                       Units     = units.copy(),
                       part      = [],
                       data      = data)
             ],
            order      = order,
            shape      = shape,
            size       = data.size,
            Units      = units.copy(),
            ndim       = ndim,
            direction  = direction,
            adims      = [],
            _FillValue = _FillValue,
            )

        if chunk:
            self.chunk()
    #--- End: def

    def __deepcopy__(self, memo):
        '''

Used if copy.deepcopy is called on the variable.

''' 
        return self.copy()
    #--- End: def

    def __nonzero__(self):
        '''
x.__nonzero__() <==> x != 0

'''
        if self.size == 1:
            return self.varray.__nonzero__()

        raise ValueError(
"The truth value of Data with more than one element is ambiguous. Use d.any() or d.all()")
    #--- End: def

    def __repr__(self):
        '''
x.__repr__() <==> repr(x)

'''
        return '<CF %s: Units = %s>' % (self.__class__.__name__, self.Units)
    #--- End: def

    def __str__(self):
        '''
x.__str__() <==> str(x)

'''
        return str({'dtype'     : getattr(self, 'dtype', None),
                    'shape'     : self.shape,
                    'order'     : self.order,
                    'ndim'      : self.ndim,
                    'size'      : self.size,
                    'direction' : self.direction,
                    'adims'     : self.adims,
                    'Units'     : self.Units,
                    '_FillValue': self._FillValue,
                    })
    #--- End: def

    def __getitem__(self, indices):
        '''
x.__getitem__(indices) <==> x[indices]

'''

        dim2position = dict([(dim, i) for i, dim in enumerate(self.order)])
        
#        print  'indices        =', indices
        indices, reverse_dimensions = _parse_indices(self, indices)
#        print  'parsed_indices =', indices

        # If it's a scalar, then just copy it. ??? dch trapped in _parse_indices??
        if not self.ndim and not indices:
            return self.copy()
        
        new            = type(self)()
        
        new.partitions = PartitionArray(_overlapping_partitions(self.partitions, 
                                                                indices, 
                                                                dim2position))

        new.shape      = map(_size_of_index, indices, self.shape)
        new.size       = long(reduce(mul, new.shape, 1))
        new.order      = self.order[:]
        new.direction  = copy(self.direction)
        new.Units      = self.Units.copy()
        new.adims      = self.adims[:]
        new._FillValue = self._FillValue
        new.ndim       = self.ndim

        new.set_location_map()
        
        # reverse dims with flipped slices
        if reverse_dimensions:
            new.reverse_dims(reverse_dimensions)
            
        # Remove null aggregating dimensions
        while (len(new.partitions) == 1 and
               not isinstance(new.partitions[0], Partition)):
            new.partitions._list = new.partitions._list[0]
            new.partitions.adims.pop()
        #--- End: while           
        
        return new
    #--- End: def

    def __setitem__(self, indices, value):
        '''
x.__setitem__(i, y) <==> x[i]=y

'''            
        def _mirror(index, size):
            '''

:Parameters:

    index : slice

    size : int

:Returns:

    out : slice

'''
            start, stop, step = index.indices(size)
            size -= 1
            start = size - start
            stop  = size - stop
            if stop < 0:
                stop = None
                
            return slice(start, stop, -1)
        #--- End: def

        indices, reverse_dimensions = _parse_indices(self, indices)

        # Find out if the value is logically scalar. If it is then
        # everything is much simpler later.
        scalar_value = None
        if isinstance(value, numpy.ndarray):
            if value.size == 1:                
                scalar_value = numpy.array(value.item())
            else:
                value = type(self)(value, units=self.Units, chunk=False)
                
        elif isinstance(value, Data):
            if value.size == 1:
                scalar_value = numpy.array(value.array.item())
            else:
                value.to_memory()

        else:
            scalar_value = numpy.array(value)
        #--- End: if

        order     = self.order
        direction = self.direction
        units     = self.Units

        save           = self.save_to_disk()
        revert_to_file = False
            
        if scalar_value is not None:
            for partition in self.partitions.flat():
                
                partition_indices, shape = _partition_overlaps(partition,
                                                               indices)
                
                if not partition_indices:
                    # This partition is not in the slice
                    continue
                
                partition.conform(order, direction, units, 
                                  save, revert_to_file)
                
                # Update the partition's data
                partition.data[tuple(partition_indices)] = scalar_value

                partition.close()
            #--- End: for

            return

        else:
            shape0 = map(_size_of_index, indices, self.shape)
            self_ndim  = self.ndim
            value_ndim = value.ndim
            align_offset = self_ndim - value_ndim
            if align_offset >= 0:
                # self has more dimensions than other
                shape0   = shape0[align_offset:]
                shape1   = value.shape
                ellipsis = None 

                reverse_dimensions = [i - align_offset
                                      for i in reverse_dimensions
                                      if i >= align_offset]                
            else:
                # value has more dimensions than self
                v_align_offset = -align_offset
                if value.shape[:v_align_offset] != [1] * v_align_offset:
                    raise ValueError(
"output operand requires a reduction, but reduction is not enabled")
                shape1       = value.shape[v_align_offset:]
                ellipsis     = Ellipsis
                align_offset = 0
            #--- End: if

            # Find out which of the dimensions of value are to be
            # broadcast, and those which are not.
            base_value_indices       = []
            non_broadcast_dimensions = []

            for i, (a, b) in enumerate(zip(shape0, shape1)):
                if b == 1:
                    base_value_indices.append(slice(None))
                elif a == b and b != 1:
                    base_value_indices.append(None)
                    non_broadcast_dimensions.append(i)
                else:
                    # The broadcasting rules have been violated
                    raise ValueError(
"output operand requires a reduction, but reduction is not enabled")
            #--- End: for

            #
            previous_location = ((-1,),) * self_ndim
            start             = [0]      * value_ndim
        
            for partition in self.partitions.flat():
                
                partition_indices, shape = _partition_overlaps(partition,
                                                               indices)
                
                if not partition_indices:
                    # This partition is not in the slice
                    continue
                
                # Find which elements of value apply to this partition
                value_indices = base_value_indices[:]

                for i in non_broadcast_dimensions:

                    j                  = i + align_offset
                    location           = partition.location[j][0]
                    reference_location = previous_location[j][0]

                    if location > reference_location:
                        stop             = start[i] + shape[j]
                        value_indices[i] = slice(start[i], stop)
                        start[i]         = stop

                    elif location == reference_location:
                        value_indices[i] = previous_slice[i]

                    elif location < reference_location:
                        stop             = shape[j]
                        value_indices[i] = slice(0, stop)
                        start[i]         = stop
                #--- End: for

                for i in reverse_dimensions:
                    value_indices[i] = _mirror(value_indices[i], shape1[i])

                if ellipsis:
                    value_indices.insert(0, ellipsis)

                v = value[tuple(value_indices)].varray
                        
                if not save:
                    v = v.copy()

                previous_location = partition.location
                previous_slice    = value_indices
                
                partition.conform(order, direction, units, 
                                  save, revert_to_file)

                 # Update the partition's data
                partition.data[tuple(partition_indices)] = v
                 
                partition.close()
            #--- End: For
                
            return

        #--- End: if

    #--- End: def
    
    def chunk(self, chunksize=None, extra_boundaries=None, chunk_dims=None):
        '''

:Parameters:

    chunksize : int, optional

    extra_boundaries : sequence of lists or tuples, optional

    chunk_dims : sequence of lists or tuples, optional

:Returns:

    extra_boundaries, chunk_dims : {list, list}

**Examples**

>>> d.chunk()
>>> d.chunk(100000)
>>> d.chunk(extra_boundaries=([3, 6],), chunk_dims=['dim0'])
>>> d.chunk(extra_boundaries=([3, 6], [40, 80]), chunk_dims=['dim0', 'dim1'])

'''
        if extra_boundaries is None:

            extra_boundaries = []
            chunk_dims       = []

            if not self.save_to_disk():
                return [], []
            
            # Assumption: there is only one partition, hmmm, could do better
            if not chunksize:
                chunksize = CHUNKSIZE()

            n_chunks = int(self.size*self.dtype.itemsize / chunksize + 0.5)

            if not n_chunks:
                return [], []

            for size, dim in zip(self.shape, self.order):
                if size <= n_chunks:
                    extra_boundaries.append(range(1, size))
                    chunk_dims.append(dim)
                    n_chunks -= size
                else:
#                    start = size // (n_chunks+1)
                    start = int(size/float(n_chunks))
                    extra_boundaries.append(range(start, size, start))
                    chunk_dims.append(dim)
                    break
            #--- End: for        
        #--- End: if
            
        for extra_bounds, dim in zip(extra_boundaries, chunk_dims):
             if dim not in self.adims:
                self.expand_aggregating_dims(dim)            

             self.add_partitions(extra_bounds, dim)
        #--- End: for

        return extra_boundaries, chunk_dims
    #--- End: def
    
    def _combine(self, other, method):
        '''

'''
           
        if not self.Units and not other.Units:
            return self.Units

        if not other.Units or not self.Units:
            raise TypeError("Can't combine %s with %s" %
                            (repr(self.Units), repr(other.Units)))

        method_type  = method[-5:-2]

        if method_type in ('mul', 'div'):
            return getattr(self.Units, method)(other.Units)
          
        if method_type in ('sub', 'add'):
          if self.Units.equivalent(other.Units):
              other.Units = self.Units
              if self.Units.is_reftime and other.Units.is_reftime:
                  return getattr(self.Units, method)(other.Units)
              else:
                  return self.Units.copy()
          else:
              if self.Units.is_reftime and other.Units.is_time:
                  other.Units  = Units(self.Units._rtime.units)
                  return self.Units.copy()
              else:
                  # Raise an exception for an illegal combination of
                  # units
                  getattr(self.Units, method)(other.Units)
        #--- End: if

        if method_type in ('_eq', '_ne', '_lt', '_le', '_gt', '_ge'):
            if self.Units.equivalent(other.Units):
                other.Units = self.Units
                return Units()
            else: 
                raise TypeError("Can't compare %s to %s" %
                                (repr(self.Units), repr(other.Units)))
        #--- End: if

        if method_type in ('and', '_or', 'xor', 'ior'):
            if self.Units.equivalent(other.Units):
                other.Units = self.Units
                return self.Units.copy()
            else: 
                raise TypeError("Can't compare %s to %s" %
                                (repr(self.Units), repr(other.Units)))
        #--- End: if

        if method_type == 'pow':
            if other.size > 1:
                raise TypeError(
"Can't raise %s to the power of an array with more than one element" %
self.__class__.__name__)

            if other.Units.equivalent(Units('1')):
                other.Units = Units('1')
            else:
                # Raise an exception for an illegal combination of
                # units
                getattr(self.Units, method)(other.Units)

            return self.Units ** other.varray.item()                    
        #--- End: if

        raise ValueError("doo bee doo")
    #--- End: def

    def _binary_operation(self, other, method):
        '''

Implement binary arithmetic and comparison operations with the numpy
broadcasting rules.

It is intended to be called by the binary arithmetic and comparison
methods, such as __sub__() and __lt__().

:Parameters:

    other : object
        The object on the right hand side of the operator.

    method : str
        The binary arithmetic or comparison method name (such as
        "__imul__" or "__ge__").

:Returns:

    new : Data
        A new Data object, or if the operation was in place, the same
        Data object.

**Examples**

>>> d = cf.Data([0, 1, 2, 3])
>>> e = cf.Data([1, 1, 3, 4])

>>> f = d._binary_operation(e, '__add__')
>>> f.array
array([1 2 5 7])

>>> e = d._binary_operation(e, '__lt__')
>>> e.array
array([ True False  True  True])

>>> d._binary_operation(2, '__imul__')
>>> d.array
array([0 2 4 6])

'''
#        other_Units  = None
        inplace      = method[2] == 'i'
        method_type  = method[-5:-2]

        if isinstance(other, self.__class__):
            # --------------------------------------------------------
            # other is Data object
            # --------------------------------------------------------
            other = other.copy()            

            result_Units = self._combine(other, method)

#            print 'jerer'
#            if method_type == 'pow':
#                if other.size > 1:
#                    raise TypeError, "Can't raise %s to the power of an array \
#with more than one element" % self.__class__.__name__
#                elif not other.Units.equivalent(Units('1')):
#                    # Raise a meaningful exception for an illegal
#                    # combination of units
#                    getattr(self.Units, method)(other.Units)
#                    
#                other.Units  = Units('1')
#                result_Units = self.Units ** other.varray.item()                    
#
#            elif method_type == 'mul':
#                result_Units = self.Units * other.Units
#
#            elif method_type == 'div':
#                result_Units = self.Units / other.Units
#
#            elif method_type == 'sub':
#                if self.Units.equivalent(other.Units):
#                    other.Units = self.Units
#                    if self.Units.is_reftime and other.Units.is_reftime:
#                        result_Units = self.Units - other.Units
#                    else:
#                        result_Units = self.Units.copy()
#                else:
#                    if self.Units.is_reftime and other.Units.is_time:
#                        other.Units  = Units(self.Units._rtime.units)
#                        result_Units = self.Units.copy()
#                    else:
#                        # Raise a meaningful exception for an illegal
#                        # combination of units
#                        getattr(self.Units, method)(other.Units)
#
#            elif method_type == 'add':
#                if self.Units.equivalent(other.Units):
#                    other.Units  = self.Units
#                    result_Units = self.Units.copy()
#                elif not other.Units:
#                    result_Units = self.Units.copy()
#                elif not self.Units:
#                    result_Units = other.Units.copy()
#                else:
#                    # Raise a meaningful exception for an illegal
#                    # combination of units
#                    getattr(self.Units, method)(other.Units)
#
#            else:
#                # being here means that the result are 
#                if self.Units.equivalent(other.Units):
#                    other.Units  = self.Units
#                    result_Units = Units()
#                elif not other.Units or not self.Units:
#                    result_Units = Units()
#                else:
#                    # Raise a meaningful exception for an illegal
#                    # combination of units
#                    getattr(self.Units, method)(other.Units)
#            #--- End: if                    

        elif isinstance(other, (float, int, long)):
            # --------------------------------------------------------
            # other is a number
            # --------------------------------------------------------
            if method_type == 'pow':
                result_Units = self.Units ** other
            elif method_type in ('_eq', '_ne', '_lt', '_le', '_gt', '_ge'):
                result_Units = Units()              
            else:
                result_Units = self.Units.copy()

            other = type(self)(other)
            
        elif isinstance(other, numpy.ndarray):
            # --------------------------------------------------------
            # other is a numpy array
            # --------------------------------------------------------
            if method_type == 'pow':
                if other.size > 1:
                    raise TypeError(
"Can't raise %s to the power of an array with more than one element" %
self.__class__.__name__)

                result_Units = self.Units ** other.varray.item()
            elif method_type in ('_eq', '_ne', '_lt', '_le', '_gt', '_ge'):
                result_Units = Units()
            else:
                result_Units = self.Units.copy()

            other = type(self)(other)
   
        elif isinstance(other, (list, tuple)):
            # --------------------------------------------------------
            #
            # --------------------------------------------------------
            if method_type == 'pow':
                if len(other) > 1:
                    raise TypeError(
"Can't raise %s to the power of a sequence with more than one element" %
self.__class__.__name__)

                result_Units = self.Units ** other.varray.item()
            elif method_type in ('_eq', '_ne', '_lt', '_le', '_gt', '_ge'):
                result_Units = Units()
            else:
                result_Units = self.Units.copy()

            other = type(self)(other)
   
        else:
            # --------------------------------------------------------
            # other is something else
            # --------------------------------------------------------
            return NotImplemented
        #--- End: if

        # ------------------------------------------------------------
        # Bring other into memory, if appropriate.
        # ------------------------------------------------------------
        other.to_memory()
        
        # ------------------------------------------------------------
        # Find which dimensions need to be broadcast in one or other
        # of the arrays.
        #
        # For each common dimension, the 'broadcast_indices' list will
        # have a value of None if there is no broadcasting required
        # (i.e. the two arrays have the same size along that
        # dimension) or a value of slice(None) if broadcasting is
        # required (i.e. the two arrays have the different sizes along
        # that dimension and one of the sizes is 1).
        #
        # For example, if c.shape is [7,1,6,1,5] and d.shape is
        # [6,4,1] then broadcast_indices will be
        # [None,slice(None),slice(None)].
        #
        # The indices to d which correspond to a partition of c, are
        # the relevant subset of partition.indices updated with the
        # non None elements of the broadcast_indices list.
        #
        # In the above example, if a partition of c has
        # partition.indices of
        # (slice(0,3),slice(0,1),slice(2,4),slice(0,1),slice(0,5)),
        # the relevant subset of these is partition.indices[2:] and
        # the corresponding indices to d are
        # (slice(2,4),slice(None),slice(None))
        # ------------------------------------------------------------
        align_offset = self.ndim - other.ndim
        if align_offset >= 0:
            # self has more dimensions than other
            shape0        = self.shape[align_offset:]
            shape1        = other.shape
            ellipsis      = None
            new_shape     = self.shape[:align_offset]
            new_order     = self.order[:]
            new_direction = copy(self.direction)
        else:
            # other has more dimensions than self
            align_offset  = -align_offset
            shape0        = self.shape
            shape1        = other.shape[align_offset:]
            ellipsis      = (Ellipsis,)
            new_shape     = other.shape[:align_offset]
            new_order     = [str(i) for i in xrange(align_offset)] + self.order[:]

            if self.is_scalar:
                new_direction = {}
            else:
                new_direction = self.direction.copy()

            for dim in other.order[:align_offset]:
                new_direction[dim] = other.direction[dim]

            align_offset  = 0
        #--- End: if
            
        broadcast_indices = []
        
        for a, b in zip(shape0, shape1):
            if a == b:
                new_shape.append(a)
                broadcast_indices.append(None)
                continue
            
            if a > 1 and b == 1:
                new_shape.append(a)
            elif b > 1 and a == 1:
                new_shape.append(b)
            else:
                raise ValueError("shape mismatch")
            
            broadcast_indices.append(slice(None))
        #--- End: for
            
        # ------------------------------------------------------------
        # Create a Data object which just contains the metadata for
        # the result. If we're doing a binary arithmetic operation
        # then result will get filled with data and returned. If we're
        # an augmented arithmetic assignment then we'll use result's
        # metadata to update self.
        # ------------------------------------------------------------
        new_ndim = len(new_shape)

        if new_shape != self.shape:
            set_location_map = True
            dummy_location   = [None] * new_ndim
        else:
            set_location_map = False

        if not set_location_map:
            new_size = long(reduce(mul, new_shape, 1))
        else:
            new_size = self.size

        result            = self.copy()
        result._FillValue = self._FillValue
        result.adims      = self.adims[:]
        result.shape      = new_shape
        result.ndim       = new_ndim
        result.size       = new_size
        result.order      = new_order
        result.direction  = new_direction        

        # ------------------------------------------------------------
        # Set flags to control whether or not the data of result and
        # self should be kept in memory
        # ------------------------------------------------------------
        result_dtype = max(self.dtype, other.dtype)
        save_result  = result.save_to_disk(result_dtype.itemsize)

        if not inplace:
            # When doing a binary arithmetic operation (like
            # result=self+2), we need to decide whether or not to keep
            # self's data in memory.
            revert_to_file = True
            save_self      = self.save_to_disk()
        else:
            # When doing an augmented arithmetic assignment (like
            # self+=2), we don't need to keep self's original data in
            # memory.
            revert_to_file = False
            save_self      = False
        #--- End: if

        order     = self.order
        direction = self.direction
        units     = self.Units

        for partition_r, partition_s in zip(result.partitions.flat(),
                                            self.partitions.flat()):
            
            array0 = partition_s.conform(order, direction, units, 
                                         save_self, revert_to_file)    

            indices = tuple([                
                    (index 
                     if not broadcast_index else
                     broadcast_index) 
                    for index, broadcast_index in zip(partition_s.indices[align_offset:],
                                                      broadcast_indices)
                    ])
            
            # If other has more dimensions than self then we must
            # prepend an Ellipsis object to the indices
            if ellipsis:
                indices = ellipsis + indices
                
            # POSSIBLE ISSUE HERE: array1 could be much larger
            # than the chunk size.
            array1 = other[indices].varray

            if not inplace:
                partition = partition_r
                partition.update_from(partition_s)
            else:
                partition = partition_s

            partition.data      = getattr(array0, method)(array1)

            partition.Units     = result_Units.copy()
            partition.order     = new_order[:]
            partition.direction = copy(new_direction)

            if set_location_map:
                partition.location = dummy_location[:]
                partition.shape    = list(partition.data.shape)
            #--- End: if

            partition.close(save_result)            

            if not inplace:
                partition_s.close()
        #--- End: for

        if result.dtype == bool:
            result_Units = Units()

        if not inplace:
            result.Units = result_Units

            if set_location_map:
                result.set_location_map()

            return result
        else:
            # Update the metadata for the new master array in place
            self.Units      = result_Units
            self.shape      = new_shape
            self.ndim       = new_ndim
            self.order      = new_order
            self.size       = new_size
            self.direction  = new_direction        

            if set_location_map:
                self.set_location_map()

            return self
    #--- End: def

    def _unary_operation(self, operation):
        '''

Implement unary arithmetic operations.

It is intended to be called by the unary arithmetic methods, such as
__abs__().

:Parameters:

    operation : str
        The unary arithmetic method name (such as "__abs__").

:Returns:

    new : Data
        A new Data array.

**Examples**

>>> d = cf.Data([1, 2, -3, -4, -5])

>>> e = d._unary_operation('__abs__')
>>> e.array
array([1 2 3 4 5])

>>> e = d.__abs__()
>>> e.array
array([1 2 3 4 5])

>>> e = abs(d)
>>> e.array
array([1 2 3 4 5])

'''
        new = self.copy()

        order        = new.order
        direction    = new.direction
        units        = new.Units

        save           = new.save_to_disk()
        revert_to_file = True

        for partition in new.partitions.flat():
            data = partition.conform(order, direction, units, 
                                     save, revert_to_file)

            partition.data = getattr(data, operation)()

            partition.close()
        #--- End: for

        return new
    #--- End: def

    def __add__(self, other):
        '''
x.__add__(y) <==> x+y
        '''
        return self._binary_operation(other, '__add__')
    #--- End: def

    def __iadd__(self, other):
        '''
x.__iadd__(y) <==> x+=y
        '''
        return self._binary_operation(other, '__iadd__')
    #--- End: def

    def __radd__(self, other):
        '''
x.__radd__(y) <==> y+x
        '''
        return self._binary_operation(other, '__radd__')
    #--- End: def

    def __sub__(self, other):
        '''
x.__sub__(y) <==> x-y
        '''
        return self._binary_operation(other, '__sub__')
    #--- End: def

    def __isub__(self, other):
        '''
x.__isub__(y) <==> x-=y
        '''
        return self._binary_operation(other, '__isub__')
    #--- End: def

    def __rsub__(self, other):
        '''
x.__rsub__(y) <==> y-x
        '''
        return self._binary_operation(other, '__rsub__')
    #--- End: def

    def __mul__(self, other):
        '''
x.__mul__(y) <==> x*y
        '''
        return self._binary_operation(other, '__mul__')
    #--- End: def

    def __imul__(self, other):
        '''
x.__imul__(y) <==> x*=y
        '''
        return self._binary_operation(other, '__imul__')
    #--- End: def

    def __rmul__(self, other):
        '''
x.__rmul__(y) <==> y*x
        '''
        return self._binary_operation(other, '__rmul__')
    #--- End: def

    def __div__(self, other):
        '''
x.__div__(y) <==> x/y
        '''
        return self._binary_operation(other, '__div__')
    #--- End: def

    def __idiv__(self, other):
        '''
x.__idiv__(y) <==> x/=y
        '''
        return self._binary_operation(other, '__idiv__')
    #--- End: def

    def __rdiv__(self, other):
        '''
x.__rdiv__(y) <==> y/x
        '''
        return self._binary_operation(other, '__rdiv__')
    #--- End: def

    def __floordiv__(self, other):
        '''
x.__floordiv__(y) <==> x//y
        '''
        return self._binary_operation(other, '__floordiv__')
    #--- End: def

    def __ifloordiv__(self, other):
        '''
x.__ifloordiv__(y) <==> x//=y
        '''
        return self._binary_operation(other, '__ifloordiv__')
    #--- End: def

    def __rfloordiv__(self, other):
        '''
x.__rfloordiv__(y) <==> y//x
        '''
        return self._binary_operation(other, '__rfloordiv__')
    #--- End: def

    def __truediv__(self, other):
        '''
x.__truediv__(y) <==> x/y
        '''
        return self._binary_operation(other, '__truediv__')
    #--- End: def

    def __itruediv__(self, other):
        '''
x.__itruediv__(y) <==> x/=y
        '''
        return self._binary_operation(other, '__itruediv__')
   #--- End: def

    def __rtruediv__(self, other):
        '''
x.__rtruediv__(y) <==> y/x
        '''
        return self._binary_operation(other, '__rtruediv__')
    #--- End: def

    def __pow__(self, other, modulo=None):
        '''
x.__pow__(y) <==> x**y
    '''  
        if modulo is not None:
            raise NotImplementedError("3-argument power not supported for '%s'" %
                                      self.__class__.__name__)

        return self._binary_operation(other, '__pow__')
    #--- End: def

    def __ipow__(self, other, modulo=None):
        '''
x.__ipow__(y) <==> x**=y
    '''  
        if modulo is not None:
            raise NotImplementedError("3-argument power not supported for '%s'" %
                                      self.__class__.__name__)

        return self._binary_operation(other, '__ipow__')
    #--- End: def

    def __rpow__(self, other, modulo=None):
        '''
x.__ipow__(y) <==> x**=y
    '''  
        if modulo is not None:
            raise NotImplementedError("3-argument power not supported for '%s'" %
                                      self.__class__.__name__)

        return self._binary_operation(other, '__rpow__')
    #--- End: def

    def __eq__(self, other):
        '''
x.__eq__(y) <==> x==y
'''
        return self._binary_operation(other, '__eq__')
    #--- End: def

    def __ne__(self, other):
        '''
x.__ne__(y) <==> x!=y
'''
        return self._binary_operation(other, '__ne__')
    #--- End: def

    def __ge__(self, other):
        '''
x.__ge__(y) <==> x>=y
'''
        return self._binary_operation(other, '__ge__')
    #--- End: def

    def __gt__(self, other):
        '''
x.__gt__(y) <==> x>y
'''
        return self._binary_operation(other, '__gt__')
    #--- End: def

    def __le__(self, other):
        '''
x.__le__(y) <==> x<=y
'''
        return self._binary_operation(other, '__le__')
    #--- End: def

    def __lt__(self, other):
        '''
x.__lt__(y) <==> x<y
'''
        return self._binary_operation(other, '__lt__')
    #--- End: def

    def __and__(self, other):
        '''
x.__and__(y) <==> x&y
'''
        return self._binary_operation(other, '__and__')
    #--- End: def

    def __or__(self, other):
        '''
x.__or__(y) <==> x|y
'''
        return self._binary_operation(other, '__or__')
    #--- End: def

    def __abs__(self):
        '''
x.__abs__() <==> abs(x)
'''
        return self._unary_operation('__abs__')
    #--- End: def

    def __neg__(self):
        '''
x.__neg__() <==> -x
'''
        return self._unary_operation('__neg__')
    #--- End: def

    def __invert__(self):
        '''
x.__invert__() <==> ~x
'''
        return self._unary_operation('__invert__')
    #--- End: def

    def __pos__(self):
        '''
x.__pos__() <==> +x
'''
        return self._unary_operation('__pos__')
    #--- End: def

    def _set(self, attr, value):
        '''

Set an attribute which actually belongs to the partition array. This
is used by the decorated properties,
'''
        setattr(self.partitions, attr, value)  
    #--- End: def

    def _get(self, attr):
        '''

Get an attribute which actually belongs to the partition array. This
is used by the decorated properties,
'''
        return getattr(self.partitions, attr)
    #--- End: def

    def _del(self, attr):
        '''

Delete an attribute which actually belongs to the partition
array. This is used by the decorated properties,
'''
        try:
            delattr(self.partitions, attr)
        except AttributeError:
            raise AttributeError("Can't delete '%s' attribute '%s'" %
                                 (self.__class__.__name__, attr))
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: dtype (can't set or delete)
    # ----------------------------------------------------------------
    @property
    def dtype(self):
        '''

Numpy data-type of the data array.

If the array is partitioned internally into sub-arrays with different
data-types, then the normal data-type coercion rules apply. For
example, if the partitions have data-types 'int32' and 'float32' then
the data array's data-type will be 'float32'.

**Examples**

>>> type(f.dtype)
<type 'numpy.dtype'>
>>> f.dtype
dtype('float64')

'''
        return self._get('dtype')
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: Units
    # ----------------------------------------------------------------
    @property
    def Units(self):
        '''

Deleting the `Units` attribute actually sets it to undefined units, so
the Data object is guaranteed to always have the `Units` attribute.

**Examples**

>>> del d.Units
>>> print d.Units
<CF Units: >

'''
        return self._get('Units')
    #--- End: def

    @Units.setter
    def Units(self, value): self._set('Units', value)
    @Units.deleter
    def Units(self)       : self.Units = Units(None)

    # ----------------------------------------------------------------
    # Property attribute: _FillValue
    # ----------------------------------------------------------------
    @property
    def _FillValue(self):
        '''

The _FillValue CF attribute.

If set to None then the default numpy fill value appropriate to the
data array's data type will be used.

Deleting the `_FillValue` attribute actually sets it to None, so the
Data object is guaranteed to always have the `_FillValue` attribute.

**Examples**

>>> del d._FillValue
>>> print d._FillValue
None

'''
        return self._get('_FillValue')
    #--- End: def

    @_FillValue.setter
    def _FillValue(self, value): self._set('_FillValue', value)
    @_FillValue.deleter
    def _FillValue(self)       : self._FillValue = None

    # ----------------------------------------------------------------
    # Property attribute: adims
    # ----------------------------------------------------------------
    @property
    def adims(self): return self._get('adims')
    @adims.setter
    def adims(self, value): self._set('adims', value)
    @adims.deleter
    def adims(self)       : self._del('adims')

    # ----------------------------------------------------------------
    # Property attribute: direction
    # ----------------------------------------------------------------
    @property
    def direction(self): return self._get('direction')
    '''

'''
    @direction.setter
    def direction(self, value): self._set('direction', value)
    @direction.deleter
    def direction(self)       : self._del('direction')

    # ----------------------------------------------------------------
    # Property attribute: ndim
    # ----------------------------------------------------------------
    @property
    def ndim(self):
        '''

Number of dimensions in the data array.

**Examples**

>>> d.shape
[73, 96]
>>> d.ndim
2

'''
        return self._get('ndim')
    @ndim.setter
    def ndim(self, value): self._set('ndim', value)
    @ndim.deleter
    def ndim(self)       : self._del('ndim')

    # ----------------------------------------------------------------
    # Property attribute: order
    # ----------------------------------------------------------------
    @property
    def order(self): return self._get('order')
    '''

'''
    @order.setter
    def order(self, value): self._set('order', value)
    @order.deleter
    def order(self)       : self._del('order')

    # ----------------------------------------------------------------
    # Property attribute: shape
    # ----------------------------------------------------------------
    @property
    def shape(self):
        '''

List of the data array's dimension sizes.

Note that this attribute is a list, not a tuple.

**Examples**

>>> d.shape
[73, 96]

'''
        return self._get('shape')
    @shape.setter
    def shape(self, value): self._set('shape', value)
    @shape.deleter
    def shape(self)       : self._del('shape')

    # ----------------------------------------------------------------
    # Property attribute: size
    # ----------------------------------------------------------------
    @property
    def size(self):
        '''

Number of elements in the data array.

**Examples**

>>> d.shape
[73, 96]
>>> d.size
7008

'''
        return self._get('size')
    @size.setter
    def size(self, value): self._set('size', value)
    @size.deleter
    def size(self)       : self._del('size')

    # ----------------------------------------------------------------
    # Property attribute: is_scalar (can't set or delete)
    # ----------------------------------------------------------------
    @property
    def is_scalar(self):
        '''

True if the data array is a 0-d scalar array.

**Examples**

>>> d.ndim
0
>>> d.is_scalar
True

>>> d.ndim >= 1
True
>>> d.is_scalar
False

'''
        return self._get('is_scalar')
    #--- End: def

    def save_to_disk(self, itemsize=None):
        '''

:Parameters:

    itemsize : int, optional

:Returns:

    out : bool

**Examples**

>>> 
>>>

'''
        return self.partitions.save_to_disk(itemsize)
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: array (can't set or delete)
    # ----------------------------------------------------------------
    @property
    def array(self):
        '''

A numpy array copy the data array.

**Examples**

>>> a = d.array
>>> type(a)
<type 'numpy.ndarray'>

'''
        array0 = numpy.ma.empty(self.shape, dtype=self.dtype)

        order        = self.order
        direction    = self.direction
        units        = self.Units

        save           = self.save_to_disk()
        revert_to_file = True

        for partition in self.partitions.flat():
            data = partition.conform(order, direction, units, 
                                     save, revert_to_file)

            array0[partition.indices] = data

            partition.close()
        #--- End: for

        return array0
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: varray (can't set or delete)
    # ----------------------------------------------------------------
    @property
    def varray(self):
        '''

A numpy array view the data array.

Note that making changes to elements of the returned view changes the
underlying data.

**Examples**

>>> a = d.varray
>>> type(a)
<type 'numpy.ndarray'>
>>> a
array([0, 1, 2, 3, 4])
>>> a[0] = 999
>>> d.varray
array([999, 1, 2, 3, 4])

'''
        order     = self.order
        direction = self.direction
        units     = self.Units

        save           = False
        revert_to_file = False

        if not self.adims:
            # If there is only one partition, then we can conform it
            # and return a view to the partition's data without having
            # to create an empty array and then filling it up
            # partition by partition.
            self.partitions[0].conform(order, direction, units,
                                       save, revert_to_file)
            return self.partitions[0].data.view()
        #--- End: if

        # Still here?  
        shape = self.shape
        array = numpy.ma.empty(shape, dtype=self.dtype)

        for partition in self.partitions.flat():

            data = partition.conform(order, direction, units, 
                                     save, revert_to_file)

            array[partition.indices] = data

#            del partition.data
        #--- End: for

        self.partitions = PartitionArray(
            [Partition(location  = [(0,n) for n in shape],
                       shape     = shape[:],
                       order     = order[:],
                       direction = copy(direction),
                       data      = array,
                       part      = [],
                       Units     = units.copy(),      
                       )
             ],
            order      = order,
            shape      = shape,
            adims      = [],
            direction  = direction,
            size       = self.size,
            Units      = units,
            ndim       = self.ndim,
            _FillValue = self._FillValue,
            )

        return array.view()
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: binary_mask (can't set or delete)
    # ----------------------------------------------------------------
    @property
    def binary_mask(self):
        '''

The binary missing data mask of the data array.

The binary mask has 0 where the data array has missing data and 1
otherwise.

**Examples**

>>> d.mask.array
array([ True, False,  True, False], dtype=bool)
>>> b = d.binary_mask.array
array([0, 1, 0, 1], dtype=int32)

'''
        binary_mask = self.copy()

        is_masked = numpy.ma.is_masked

        order     = mask.order
        direction = mask.direction
        units     = mask.Units

        save           = self.save_to_disk(itemsize = numpy.dtype('int32').itemsize)
        revert_to_file = False

        self.to_memory()

        for partition in binary_mask.partitions.flat():
            
            data = partition.conform(order, direction, units, 
                                     save, revert_to_file)
            
            if is_masked(data):
                # data is a numpy.ma.array instance and has a mask
                # with at least one True element.
                partition.data = numpy.asarray(~data.mask, dtype='int32')
            else:
                # data is not a numpy.ma.array instance or data is a
                # numpy.ma.array instance and has a mask of
                # numpy.ma.nomask.
                partition.data = numpy.ones(data.shape, dtype=bool)

            partition.Units = Units('1')

            partition.close()
        #--- End: for

        binary_mask.Units = Units('1')

        return binary_mask
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: is_masked (can't set or delete)
    # ----------------------------------------------------------------
    @property
    def is_masked(self):
        '''

True if the data array has any masked values.

**Examples**

>>> d.is_masked
True
'''
        is_masked = numpy.ma.is_masked

        order          = self.order
        direction      = self.direction
        units          = self.Units

        save           = self.save_to_disk()
        revert_to_file = True

        nomask = numpy.ma.nomask

        for partition in self.partitions.flat():

            data = partition.conform(order, direction, units,
                                     save, revert_to_file)
            partition.close()

            if (is_masked(data) and 
                (data.mask is not nomask or data.mask.any())):
                return True
        #--- End: for

        return False
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: mask
    # ----------------------------------------------------------------
    @property
    def mask(self):
        '''

The boolean missing data mask of the data array.

The boolean mask has True where the data array has missing data and
False otherwise.

The mask may be set to the equivalent of 'no missing data' (i.e. all
elements are False) by deleting the attribute.

**Examples**

>>> d.shape
[12, 73, 96]
>>> m = d.mask
>>> m.dtype
dtype('bool')
>>> m.shape
[12, 73, 96]

>>> del d.mask
>>> d.array.mask
False
>>> import numpy
>>> a.array.mask is numpy.ma.nomask
True
'''
        mask = self.copy()

        is_masked = numpy.ma.is_masked

        order     = mask.order
        direction = mask.direction
        units     = mask.Units

        save           = self.save_to_disk(itemsize = numpy.dtype(bool).itemsize)
        revert_to_file = False

        self.to_memory()

        for partition in mask.partitions.flat():

            data = partition.conform(order, direction, units, 
                                     save, revert_to_file)
            
            if is_masked(data):
                # data is a numpy.ma.array instance and has a mask
                # with at least one True element.
                partition.data = data.mask.copy()
            else:
                # data is not a numpy.ma.array instance or data is a
                # numpy.ma.array instance and has a mask of
                # numpy.ma.nomask.
                partition.data = numpy.zeros(data.shape, dtype=bool)

            partition.Units = Units()

            partition.close()
        #--- End: for

        mask.Units = Units()

        return mask
    #--- End: def

    @mask.setter
    def mask(self, new_mask):
        raise NotImplementedError("Can't set a mask ... yet. DCH")
    #--- End: def

    @mask.deleter
    def mask(self):
        is_masked    = numpy.ma.is_masked
        nomask       = numpy.ma.nomask

        order        = self.order
        direction    = self.direction
        units        = self.Units

        save           = self.save_to_disk()
        revert_to_file = False

        for partition in self.partitions.flat():

            data = partition.conform(order, direction, units, 
                                     save, revert_to_file)

            if is_masked(data):
                data.mask = nomask
                data.shrink_mask()
            #--- End: if

            partition.close()
        #--- End: for
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: first_datum (can't delete)
    # ----------------------------------------------------------------
    @property
    def first_datum(self):
        ''' 

The first element of the data array.

May be retrieved or set.

Equivalent to ``x[(0,) * x.ndim].array.item()`` or ``x[(0,) * x.ndim]
= y``

**Examples**

>>> d.array
array([[1, 2],
       [3, 4]])
>> d.first_datum
1
>> d.first_datum = 999
>> d.array
array([[999,   2],
       [  3,   4]])

'''
        return self[(slice(0,1),) * self.ndim].array.item()
    #--- End: def

    @first_datum.setter
    def first_datum(self, value):
        self[(slice(0,1),) * self.ndim] = value
    #--- End: def
    
    # ----------------------------------------------------------------
    # Property attribute: last_datum (can't delete)
    # ----------------------------------------------------------------
    @property
    def last_datum(self):
        ''' 

The last element of the data array.

May be retrieved or set.

Equivalent to ``x[(-1,) * x.ndim].array.item()`` or ``x[(-1,) *
x.ndim] = y``

**Examples**

>>> d.array
array([[1, 2],
       [3, 4]])
>> d.last_datum
4
>> d.last_datum = 999
>> d.array
array([[  1,   2],
       [  3, 999]])

'''
        return self[(slice(-1,None),) * self.ndim].array.item()
    #--- End: def

    @last_datum.setter
    def last_datum(self, value):
        self[(slice(-1,None),) * self.ndim] = value
    #--- End: def

    def all(self):
        '''

Test whether all array elements evaluate to True.

Masked values are considered as True during computation.

**Examples**

>>> d.array
array([0, 3, 0])
>>> d.all()
False

>>> d.array
array([1, 3, 2])
>>> d.all()
True

'''         
        order, direction, units = self.order, self.direction, self.Units
        save, revert_to_file    = self.save_to_disk(), True
              
        for partition in self.partitions.flat():
            array = partition.conform(order, direction, units, 
                                      save, revert_to_file)
            partition.close()
            if not array.all():
                return
        #--- End: for

        return True
    #--- End: def

    def any(self):
        '''

Test whether any array elements evaluate to True.

Masked values are considered as True during computation.

**Examples**

>>> d.array
array([0, 0, 0])
>>> d.any()
False

>>> d.array
array([0, 3, 0])
>>> d.any()
True

'''         
        order, direction, units = self.order, self.direction, self.Units
        save, revert_to_file    = self.save_to_disk(), True
               
        for partition in self.partitions.flat():
            array = partition.conform(order, direction, units, 
                                      save, revert_to_file)
            partition.close()
            if array.any():
                return True
        #--- End: for

        return
    #--- End: def

    def override_units(self, new_units):
        '''

Override the data array units in place.

Not to be confused with setting the `Units` attribute to units which
are equivalent to the original units.

This is different to setting the `Units` attribute, as the new units
need not be equivalent to the original ones and the data array
elements will not be changed to reflect the new units.

:Parameters:

    new_units : str or Units
        The new units for the data array.

:Returns:

    None

**Examples**

>>> d.Units
<CF Units: hPa>
>>> d.first_datum
100000.0
>>> d.override_units('km')
>>> d.Units
<CF Units: km>
>>> d.first_datum
100000.0
>>> d.override_units(cf.Units('watts'))
>>> d.Units
<CF Units: watts>
>>> d.first_datum
100000.0

'''
        if isinstance(new_units, basestring):
            new_units = Units(new_units)

        order        = self.order
        direction    = self.direction
        units        = self.Units

        save           = self.save_to_disk()
        revert_to_file = False
               
        for partition in self.partitions.flat():
            
            if partition.Units == units:
                partition.Units = new_units.copy()
                continue
            #--- End: if

            partition.conform(order, direction, units, 
                              save, revert_to_file)

            partition.Units = new_units.copy()

            partition.close()
        #--- End: for

        self.Units = new_units.copy()
    #--- End: def

    def to_disk(self):
        '''

'''
        self.partitions.to_disk()
    #--- End: if

    def to_memory(self, regardless=False):
        '''

Store the data array in memory if it is smaller than the chunk size.

:Parameters:
    
    regardless : bool, optional
        If True then store the data array in memory regardless of its
        size.

:Returns:

    None

**Examples**

>>> d.to_memory()
>>> d.to_memory(True)

'''
        self.partitions.to_memory(regardless)
    #--- End: def

    def copy(self):
        '''

Return a deep copy.

Equivalent to ``copy.deepcopy(d)``.

:Returns:

    out : 
        The deep copy.

**Examples**

>>> e = d.copy()

'''
        new            = type(self)()        
        new.partitions = self.partitions.copy()
        return new
    #--- End: def

    def dump(self, id=None, omit=()):
        '''

'''
        if id is None:
            id = self.__class__.__name__
            
        string = ['%s.shape = %s' % (id, tuple(self.shape))]

#        if self.ndim <= 1:
#            if self.size == 1:
#                string.append('%s.first_datum = %s' % (id, self.first_datum))
#            else:
#                string.append('%s[[0,-1]] = %s' % \
#                                  (id, numpy.ma.array([self.first_datum,  
#                                                       self.last_datum])))
#        else:
##            fs = ','.join([' 0'] * self.ndim)
##            ls = ','.join(['-1'] * self.ndim)
#            string.append('%s.first_datum = %s' % (id, self.first_datum))
#            string.append('%s.last_datum  = %s' % (id, self.last_datum))
#        #-- End: if
                
        if self.size == 1:
            string.append('%s.first_datum = %s' % (id, self.first_datum))
        else:
            #            fs = ','.join([' 0'] * self.ndim)
            #            ls = ','.join(['-1'] * self.ndim)
            string.append('%s.first_datum = %s' % (id, self.first_datum))
            string.append('%s.last_datum  = %s' % (id, self.last_datum))
        #-- End: if
            
        for attr in ('_FillValue', 'Units'):
            if hasattr(self, attr) and attr not in omit:
                string.append('%s.%s = %s' % (id, attr,
                                              repr(getattr(self, attr))))
        #--- End: for

        return '\n'.join(string)
    #--- End: def
    
    def _equivalent(self, other, dim_name_map,
                    rtol=None, atol=None,
                    _check_units=True, use_directions=False):
        '''
'''
        self_size  = self.size
        self_ndim  = self.ndim
        other_ndim = other.ndim

        if self_size != other.size:
            return
        
        if _check_units and not self.Units.equivalent(other.Units):
            return 

        if self_size == 1:
            if self_ndim > 1 or other_ndim > 1:
                return

            other = other.copy()
               
            if not self_ndim and other_ndim:
                other.squeeze()
            elif not other_ndim and self_ndim:
                other.expand_dims()

        else:
            if self_ndim != other_ndim:
                # Data with unequal numbers of dimensions are not
                # equivalent
                return
            
            self_order  = self.order
            other_order = other.order

            try:
                axes = [self_order.index(dim_name_map[dim1])
                        for dim1 in other_order]
            except ValueError:
                return

            other = other.copy()

            other.transpose(axes)

            if use_directions:            
                self_direction  = self.direction
                other_direction = other.direction

                axes = [i
                        for i, (dim1, dim0) in enumerate(zip(other_order, 
                                                             self_order))
                        if other_direction[dim1] != self_direction[dim0]]
            
                if axes:
                    other.reverse_dims(axes)
            #--- End: if
        #--- End: if
        
        other.Units = self.Units.copy()

        return self.equals(other, rtol=rtol, atol=atol, traceback=False)
    #--- End: def

    def equals(self, other, rtol=None, atol=None, traceback=False):
        '''

True if two data arrays are logically equal, False otherwise.

:Parameters:

    other : 
        The object to compare for equality.

    atol : float, optional
        The absolute tolerance for all numerical comparisons, By
        default the value returned by the `ATOL` function is used.

    rtol : float, optional
        The relative tolerance for all numerical comparisons, By
        default the value returned by the `RTOL` function is used.

    traceback : bool, optional
        If True then print a traceback highlighting where the two
        instances differ.

:Returns: 

    out : bool
        Whether or not the two instances are equal.

**Examples**

>>> d.equals(d)
True
>>> d.equals(d + 1)
False

'''
        # Check each instance's id
        if self is other:
            return True
 
        # Check that each instance is the same type
        if self.__class__ != other.__class__:
            return

        # Check that each instance has the same shape
        if self.shape != other.shape:
            if traceback:
                print \
"%s: Different shape: %s, %s" % (self.__class__.__name__,
                                 self.shape, other.shape)
            return
        #--- End: if

        # Check that each instance has the same units
        if self.Units != other.Units:
            if traceback:
                print \
"%s: Different Units: %s, %s" % (self.__class__.__name__,
                                 repr(self.Units), repr(other.Units))
            return
        #--- End: if
        # Check that each instance has the same fill value
        if self._FillValue != other._FillValue:
            if traceback:
                print \
"%s: Different _FillValue: %s, %s" % (self.__class__.__name__,
                                      self._FillValue, other._FillValue)
            return
        #--- End: if

        # ------------------------------------------------------------
        # Check that each instance has the array values
        # ------------------------------------------------------------
        # Set default tolerances
        if rtol is None:
            rtol = RTOL()
        if atol is None:
            atol = ATOL()        

        order        = self.order
        direction    = self.direction
        units        = self.Units

        save           = self.save_to_disk()
        revert_to_file = True

        other.to_memory()

        for partition in self.partitions.flat():

            array0 = partition.conform(order, direction, units, 
                                       save, revert_to_file)
            array1 = other[partition.indices].varray

            partition.close()

            if not array_allclose(array0, array1,
                                  rtol=rtol, atol=atol):
                if traceback:
                    print \
"%s: Different data arrays" % self.__class__.__name__
                return
        #--- End: for

        # ------------------------------------------------------------
        # Still here? Then the two instances are equal
        # ------------------------------------------------------------
        return True            
    #--- End: def

    def ufunc(self, func, *args, **kwargs):
        '''

Return a 
'''
        new = self.copy()

        order, direction, units = new.order, new.direction, new.Units
        save           = new.save_to_disk()
        revert_to_file = True
             
        self.to_memory()
  
        for partition in new.partitions.flat():
            array = partition.conform(order, direction, units, 
                                      save, revert_to_file)
            partition.data = func(array, *args, **kwargs)
            partition.close()
        #--- End: for

        return new
    #--- End: def

    def set_location_map(self):
        '''
'''
        self.partitions.set_location_map()
    #--- End: def

    def change_dimension_names(self, dim_name_map):
        '''

Change the dimension names.

The dimension names are arbitrary (though unique), so mapping them to
another arbitrary (though unique) set does not change the data array
values, units, dimension directions nor dimension order.

**Examples**

>>> d.order
['dim0', 'dim1', 'dim2']
>>> dim_name_map
{'dim0': 'dim1',
 'dim1': 'dim0',
 'dim2': 'dim2',
 'dim3': 'dim3'}
>>> d.change_dimension_names(dim_name_map)
>>> d.order
['dim1', 'dim0', 'dim2']

'''
        self.partitions.change_dimension_names(dim_name_map)
    #--- End: def

    def hash(self):
        '''

Return a hash value for the data array.

Note that generating the hash realizes the entire array in memory,
which may not be possible for large arrays.

:Returns:

    out : int
        The hash value.

**Examples**

>>> d.hash()
3632586161869339209

'''
        return hash(tuple(self.array.flat))
    #--- End: def

    def squeeze(self, axes=None):
        '''

Remove size 1 dimensions from the shape of the data in place.

:Parameters:

    axes : int or tuple of ints, optional

        The axes to be squeezed given by their positions. If unset
        then all size one dimensions of the data array are removed.

:Returns:

    out : tuple of ints
        The axes which were squeezed as a tuple of their positions.

**Examples**

>>> v.shape
[1]
>>> v.squeeze()
>>> v.shape
[]

>>> v.shape
[1, 2, 1, 3, 1, 4, 1, 5, 1, 6, 1]
>>> v.squeeze(axis=2).shape
[1, 2, 3, 1, 4, 1, 5, 1, 6, 1]
>>> v.squeeze(axis=(0,)).shape
[2, 3, 1, 4, 1, 5, 1, 6, 1]
>>> v.squeeze(axis=(2, 4)).shape
[2, 3, 4, 5, 1, 6, 1]
>>> v.squeeze().shape
[2, 3, 4, 5, 6]

'''
        if self.is_scalar:
            if axes or axes == 0:
                raise ValueError("Can't remove a dimension from a scalar %s" %
                                 self.__class__.__name__)
            return ()

        shape = self.shape

        if axes is None:
            axes = [i for i in range(self.ndim-1,-1,-1) if shape[i] == 1]
        else:
            if isinstance(axes, (int, long)):
                axes = (axes,)

            for index in axes:
                try:
                    if shape[index] > 1:
                        raise ValueError("Can't remove a dimension of size > 1 from %s" %
                                         self.__class__.__name__)
                except IndexError:
                    raise ValueError("Axes don't match data: %s" % (axes,))
        #--- End: if

        if not axes:
            return axes

        # Still here?
        order     = self.order
        direction = self.direction

        for index in axes:
            shape.pop(index)
            dir = direction.pop(order.pop(index))
            for partition in self.partitions.flat():
                partition.location.pop(index)
                partition.shape.pop(index)
        #--- End: for

        self.ndim = len(shape)

        # If the data has been squeezed to a scalar
        if self.is_scalar:
            self.direction = dir

        return axes
    #--- End: def

    def transpose(self, axes=None):
        '''

axes: list of ints, optional
        By default, reverse the dimensions, otherwise permute the axes
        according to the values given.

'''
        ndim  = self.ndim
        order = self.order
        
        if axes is None:
            # By default, reverse the dimensions
            if ndim <= 1:
                return
            axes = range(ndim-1, -1, -1)
        else:
            if isinstance(axes, (str, int, long)):
                # Convert an integer to a single element sequence
                axes = [axes]
            else:
                axes = list(axes)
                
            for i, axis in enumerate(axes):
                if axis in order:
                    axes[i] = order.index(axis)
                elif not isinstance(axis, (int, long)):
                    raise TypeError(
"Transpose axis must be an integer or a dimension name: %s" % repr(axis))
            #--- End: for

            # Return unchanged if axes are in the same order as the data
            if axes == range(ndim):
                return

            if len(axes) != ndim:
                raise ValueError("Axes don't match array: %s" % (axes,))
        #--- End: if

        # Permute the order
        try:
            self.order = [order[i] for i in axes]
        except IndexError:
            raise ValueError("Axes don't match array: %s" % (axes,))

        # Permute the shape
        shape      = self.shape
        self.shape = [shape[i] for i in axes]
        
        # Permute the locations map
        for partition in self.partitions.flat():
            location = partition.location
            shape    = partition.shape
            partition.location = [location[i] for i in axes]
            partition.shape    = [shape[i]    for i in axes]
    #--- End: def

    def partition_boundaries(self):
        '''
'''            
        return self.partitions.partition_boundaries()
    #--- End: def

    def reverse_dims(self, axes=None):
        '''

Reverse the directions of data array dimensions in place.

Equivalent to indexing the specified dimensions with ``::-1``.

:Parameters:

    axes : int or sequence of ints
        Reverse the dimensions whose positions are given. By default
        all dimensions are reversed.

:Returns:

    out : list of ints
        The axes which were reversed, in arbitrary order.

**Examples**

>>> d.ndim
3
>>> d.reverse_dims()
>>> d.reverse_dims(1)
>>> e = d[::-1, ::-1, :]
>>> d.reverse_dims([0, 1]).equals(e)
True

'''
        def _reverse_across_partitions(partitions, adims, adim):
            '''
            in place
            '''
            if not adims:
                return    
            if adim == adims[-1]:
                partitions.reverse()
            else:
                adims = adims[0:-1]
                if not adims:
                    return    
                for nested_partitions in partitions:
                    # Recursive call
                    _reverse_across_partitions(nested_partitions, adims, adim)
        #--- End: def

        if axes is None or (not axes and axes != 0):
            return

        # Scalar data. Allow axis in (0, None) to reverse the
        # direction
        if self.is_scalar:
            if axes:
                raise ValueError("Invalid axis for this array: %s" % (axes,))
            self.direction = not self.direction
            return 0  # DCH
        #--- End: if

        order = self.order        

        if axes is None:
#            axes = range(self.ndim)
            axes2 = dict((i, dim) for i, dim in enumerate(order))
        else:
            # Make sure that axes is a single element sequence
            if isinstance(axes, (int, long, str)):
                axes = (axes,)

            axes2 = {}
            for i, axis in enumerate(axes):
                if axis in order:
                    axes2[order.index(axis)] = axis
                elif isinstance(axis, (int, long)):
                    axes2[i] = order[axis]
                else:
                    raise TypeError(
"Reverse axis must be an integer or a dimension name: %s" % repr(axis))
            #--- End: for

            # Check for duplicate axes
            if len(axes2) != len(axes):
                raise ValueError("Repeated axis in reverse")

            # Check for out of range axes
            max_axis = max(axes2)
            if max_axis >= self.ndim:
                raise ValueError("Invalid axis for this array: %d" % max_axis)
        #--- End: if

        # Reverse each of the requested dimensions    

        partitions = self.partitions
        direction  = self.direction    # This will always be a dict when len(axes)>0
        shape      = self.shape        
        adims      = self.adims
        reversed_partition_array = False

        for i, dim in axes2.iteritems():
            if shape[i] == 1:
                continue

            direction[dim] = not direction[dim]
            if dim in adims:
                reversed_partition_array = True
                _reverse_across_partitions(partitions, adims, dim)
        #--- End: for

        # Recreate the location map
        if reversed_partition_array:
            self.set_location_map()

        return axes2.keys()
    #--- End: def        

    def expand_dims(self, axis=0, dim='None', direction=True):
        '''
        no check is done for dim already being in self.order

'''
        # Expand direction
        if self.is_scalar:
            self.direction = {dim: direction}
        else:
            self.direction[dim] = direction

        # Expand order, shape and ndim
        self.order.insert(axis, dim)        
        self.shape.insert(axis, 1)   
        self.ndim += 1
        
        # Expand the locations map
        range01 = (0, 1)
        for partition in self.partitions.flat():
            partition.location.insert(axis, range01)
            partition.shape.insert(axis, 1)
    #--- End: def

    def expand_aggregating_dims(self, adim):
        '''

'''
        self.partitions.expand_dims(adim)
    #--- End: def
        
    def add_partitions(self,
                       extra_boundaries, 
                       adim,
                       existing_boundaries=None):
        '''

**Examples**

>>> d.add_partitions(    )

'''            
        self.partitions.add_partitions(extra_boundaries, 
                                       adim,
                                       existing_boundaries=existing_boundaries)
    #--- End: def

    def new_dimension_name(self):
        '''

Return a dimension name not being used by the data array.

Note that a partition of the data array may have dimensions which
don't belong to the data array itself.

:Returns:

    out : str
        The new dimension name.

**Examples**

>>> d.order
['dim1', 'dim0']
>>> d.partitions.info('order')
[['dim0', 'dim0'],
 ['dim1', 'dim0', 'dim2']]
>>> d.new_dimension_name()
'dim3'

'''
        all_dim_names = set()
        for partition in self.partitions.flat():
            all_dim_names.update(set(partition.order))

        N   = len(all_dim_names)
        dim = 'dim%d' % N
        while dim in all_dim_names:
            N += 1
            dim = 'dim%d' % N
        #--- End: while

        return dim
    #--- End: def

    def _insert_data(self, other, indices,
                     ADim, ADim_direction, dim_name_map):
        '''
'''
        data0 = self
        data1 = other
            
        # ------------------------------------------------------------
        # 1. Change the dimension names in data1 to match those in the
        # the space of data0.
        # ------------------------------------------------------------
        data1.change_dimension_names(dim_name_map)


        # ------------------------------------------------------------
        # 2. Insert (size 1) dimensions into data0 which are present
        # in data1 but not in data0
        #
        # Note that the dimensions to be inserted into data0 will have
        # size 1 in data1 (because the two spaces are aggregatable).
        #
        # Note that after this process data1.order will be a subset of
        # data0.order (because the two spaces are aggregatable and so
        # have the same rank).
        # ------------------------------------------------------------
        order0     = data0.order
        direction1 = data1.direction
        for dim1 in data1.order[::-1]:
            if dim1 not in order0:
                data0.expand_dims(0, dim1, direction1[dim1])
        #--- End: for
    
        # ------------------------------------------------------------
        # 3. If the aggregating dimension is missing from data0 then
        # insert it (with size 1 in the slowest varying position).
        # 
        # Note that after this process data1.order will be a subset of
        # data0.order
        # ------------------------------------------------------------
        if ADim not in data0.order:
            data0.expand_dims(0, ADim, ADim_direction)

        # ------------------------------------------------------------
        # 4. Permute the dimensions of data1 to match data0.
        #
        # For example: if data0.order=['dim0','dim1','dim2'] and
        #              data1.order=['dim2','dim0'] to start with then
        #              data1.order will become ['dim0','dim2']
        #
        # Note that by now, data1.order is a subset of data0.order
        # (this was was ensured by steps 2 and 3).
        # ------------------------------------------------------------
        order1 = data1.order
        axes   = [order1.index(dim)
                  for dim in data0.order if dim in order1] 
        data1.transpose(axes)

        # ------------------------------------------------------------
        # 5. Insert (size 1) dimensions into data1 which are present in
        # data0 but not in data1.
        #
        # For example: if data0.order=['dim0','dim1','dim2'] and
        #              data1.order=['dim0','dim2'] to start with then
        #              data1.order will become ['dim0','dim1','dim2']
        #
        # Note that after this process data0 and data1 will span the
        # same dimensions.
        #
        # Note that the orders of dimensions in data0 and data1 have
        # to be the same (so that their partitions' locations
        # conform).
        #
        # Note that after this process the Aggregating Dimension will
        # definitely be in both data0 and data1.
        # ------------------------------------------------------------
        direction0 = data0.direction
        order1     = data1.order
        for i, dim in enumerate(data0.order):
            if dim not in order1:
                data1.expand_dims(i, dim, direction0)
        #--- End: for

        # ------------------------------------------------------------
        # 6. Make sure that the aggregating dimensions of data0 are
        # the same as data1 (although, for now, they may have
        # different orders).
        #
        # Note that this may involve adding new aggregating dimensions
        # to either or both of data0 and data1.
        #            
        # Note that if the Aggregating Dimension needs to be added it
        # is inserted as the outer (slowest varying) dimension to
        # reduce the likelihood of having to transpose (expensively)
        # the partition array.
        # ------------------------------------------------------------
        for f, g in zip((data0, data1), 
                        (data1, data0)):

            f_adims = f.adims
            g_adims = g.adims[:]
            try:
                g_adims.remove(ADim)
            except ValueError:
                pass
    
            for adim in g_adims[::-1]:
                if adim not in f_adims:
                    f.expand_aggregating_dims(adim)

            if ADim not in f.adims:
                f.expand_aggregating_dims(ADim)
        #--- End: for

        # ------------------------------------------------------------
        # 7. Permute the aggregating dimensions of the partition array
        # of data0 so that the Aggregating Dimension is the outermost
        # (slowest varying) dimension.
        # ------------------------------------------------------------        
        axis = data0.adims.index(ADim)
        data0.partitions.rollaxis(axis, 0)
            
        # ------------------------------------------------------------
        # 8. Permute the aggregating dimensions of the partition array
        # of data1 to match data0.
        # ------------------------------------------------------------
        adims1 = data1.adims
        axes   = [adims1.index(dim) for dim in data0.adims]
        data1.partitions.transpose(axes)
            
        # ------------------------------------------------------------
        # 9. Create new partition boundaries in the partition arrays
        # of data0 and data1 so that their partition arrays may be
        # considered as different slices of a common, larger
        # hyperrectangular partition array.
        #
        # Note that there is no need to add any boundaries across the
        # Aggregating Dimension.
        # ------------------------------------------------------------
        boundaries0 = data0.partition_boundaries()        
        boundaries1 = data1.partition_boundaries()

        for dim in data0.adims[1:]:
            
            # Still here? Then see if there are any partition
            # boundaries to be created for this aggregating dimension
            bounds0 = boundaries0[dim]
            bounds1 = boundaries1[dim]

            symmetric_diff = set(bounds0).symmetric_difference(bounds1)            
            if not symmetric_diff:
                # The partition boundaries for this aggregating
                # dimension are already the same in data0 and data1
                continue
            
            # Still here? Then there are some partition boundaries to
            # be created for this aggregating dimension in data0
            # and/or data1.
            for f, g, bf, bg in zip((data0, data1, bounds0, bounds1), 
                                    (data1, data0, bounds1, bounds0)):                
                extra_bounds = [i
                                for i in bg
                                if i in symmetric_diff]
                
                f.add_partitions(extra_bounds, dim, boundaries=bf)
            #--- End: for
        #--- End: for
                
        # ------------------------------------------------------------
        # 10. Merge data1 into data0 along the Aggregating Dimension.
        #
        # If no indices have been provided then, along the Aggregating
        # Dimension, concatenate the partitions of data1 onto the end of the
        # partitions of data0.
        # ------------------------------------------------------------
        if indices is None:
            npartitions0 = data0.npartitions
            indices = xrange(npartitions0, 
                             npartitions0 + data1.npartitions)
        #--- End: if

        partitions0 = data0.partitions
        partitions1 = data1.partitions
        for i, index in enumerate(indices):
            partitions0.insert(index, partitions1[i])
                               
        # Update the size and shape of data0
        index               = data1.order.index(ADim)
        data0.shape[index] += data1.shape[index]
        data0.size         += long(data1.size)

        # Update the location map of the partition array of data0
        data0.set_location_map()

        # ------------------------------------------------------------
        # If the new array is larger than the chunk size then save any
        # numpy arrays to disk
        # ------------------------------------------------------------
        if data0.save_to_disk():
            data0.to_disk()
        
        # ------------------------------------------------------------
        # Done
        # ------------------------------------------------------------
    #--- End: def

#--- End: class

    
def _size_of_index(index, size=None):
    '''

Return the number of elements resulting in applying an `index` to a
sequence of length `size`

'''
    if isinstance(index, slice):
        start, stop, step = index.indices(size)
        div, mod = divmod(stop-start, step)
        if mod != 0:
            div += 1
        return div
    else:
        return len(index)
#--- End: def

def _parse_indices(data, indices):
    '''

'''
    parsed_indices = parse_indices(data, indices)

    reverse_dimensions = []

    for i, index in enumerate(parsed_indices):

        

        if isinstance(index, slice):
            size = data.shape[i]
            if index.step < 0:              
                # If the slice step is negative, then transform the
                # original slice to a new slice with a positive step
                # such that the result of the new slice is the reverse
                # of the result of the original slice.
                #
                # For example, if the original slice is slice(6,0,-2)
                # then the new slice will be slice(2,7,2):
                #
                # >>> a = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
                # >>> a[slice(6, 0, -2)]
                # [6, 4, 2]
                # >>> a[slice(2, 7, 2)]
                # [2, 4, 6]
                # a[slice(6, 0, -2)] == list(reversed(a[slice(2, 7, 2)]))
                # True
                start, stop, step = index.indices(size)
                step    *= -1
                div, mod = divmod(start-stop-1, step)
                div_step = div*step
                start   -= div_step
                stop     = start + div_step + 1
                
                index = slice(start, stop, step)
                reverse_dimensions.append(i)         
            #--- End: if      

            # If step is greater than one, make sure that index.stop isn't
            # bigger than it needs to be.
            if index.step > 1:
                start, stop, step = index.indices(size)
                div, mod = divmod(stop-start-1, step)
                stop     = start + div*step + 1
                index    = slice(start, stop, step)
            #--- End: if

            parsed_indices[i] = index    

        else:
            # --------------------------------------------------------
            # Check that an integer list is strictly monotonic and if
            # it's descending reverse it so that it's ascending
            # --------------------------------------------------------
            step = index[1] - index[0]

            if step > 0:                        
                if not all(x < y for x, y in zip(index, index[1:])):
                    raise ValueError("Bad slice (not strictly monotonic): %s" % index)

            elif step < 0:
                if not all(x > y for x, y in zip(index, index[1:])):
                    raise ValueError("Bad slice (not strictly monotonic): %s" % index)
                
                # Reverse the list so that it's strictly monotonically
                # increasing and make a note that this dimension will
                # need reversing later
                index.reverse()
                reverse_dimensions.append(i)
            else:
                # Step is 0
                raise ValueError("Bad slice (not strictly monotonic): %s" % index)
        #--- End: if
    #--- End: for
    
    return parsed_indices, reverse_dimensions
#--- End: def

def _partition_overlaps(partition, indices):
    '''

:Parameters:

   partition : Partition

   indices : tuple
       A tuple of indices describing a subset of the master array.

:Returns:

    partition_indices, shape : {list, list} or {None, None}
    
        If the partition overlaps the indices then return a list of
        indices which will subset the partition's data to where it
        overlaps the master indices and the subsetted partition's
        shape as a list. Otherwise return`(None, None)`.

**Examples**

>>> indices = (slice(None), slice(5, 1, -2), [1,3,4,8])
>>> _partition_overlaps(p, indices)
[slice(), ddfsfsd], [3, 5, 4]

'''
    partition_indices = []
    shape    = []

    for index, (r0, r1) in zip(indices,
                               partition.location):
        
        if isinstance(index, slice):
            
            size = r1 - r0
            stop = size
            if index.stop < r1:
                stop -= (r1 - index.stop)
              
            start = index.start - r0
            if start < 0:
                start %= index.step   # start is now +ve
            
            if start >= stop:
                # This partition is not in the slice
                return None, None
                
            # Still here?
            index      = slice(start, stop, index.step)
            index_size = _size_of_index(index, size)

        else:
            if not set(index).intersection(range(r0, r1)):
                # This partition is not in the slice
                return None, None
            
            # Still here?
            index = [x - r0
                     for x in index
                     if r0 <= x < r1]
            index_size = len(index)
            if index == range(index_size):
                index = slice(0, index_size, 1)
            else:
                if index == range(index_size-1,-1,-1):
                    index = slice(index_size-1, None, -1)
        #--- End: if

        # Still here?
        partition_indices.append(index)
        shape.append(index_size)
    #--- End: for
        
    # Still here? Then the elements of this partition specified by
    # partition_indices are in the slice
    return partition_indices, shape
#--- End: def

def _overlapping_partitions(partitions, indices, dim2position):
    '''

Return the nested list of (modified) partitions which overlap the
indices.

:Parameters:

    partitions : PartitionArray

    indices : tuple

    dim2position : dict

:Returns:

    out : list
        A (possibly nested) list of Partition objects.

**Examples**

>>> type f.Data
<class 'cf.data.Data'>
>>> d.order
['dim1', 'dim2', 'dim0']
>>> dim2position = {'dim0': 2, 'dim1': 0, 'dim2' : 1}
>>> indices = (slice(None), slice(5, 1, -2), [1,3,4,8])
>>> x = _overlapping_partitions(d.partitions, indices, dim2position)
>>> type(x)
list

'''
    new_list = []
    if isinstance(partitions[0], Partition):

        for partition in partitions:

            # Find out if this partition overlaps the original slice
            _indices, shape = _partition_overlaps(partition, indices)

            if not _indices:
                # This partition is not in the slice
                continue

            # Still here? Create a new partition                      
            partition      = partition.copy()

            partition.part = partition.new_part(_indices, 
                                                dim2position,
                                                partitions.direction)
            partition.shape = shape

            new_list.append(partition)
        #--- End: for

    else:
        for nested_partitions in partitions:
            # Recursive call
            om = _overlapping_partitions(nested_partitions,
                                         indices,
                                         dim2position)
            if om:
                new_list.append(om)
        #--- End: if    
    #--- End: for

    return new_list
#--- End: def
