from copy import deepcopy
from .utils       import CHUNKSIZE, parse_indices
from .variable    import Variable, SubsetVariable
from .space       import Space
from .data        import Data
from .flags       import Flags
from .cellmethods import CellMethods


# ====================================================================
#
# Field object
#
# ====================================================================

class Field(Variable):
    '''

A field construct according to the CF data model.

A field is a container for a data array and metadata comprising
properties to describe the physical nature of the data and a
coordinate system (called a space) which describes the positions of
each element of the data array.

It is structured in exactly the same way as a field construct defined
by the CF data model.

The field's space may contain coordinates and cell measures (which
themselves contain data arrays and properties to describe them) and
transforms to describe how other auxiliary coordinates may be
computed.

As in the CF data model, all components of a field are optional.

'''

    _special_attributes = Variable._special_attributes.union(set(('ancillary_variables',
                                                                  'cell_methods',
                                                                  'flag_values',
                                                                  'flag_masks',
                                                                  'flag_meanings',
                                                                  ))
                                                             )

    def _binary_operation(self, other, method):
        '''

Implement binary arithmetic and comparison operations on the data
array with metadata-aware broadcasting.

It is intended to be called by the binary arithmetic and comparison
methods, such as __sub__(), __imul__() and __lt__().

:Parameters:

    operation : str
        The binary arithmetic or comparison method name (such as
        '__rdiv__' or '__ge__').

:Returns:

    out : Field
        The new Field, or the same Field if the operation was
        in-place.

**Examples**

>>> h = f._binary_operation(g, '__add__')
>>> h = f._binary_operation(g, '__ge__')
>>> f._binary_operation(g, '__isub__')

>>> g = f._binary_operation(2, '__add__')
>>> g = f._binary_operation(2, '__ge__')
>>> f._binary_operation(2, '__isub__')

'''
        if (isinstance(other, (float, int, long))       or
            isinstance(other, Data) and other.size == 1 or
            other is self):
            # ========================================================
            # Combine the field with one of the following:
            #
            #   * A python number
            #   * A size 1 Data object
            #   * Itself
            #
            # These cases are special because they don't involve any
            # changes to the field's space and so can use the
            # metadata-UNaware Variable._binary_operation method.
            # ========================================================
            return super(Field, self)._binary_operation(other, method)
        #--- End: if

        if not isinstance(other, self.__class__):
            raise TypeError(
                "Can't combine '%s' with '%s'" %
                (self.__class__.__name__, other.__class__.__name__))

        # ============================================================
        # Combine the field with another field
        # ============================================================

        # ------------------------------------------------------------
        # Analyse each field's space.
        #
        # This involves
        #
        # * Finding a standard
        # ------------------------------------------------------------
 
        # For each field, map standard names for its dimensions to
        # their space identifiers. For example:
        #
        # >>> stdname_to_dim0
        # {'time': 'dim0'}
        stdname_to_dim0 = {}
        stdname_to_dim1 = {}

        # For each field, map its dimensions' standard names to their
        # space identifiers when the dimensions' standard names are
        # defined by a 1-d auxiliary coordinate.
        #
        # Note that these dictionaries are subsets of their
        # stdname_to_dim counterparts. For example:
        #
        # >>> stdname_to_aux0
        # {'region': 'dim3'}
        stdname_to_aux0 = {}
        stdname_to_aux1 = {}

        # For each field, map its dimensions' standard names to the
        # coordinates which provide those names. For example:
        #
        # >>> stdname_to_variable0
        # {'time': <CF Coordinate: time(12)>}
        stdname_to_variable0 = {}
        stdname_to_variable1 = {}


        aux_coords0         , aux_coords1          = {}, {}

        # For each field list the dimensions which are undefined (in
        # that no standard name can be assigned to them). For example:
        #
        # >>> undefined_dim0
        # ['dim1']
        undefined_dim0 = []
        undefined_dim1 = []

        stdname_to_dim      = stdname_to_dim0
        stdname_to_aux      = stdname_to_aux0
        stdname_to_variable = stdname_to_variable0
        undefined_dim       = undefined_dim0
        aux_coords          = aux_coords0

        for space in (self.space, other.space):

            for dim in space.dimension_sizes:

                # Find this dimension's 1-d and N-d auxiliary
                # coordinates
                aux_coords['N-d']      = {}
                aux_coords[dim]        = {}
                aux_coords[dim]['1-d'] = set()
                aux_coords[dim]['N-d'] = set()
                for aux, aux_coord in space.aux_coords(dim):
                    if aux_coord.ndim > 1:
                        aux_coords['N-d'][aux] = None
                        aux_coords[dim]['N-d'].add(aux)
                    else:
                        aux_coords[dim]['1-d'].add(aux)
                #--- End: for

                if dim in space:
                    # This dimension of the space has a dimension
                    # coordinate
                    dim_coord     = space[dim]
                    standard_name = dim_coord.getattr('standard_name', None)
                    data          = getattr(dim_coord, 'Data', None)
                    if standard_name is not None and data is not None:
                        if standard_name in stdname_to_dim:
                            raise ValueError(
"Can't combine fields:  Field has more than one '%s' dimension" %
locals())

                        stdname_to_dim[standard_name]      = dim
                        stdname_to_variable[standard_name] = dim_coord
                        continue

                elif len(aux_coords[dim]['1-d']) == 1:
                    # This dimension of the space does not have a
                    # dimension coordinate but it does have exactly
                    # one 1-d auxiliary coordinate, so that will do.
                    aux       = list(aux_coords[dim]['1-d'])[0]
                    aux_coord = space[aux]

                    standard_name = aux_coord.getattr('standard_name', None)
                    data          = getattr(aux_coord, 'Data', None)
                    if standard_name and data:
                        if standard_name in stdname_to_dim:
                            raise ValueError(
"Can't combine fields: Field has more than one '%s' dimension" %
standard_name)

                        stdname_to_aux[standard_name]      = aux
                        stdname_to_dim[standard_name]      = dim
                        stdname_to_variable[standard_name] = aux_coord
                        continue
                #--- End: if

                # Still here? Then this dimension is undefined
                undefined_dim.append(dim)
            #--- End: for

            stdname_to_dim      = stdname_to_dim1
            stdname_to_aux      = stdname_to_aux1
            stdname_to_variable = stdname_to_variable1
            undefined_dim       = undefined_dim1
            aux_coords          = aux_coords1
        #--- End: for

        # Check that at most one field has undefined dimensions
        if undefined_dim0 and undefined_dim1:
            raise ValueError(
"Can't combine fields: Both fields have undefined dimensions")

        # Find the dimension names which are present in both fields -
        # called 'matching standard names'
        matching_standard_names = \
            set(stdname_to_dim0).intersection(stdname_to_dim1)

        # Check that any matching dimensions defined by an auxiliary
        # coordinate are done so in both fields.
        for standard_name in set(stdname_to_aux0).symmetric_difference(stdname_to_aux1):
            if standard_name in matching_standard_names:
                raise ValueError(
"Can't combine fields: '%s' dimension defined by auxiliary in only 1 field" %
standard_name)
        #--- End: for

        # ------------------------------------------------------------
        # For matching dimension coordinates check that one of the
        # following is true:
        #
        # 1) They have equal size > 1 and their data arrays are
        # equivalent
        #
        # 2) They have unequal sizes and one of them has size 1
        #
        # 3) They have equal size = 1. In this case, if the data
        # arrays are not equivalent then the dimension will be omitted
        # from the result field's space.
        # ------------------------------------------------------------

        # List of size 1 dimensions to be completely removed from the
        # result field. Such a dimension's size 1 defining coordinates
        # have unequivalent data arrays.
        #
        # For example:
        # >>> remove_size1_dims
        # ['dim2']
        remove_size1_dims = []

        # List of matching dimensions with equivalent defining
        # dimension coordinate data arrays. 
        #
        # Note that we don't need to include matching dimensions with
        # equivalent defining *auxiliary* coordinate data arrays.
        #
        # For example:
        # >>> equivalent_data_matching_dimensions
        # [('dim2', 'dim0')]
        equivalent_data_matching_dimensions = []

        # For each field, list those of its matching dimensions which
        # need to be broadcast against the other field. I.e. those
        # dimensions which are size 1 but size > 1 in the other field.
        #
        # For example:
        # >>> broadcast_dims0
        # ['dim1']
        broadcast_dims0 = []
        broadcast_dims1 = []

        for standard_name in matching_standard_names:

            coord0 = stdname_to_variable0[standard_name]
            coord1 = stdname_to_variable1[standard_name]

            size0 = coord0.size
            size1 = coord1.size
            if size0 != size1:
                # Defining coordinates have different sizes
                if size0 == 1:
                    broadcast_dims0.append(stdname_to_dim0[standard_name])
                elif size1 == 1:
                    broadcast_dims1.append(stdname_to_dim1[standard_name])
                else:
                    raise ValueError(
"Can't combine fields: Incompatible '%s' dimension sizes: %d, %d" %
(standard_name, size0, size1))

                continue
            #--- End: if

            # Still here? Then the defining coordinates have equal
            # size, so check that their data arrays are compatible.
            temp_dim1_name_map = \
                {stdname_to_dim1[standard_name]: stdname_to_dim0[standard_name]}
            
            if coord0._equivalent_data(coord1, temp_dim1_name_map):
                # Equal sized defining coordinates with equivalent
                # data arrays
                dim0 = stdname_to_dim0[standard_name]
                dim1 = stdname_to_dim0[standard_name]
                if standard_name not in stdname_to_aux0:
                    # Defining coordinates are dimension coordinates
                    equivalent_data_matching_dimensions.append((dim0, dim1))

            else:
                if coord0.size > 1:
                    # Equal sized > 1 defining coordinates with
                    # unequivalent data arrays
                    raise ValueError(
"Can't combine fields: Incompatible '%s' coordinate data arrays" % standard_name)
                
                # We have two size 1 defining coordinates with
                # unequivalent data arrays, so flag this dimension to
                # be omitted from the result field.
                dim0 = stdname_to_dim0[standard_name]
                remove_size1_dims.append(dim0)
        #--- End: for

        # --------------------------------------------------------
        # Still here? Then the two fields are combinable!
        # --------------------------------------------------------

        # ------------------------------------------------------------
        # 2.1 Create copies of the two fields, unless it is an in
        # place combination, in which case we don't want to copy self)
        # ------------------------------------------------------------
        field1 = other.copy()

        inplace = method[2] == 'i'
        if not inplace:
            field0 = self.copy()
        else:
            field0 = self

        # Aliases for the field's space and data array
        space0 = field0.space
        space1 = field1.space
        data0  = field0.Data
        data1  = field1.Data

        # ------------------------------------------------------------
        # 2.2 Map the dimension names of space1 to the dimension names
        # of space0.
        #
        # Where this is not possible (because space1 has a dimension
        # which space0 hasn't) then a new space0 dimension name is
        # created.
        # ------------------------------------------------------------

        # Map dimensions in field1 to dimensions in field0
        #
        # For example:
        # >>> dim1_name_map
        # {'dim1': 'dim0', 'dim2': 'dim1', 'dim0': 'dim2'}
        dim1_name_map = {}

        new_dims0 = []

        # Invert the mappings between dimensions and standard names
        dim1_to_stdname = dict((v, k) for k, v in stdname_to_dim1.iteritems())
        dim0_to_stdname = dict((v, k) for k, v in stdname_to_dim0.iteritems())

        for dim1 in space1.dimension_sizes:
            if dim1 in dim1_to_stdname:
                standard_name = dim1_to_stdname[dim1]
                if standard_name in matching_standard_names:
                    dim0                = stdname_to_dim0[standard_name]
                    dim1_name_map[dim1] = dim0
                    continue
            #--- End: if

            # Still here? Then dim1 has no match in space0, so create
            # a new space0 dimension name.
            dim0                = space0.new_dimension()
            dim1_name_map[dim1] = dim0

            # Include this new dimension as a broadcast dimension,
            # regardless of the size of the dimension in space1. This
            # is a sneaky way of, a bit later, getting the coordinates
            # from space1 copied over to space0.
            new_dims0.add(dim0)
        #--- End: for

        # Invert the mappings between
        dim0_name_map = dict((v, k) for k, v in dim1_name_map.iteritems())

        # ------------------------------------------------------------
        # Permute the dimensions of data0 so that
        #
        # 1) All of the matching dimensions are the inner (fastest
        #    varying) dimensions
        #
        # 2) All of the undefined dimensions are the outer (slowest
        #    varying) dimensions
        #
        # 3) All of the defined but unmatched dimensions are in the
        #    middle
        # ------------------------------------------------------------
        axes_U   = []  # Undefined axes
        axes_unM = []  # Defined but unMatched axes
        axes_M   = []  # Defined and Matched axes
        for i, dim0 in enumerate(data0.order):
            if dim0 in undefined_dim0:
                axes_U.append(i)
            elif dim0_to_stdname[dim0] in matching_standard_names:
                axes_M.append(i)
            else:
                axes_unM.append(i)
        #--- End: for
        axes                = axes_U + axes_unM + axes_M
        start_of_unmatched0 = len(axes_U)
        start_of_matched0   = start_of_unmatched0 + len(axes_unM)
        data0.transpose(axes)

        # ------------------------------------------------------------
        # Permute the dimensions of data1 so that
        #
        # * All of the matching dimensions are the inner (fastest
        #   varying) dimensions, in corresponding positions to data0
        #
        # * All of the undefined dimensions are the outer (slowest
        #   varying) dimensions
        #
        # * All of the defined but unmatched dimensions are in the
        #   middle
        # ------------------------------------------------------------
        axes_U   = []
        axes_unM = []
        axes_M   = []
        for i, dim1 in enumerate(data1.order):
            if dim1 in undefined_dim1:
                axes_U.append(i)
            elif dim1_to_stdname[dim1] in matching_standard_names:
                continue
            else:
                axes_unM.append(i)
        #--- End: for
        for dim0 in data0.order[start_of_matched0:]:
            axes_M.append(data1.order.index(dim0_name_map[dim0]))
        #--- End: for
        axes                = axes_U + axes_unM + axes_M
        start_of_unmatched1 = len(axes_U)
        start_of_matched1   = start_of_unmatched1 + len(axes_unM)
        data1.transpose(axes)

        # 2e. For the matching dimensions, make sure that the
        # directions are the same in both data0 and data1
        for i, dim1 in enumerate(data1.order[start_of_matched1:]):
            dim0 = dim1_name_map[dim1]
            if data1.direction[dim1] != data0.direction[dim0]:
                data1.reverse_dims(start_of_matched1+i)
        #--- End: for

        # ------------------------------------------------------------
        # 2f. Insert size 1 dimensions into data0 to correspond to
        # defined but unmatched dimensions in data1
        #
        # For example, if   data0 is          T Y X
        #              and  data1 is        P Z Y X
        #              then data0 becomes P Z T y X
        # ------------------------------------------------------------
        original_start_of_unmatched0 = start_of_unmatched0
        for dim1 in data1.order[start_of_unmatched1:start_of_matched1]:
            data0.expand_dims(start_of_unmatched0,
                              dim1_name_map[dim1],
                              data1.direction[dim1])
            start_of_unmatched0 += 1
            start_of_matched0   += 1
        #--- End: for

        # ------------------------------------------------------------
        # Insert size 1 dimensions into data1 to correspond to defined
        # but unmatched dimensions in data0
        #
        # For example, if   data0 is      P Z T Y X
        #              and  data1 is        P Z Y X
        #              then data1 becomes P Z T Y X
        # ------------------------------------------------------------
        for dim0 in data0.order[start_of_unmatched0:start_of_matched0]:
            data1.expand_dims(start_of_unmatched1,
                              start_of_unmatched1,  # Arbitrary, but unique, name
                              data0.direction[dim0])
            start_of_unmatched1 += 1
        #--- End: for

        # ------------------------------------------------------------
        # Insert size 1 dimensions into data0 to correspond to
        # undefined dimensions (of any size) in data1.
        #
        # For example, if   data0 is          P Z T Y X
        #              and  data1 is      4 1 P Z T Y X
        #              then data0 becomes 1 1 P Z T Y X
        # ------------------------------------------------------------
        start_of_unmatched0 = original_start_of_unmatched0
        for dim1 in data1.order[:start_of_unmatched1]:
            dim0 = dim1_name_map[dim1]
            data0.expand_dims(start_of_unmatched0,
                              dim0,
                              data1.direction[dim1])
            start_of_unmatched0 += 1
        #--- End: for

        # ============================================================
        # 3. Combine the data objects
        #
        # By now, data0.ndim >= data1.ndim and all of the dimension
        # names in data0 are correct for space0. Therefore, the
        # dimension names of the combined data will all be ok.
        # ============================================================
        data0                     = getattr(data0, method)(data1)
        field0.Data               = data0
        space0.dimensions['data'] = data0.order[:]

        # ============================================================
        # 4. Adjust the space of field0 to accommodate its new data
        # ============================================================

        # ------------------------------------------------------------
        # 4a. Remove any size 1 dimensions which are matching
        # dimensions but with different coordinate data array values
        # ------------------------------------------------------------
        for dim in remove_size1_dims:
            field0.squeeze(dim)
            space0.squeeze(dim)
        #--- End: for

        # ------------------------------------------------------------
        # 4b. If broadcasting has grown any size 1 dimensions in
        # space0 then replace their size 1 coordinates with the
        # corresponding size > 1 coordinates from space1
        # ------------------------------------------------------------
        for dim0 in broadcast_dims0 + new_dims0:

            dim1                         = dim0_name_map[dim0]
            space0.dimension_sizes[dim0] = space1.dimension_sizes[dim1]

            # If it exists, copy the space1 dimension coordinate to
            # space0
            if dim1 in space1:
                space0.insert_coordinate(space1[dim1], dim1_name_map,
                                         dim=True,
                                         space=space1, key=dim0)
            #--- End: if

            # Remove from space0 any 1-d auxiliary coordinates for
            # this dimension
            if dim0 in aux_coords0:
                for aux in aux_coords0[dim0]['1-d'].copy():
                    aux_coords0[dim0]['1-d'].remove(aux)
                    space0.remove_coordinate(aux)
            #--- End: if

            # Copy to space0 any space1 1-d auxiliary coordinates for
            # this dimension
            if dim1 in aux_coords1:
                for aux in aux_coords1[dim1]['1-d']:
                    space0.insert_coordinate(space1[aux], dim1_name_map,
                                             aux=True, dimensions=[dim0],
                                             space=space1)
        #--- End: for

        # ------------------------------------------------------------
        # Consolidate any 1-d auxiliary coordinates for matching
        # dimensions whose defining dimension coordinates have
        # equivalent data arrays.
        #
        # A space0 1-d auxiliary coordinate is retained if there is a
        # corresponding space1 1-d auxiliary with the same standard
        # name and equivalent data array.
        # ------------------------------------------------------------
        for dim0, dim1 in equivalent_data_matching_dimensions:

            for aux0 in aux_coords0[dim0]['1-d']:
                # Remove this space0 1-d auxiliary coordinate if it
                # has no standard name
                if space0[aux0].getattr('standard_name', None) is None:
                    space0.remove_coordinate(aux0)
                    continue

                # Still here?
                aux0_has_equivalent_pair = False

                for aux1 in aux_coords1[dim1]['1-d'].copy():
                    if space0[aux0]._equivalent(space1[aux1], dim1_name_map):
                        aux_coords1[dim1]['1-d'].remove(aux1)
                        aux0_has_equivalent_pair = True
                        break
                #--- End: for

                # Remove this space0 1-d auxiliary coordinate if it
                # has no equivalent in space1
                if not aux0_has_equivalent_pair:
                    space0.remove_coordinate(aux0)
        #--- End: for

        # ------------------------------------------------------------
        # Consolidate N-d auxiliary coordinates for matching
        # dimensions which have the same size
        # ------------------------------------------------------------
        # Remove any N-d auxiliary coordinates which span broadcasted
        # dimensions
        for broadcast_dims, aux_coords, space in zip((broadcast_dims0, broadcast_dims1),
                                                     (aux_coords0    , aux_coords1    ),
                                                     (space0         , space1         )):
            for dim in broadcast_dims:
                if dim not in aux_coords:
                    continue

                for aux in aux_coords[dim]['N-d']:
                    del aux_coords['N-d'][aux]
                    space.remove_coordinate(aux)
        #--- End: for

        # Remove any N-d auxiliary coordinates which span a mixture of
        # matching and non-matching dimensions
        for aux_coords, space, dim_to_stdname in zip((aux_coords0    , aux_coords1    ),
                                                     (space0         , space1         ),
                                                     (dim0_to_stdname, dim1_to_stdname)):
            for aux in aux_coords['N-d'].keys():

                # Count how many of this N-d auxiliary coordinate's
                # dimensions are matching dimensions
                aux_coords['N-d'][aux] = set()
                for dim in space.dimensions[aux]:
                    standard_name = dim_to_stdname[dim]
                    if standard_name in matching_standard_names:
                        aux_coords['N-d'][aux].add(standard_name)
                #--- End: for

                n_matching_dims = len(aux_coords['N-d'][aux])
                if 1 <= n_matching_dims < len(space.dimensions[aux]):
                    del aux_coords['N-d'][aux]
                    if space is space0:
                        space.remove_coordinate(aux)
            #--- End: for
        #--- End: for

        # Forget about
        for aux in aux_coords0['N-d'].keys():
             n_matching_dims = len(aux_coords0['N-d'][aux])
             if n_matching_dims == 0:
                 del aux_coords0['N-d'][aux]
        #--- End: for

        # Copy to space0 any space1 N-d auxiliary coordinates which do
        # not span any matching dimensions
        for aux in aux_coords1['N-d'].keys():
             n_matching_dims = len(aux_coords1['N-d'][aux])
             if n_matching_dims == 0:
                 del aux_coords1['N-d'][aux]
                 dims = [dim1_name_map[dim] for dim in space1.dimensions[aux]]
                 space0.insert_coordinate(space1[aux], dim1_name_map,
                                          aux=True, dimensions=dims,
                                          space=space1)
        #--- End: for

        # By now, aux_coords0['N-d'] contains only those N-d auxiliary
        # coordinates which span equal sized matching dimensions.

        # Remove from space0 any N-d auxiliary coordinates which span
        # same-size matching dimensions and do not have an equivalent
        # N-d auxiliary coordinate in space1 (i.e. one which spans the
        # same dimensions, has the same standard name and has
        # equivalent data)
        for aux0 in aux_coords0['N-d']:

            # Remove space0 N-d auxiliary coordinate if it has no
            # standard name
            if not space0[aux0].getattr('standard_name', None):
                space0.remove_coordinate(aux0)
                continue

            # Still here?
            aux0_has_equivalent_pair = False
            for aux1 in aux_coords1['N-d'].keys():
                if space0[aux0]._equivalent(space1[aux1], dim1_name_map):
                    del aux_coords1['N-d'][aux1]
                    aux0_has_equivalent_pair = True
                    break
            #--- End: for

            # Remove space0 N-d auxiliary coordinate if it has no
            # equivalent in space1
            if not aux0_has_equivalent_pair:
                space0.remove_coordinate(aux0)
        #--- End: for

        return field0
    #--- End: def

    def __str__(self):
        '''
x.__str__() <==> str(x)

'''
        string = []

        try:
            space = self.space
        except AttributeError:
            # Field has no space
            dim_names = self.shape
        else:
            # Field has a space
            dim_names = []
            for dim in space.dimensions['data']:
                try:
                    dim_names.append(space[dim].name(ncvar=True,
                                                     default=dim))
                except (KeyError, AttributeError):
                    try:
                        dim_names.append(space.nc_dimensions[dim])
                    except (KeyError, AttributeError):
                        dim_names.append(dim)
            #--- End: for
            dim_names = tuple(dim_names)
        # ---End: try

        dim_names = str(dim_names).replace("'", "")
        dim_names = dim_names.replace(',)', ')', 1)

        string = ['Data            : %s%s' %(self.name(long_name=True,
                                                       ncvar=True,
                                                       default=''),
                                             dim_names)]

        # Cell methods
        try:
            string.append(str(self.cell_methods))
        except AttributeError:
            # Field has no cell methods
            pass

        # Space
        try:
            string.append(str(space))
        except NameError:
            # Space has no space
            string.append('Dimensions      : \nAuxiliary coords: \nCell measures   : ')

        return '\n'.join(string)
    #--- End def

    # ----------------------------------------------------------------
    # Property attribute: Conventions
    # ----------------------------------------------------------------
    @property
    def Conventions(self):
        '''

The Conventions CF attribute.

**Examples**

>>> f.Conventions = 'CF-1.5'
>>> f.Conventions
'CF-1.5'
>>> del f.Conventions

'''
        return self.getattr('Conventions')
    #--- End: def
    @Conventions.setter
    def Conventions(self, value): self.setattr('Conventions', value)
    @Conventions.deleter
    def Conventions(self):        self.delattr('Conventions')

    # ----------------------------------------------------------------
    # Property attribute: ancillary_variables
    # ----------------------------------------------------------------
    @property
    def ancillary_variables(self):
        return self._get_special_attr('ancillary_variables')
    @ancillary_variables.setter
    def ancillary_variables(self, value):
        self._set_special_attr('ancillary_variables', value)
    @ancillary_variables.deleter
    def ancillary_variables(self):
        self._del_special_attr('ancillary_variables')

    # ----------------------------------------------------------------
    # Property attribute: Flags
    # ----------------------------------------------------------------
    @property
    def Flags(self):
        '''

A Flags object containing self-describing CF flag values.

Stores the flag_values, flag_meanings and flag_masks CF attributes in
an internally consistent manner.

**Examples**

>>> f.Flags
<CF Flags: flag_values=[0 1 2], flag_masks=[0 2 2], flag_meanings=['low' 'medium' 'high']>

'''
        return self._get_special_attr('Flags')
    @Flags.setter
    def Flags(self, value):
        self._set_special_attr('Flags', value)
    @Flags.deleter
    def Flags(self):
        self._del_special_attr('Flags')

    # ----------------------------------------------------------------
    # Property attribute: flag_values
    # ----------------------------------------------------------------
    @property
    def flag_values(self):
        '''
The flag_values CF attribute.

Stored as a 1-d numpy array but may be set as any array-like object.

**Examples**

>>> f.flag_values = ['a', 'b', 'c']
>>> f.flag_values
array(['a', 'b', 'c'], dtype='|S1')
>>> f.flag_values = numpy.arange(4)
>>> f.flag_values
array([1, 2, 3, 4])
>>> f.flag_values = 1
>>> f.flag_values
array([1])

'''
        try:
            return self.Flags.flag_values
        except AttributeError:
            raise AttributeError(
                "Can't get '%s' attribute 'flag_values'" %
                self.__class__.__name__)
    #--- End: def
    @flag_values.setter
    def flag_values(self, value):
        try:
            self.Flags.flag_values = value
        except AttributeError:
            self.Flags = Flags(flag_values=value)
    #--- End: def
    @flag_values.deleter
    def flag_values(self):
        try:
            del self.Flags.flag_values
        except AttributeError:
            raise AttributeError(
                "Can't delete '%s' attribute 'flag_values'" %
                self.__class__.__name__)
        else:
            if not self.Flags:
                del self.Flags
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: flag_masks
    # ----------------------------------------------------------------
    @property
    def flag_masks(self):
        '''
The flag_masks CF attribute.

Stored as a 1-d numpy array but may be set as array-like object.

**Examples**

>>> f.flag_masks = numpy.array([1, 2, 4], dtype='int8')
>>> f.flag_masks
array([1, 2, 4], dtype=int8)
>>> f.flag_masks = 1
>>> f.flag_masks
array([1])

'''
        try:
            return self.Flags.flag_masks
        except AttributeError:
            raise AttributeError(
                "Can't get '%s' attribute 'flag_masks'" %
                self.__class__.__name__)
    #--- End: def
    @flag_masks.setter
    def flag_masks(self, value):
        try:
            self.Flags.flag_masks = value
        except AttributeError:
            self.Flags = Flags(flag_masks=value)
    #--- End: def
    @flag_masks.deleter
    def flag_masks(self):
        try:
            del self.Flags.flag_masks
        except AttributeError:
            raise AttributeError(
                "Can't delete '%s' attribute 'flag_masks'" %
                self.__class__.__name__)
        else:
            if not self.Flags:
                del self.Flags
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: flag_meanings
    # ----------------------------------------------------------------
    @property
    def flag_meanings(self):
        '''

The flag_meanings CF attribute.

Stored as a 1-d numpy string array but may be set as a space delimited
string or any array-like object.

**Examples**

>>> f.flag_meanings = 'low medium      high'
>>> f.flag_meanings
array(['low', 'medium', 'high'],
      dtype='|S6')
>>> f.flag_meanings = ['left', 'right']
>>> f.flag_meanings
array(['left', 'right'],
      dtype='|S5')
>>> f.flag_meanings = 'ok'
>>> f.flag_meanings
array(['ok'],
      dtype='|S2')
>>> f.flag_meanings = numpy.array(['a', 'b'])
>>> f.flag_meanings
array(['a', 'b'],
      dtype='|S1')

'''
        try:
            return self.Flags.flag_meanings
        except AttributeError:
            raise AttributeError(
                "Can't get '%s' attribute 'flag_meanings'" %
                self.__class__.__name__)
    #--- End: def
    @flag_meanings.setter
    def flag_meanings(self, value):
        try:
            self.Flags.flag_meanings = value
        except AttributeError:
            self.Flags = Flags(flag_meanings=value)
    #--- End: def
    @flag_meanings.deleter
    def flag_meanings(self):
        try:
            del self.Flags.flag_meanings
        except AttributeError:
            raise AttributeError(
                "Can't delete '%s' attribute 'flag_meanings'" %
                self.__class__.__name__)
        else:
            if not self.Flags:
                del self.Flags
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: cell_methods
    # ----------------------------------------------------------------
    @property
    def cell_methods(self):
        '''

The CellMethods object containing the CF cell methods of the data array.

**Examples**

>>> f.cell_methods
<CF CellMethods: time: mean (interval: 1.0 month)>

'''
        value = self._get_special_attr('cell_methods')
#        if isinstance(value, basestring):
#            cm = CellMethods()
#            value = cm.parse(value, field=self)
        return value
    #--- End: def

    @cell_methods.setter
    def cell_methods(self, value):
        self._set_special_attr('cell_methods', value)

    @cell_methods.deleter
    def cell_methods(self):
        self._del_special_attr('cell_methods')

    # ----------------------------------------------------------------
    # Property attribute: history	
    # ----------------------------------------------------------------
    @property
    def history(self):
        '''

The history CF attribute.

**Examples**

>>> f.history = 'created on 2012/10/01'
>>> f.history
'created on 2012/10/01'
>>> del f.history

'''
        return self.getattr('history')
    #--- End: def

    @history.setter
    def history(self, value): self.setattr('history', value)
    @history.deleter
    def history(self):        self.delattr('history')

    # ----------------------------------------------------------------
    # Property attribute: institution (a simple attribute)
    # ----------------------------------------------------------------
    @property
    def institution(self):
        '''

The institution CF attribute.

**Examples**

>>> f.institution = 'University of Reading'
>>> f.institution
'University of Reading'
>>> del f.institution

'''
        return self.getattr('institution')
    #--- End: def
    @institution.setter
    def institution(self, value): self.setattr('institution', value)
    @institution.deleter
    def institution(self):        self.delattr('institution')

    # ----------------------------------------------------------------
    # Property attribute: references (a simple attribute)
    # ----------------------------------------------------------------
    @property
    def references(self):
        '''

The references CF attribute.

**Examples**

>>> f.references = 'some references'
>>> f.references
'some references'
>>> del f.references

'''
        return self.getattr('references')
    #--- End: def
    @references.setter
    def references(self, value): self.setattr('references', value)
    @references.deleter
    def references(self):        self.delattr('references')

    # ----------------------------------------------------------------
    # Property attribute: standard_error_multiplier	
    # ----------------------------------------------------------------
    @property
    def standard_error_multiplier(self):
        '''

The standard_error_multiplier CF attribute.

**Examples**

>>> f.standard_error_multiplier = 2.0
>>> f.standard_error_multiplier
2.0
>>> del f.standard_error_multiplier

'''
        return self.getattr('standard_error_multiplier')
    #--- End: def

    @standard_error_multiplier.setter
    def standard_error_multiplier(self, value):
        self.setattr('standard_error_multiplier', value)
    @standard_error_multiplier.deleter
    def standard_error_multiplier(self):
        self.delattr('standard_error_multiplier')

    # ----------------------------------------------------------------
    # Property attribute: source	
    # ----------------------------------------------------------------
    @property
    def source(self):
        '''

The source CF attribute.

**Examples**

>>> f.source = 'radiosonde'
>>> f.source
'radiosonde'
>>> del f.source

'''
        return self.getattr('source')
    #--- End: def

    @source.setter
    def source(self, value): self.setattr('source', value)
    @source.deleter
    def source(self):        self.delattr('source')

    # ----------------------------------------------------------------
    # Property attribute: title
    # ----------------------------------------------------------------
    @property
    def title(self):
        '''

The title CF attribute.

**Examples**

>>> f.title = 'model data'
>>> f.title
'model data'
>>> del f.title

'''
        return self.getattr('title')
    #--- End: def

    @title.setter
    def title(self, value): self.setattr('title', value)
    @title.deleter
    def title(self):        self.delattr('title')

    # ----------------------------------------------------------------
    # Property attribute: space
    # ----------------------------------------------------------------
    @property
    def space(self):
        '''

The Space object containing the coordinate system (space).

**Examples**

>>> f.space
<CF Space: (12, 19, 73, 96)>

'''
        return self._get_special_attr('space')
    #--- End: def
    @space.setter
    def space(self, value):
        if not isinstance(value, Space):
            print  value, type(value)
            raise TypeError("asdasdasdsasdas dch ")
        self._set_special_attr('space', value)
    #--- End: def
    @space.deleter
    def space(self):
        self._del_special_attr('space')
    #--- End: def

    def equals(self, other, rtol=None, atol=None, traceback=False):
        '''

True if two fields are logically equal, False otherwise.

The `Conventions` attribute is ignored in the comparison.

:Parameters:

    other :
        The object to compare for equality.

    atol : float, optional
        The absolute tolerance for all numerical comparisons, By
        default the value returned by the `ATOL` function is used.

    rtol : float, optional
        The relative tolerance for all numerical comparisons, By
        default the value returned by the `RTOL` function is used.

    traceback : bool, optional
        If True then print a traceback highlighting where the two
        instances differ.

:Returns: 

    out : bool
        Whether or not the two instances are equal.

**Examples**

>>> f.Conventions
'CF-1.0'
>>> g = f.copy()
>>> g.Conventions = 'CF-1.5'
>>> f.equals(g)
True

In this example, two fields differ only by the long name of their time
coordinates. The traceback shows that they differ in their spaces,
that they differ in their time coordinates and that the long name has
not been matched.

>>> g = f.copy()
>>> g.coord('time').long_name = 'something_else'
>>> f.equals(g, traceback=True)
Coordinate: Different long_name: 'time', 'something else'
Coordinate: Different long_name: 'time', 'latitude in rotated pole grid'
Coordinate: Different long_name: 'time', 'longitude in rotated pole grid'
Space: Different coordinate: <CF Coordinate: time(12)>
Field: Different 'space': <CF Space: (73, 96, 12)>, <CF Space: (73, 96, 12)>
False

'''
        ignore = set(('Conventions',))

        return super(Field, self).equals(other, rtol=rtol, atol=atol, 
                                         traceback=traceback,
                                         ignore=ignore)
    #---End: def

    def reverse_dims(self, axes=None):
        '''

Reverse the directions of data array axes in place.

Equivalent to indexing the specified dimensions with ``::-1``.

:Parameters:

    axes : sequence, optional
        The dimensions to have reversed directions. By default,
        reverse the directions of all dimensions. Dimensions for
        direction reversal may be identified with one of, or a
        sequence of any combination of:

            * A dimension's standard name.
            * The integer position of a dimension in the data array.
            * The field's space's internal name of a dimension

:Returns:

    None

**Examples**

>>> f.reverse_dims()
>>> f.reverse_dims('time')
>>> f.reverse_dims(1)
>>> f.reverse_dims('dim2')
>>> f.reverse_dims(['time', 1, 'dim2'])

'''
        space      = self.space
        dimensions = space.dimensions

        # Convert the axes to integer positions
        axes_integers = []
        if isinstance(axes, (str, int, long)):
            axes = (axes,)
        for axis in axes:
            if isinstance(axis, (int, long)):
                axes_integers.append(axis)
            else:
                key = self.coord(axis, key=True)
                if key is None:
                    raise ValueError("asdasdasdas 98123 gto dch axis")
                axes_integers.append(dimensions[key][0]) # RUM!
        #--- End: for
        axes = axes_integers

        # Reverse the requested dimensions in the data ### ONLY IF
        # THEY'RE IN THE DATA!
        axes = super(Field, self).reverse_dims(axes)

        # Reverse any coordinate and cell measures which span the
        # reverse dimensions
        dimensions   = space.dimensions['data']
        reverse_dims = set((dimensions[i] for i in axes))

        dimensions = space.dimensions

        for key, variable in space.iteritems():
            axes = [i
                    for i, dim in enumerate(dimensions[key])
                    if dim in reverse_dims]
            if axes:
                variable.reverse_dims(axes)
    #--- End: def

    def match(self, attr={}, priv={}, coord={}, cellsize={}):
        '''

Determine whether or not a field satisfies conditions.

Conditions may be specified on the field's attributes, coordinate
values and coordinate cell sizes.

:Parameters:

    attr : dict, optional
        Dictionary for which each key/value pair is a standard or
        non-standard CF attribute name and a condition for the
        attribute to be tested against. If the value is a sequence of
        conditions then the attribute matches if at least one of the
        conditions is passed.

        In general, a condition is any object and it is passed if the
        attribute is equal to the object, with the following
        exception:

            * If the attribute is string-valued, then the condition
              may be a regular expression pattern recognised by the
              `re` module and the condition is passed if the attribute
              matches the regular expression. Special characters for
              the start and end of the string are assumed and need not
              be included. For example, '.*wind' is equivalent to
              '^.*wind$'.

    priv : dict, optional
        Dictionary for which each key/value pair is an attribute name
        other than their standard and non-standard CF attribute and a
        condition for the attribute to be tested against. If the value
        is a sequence of conditions then the attribute matches if at
        least one of the conditions is passed.

        In general, a condition behaves as for `attr`, with the
        following exception:

            * For the `Units` attribute, the condition is passed if
              the attribute is equivalent (rather than equal) to the
              object. (Note that this behaviour does not apply to the
              `units` attribute.)

    coord : dict, optional
        Dictionary for which each key/value pair is a coordinate name
        and a condition for the coordinate to be tested against. A
        match occurs if at least one element of the coordinate's array
        passes the condition. If the value is a sequence of conditions
        then the coordinate matches if at least one of the conditions
        is passed.

    cellsize : dict, optional
        Dictionary for which each key/value pair is a coordinate name
        and a condition for the coordinate's cell sizes to be tested
        against. A match occurs if all of the coordinate's cell's
        sizes pass the condition. Only one cell size condition may be
        given per coordinate.

        If the same coordinate is specified in the *coord* and
        *cellsize* dictionaries then the cell size condition only
        needs to be passed for cells which pass all if the
        coordinate's conditions.

:Returns:

    out : bool
        True if the field satisfies the given criteria, False
        otherwise.

**Examples**

>>> print f
Data            : air_temperature(time, latitude, longitude)
Cell methods    : time: mean
Dimensions      : time(12) = [15, ..., 345] days since 1860-1-1
                : latitude(73) = [-90, ..., 90] degrees_north
                : longitude(96) = [0, ..., 356.25] degrees_east
                : height(1) = [2] m

>>> f.match(attr={'standard_name': 'air_temperature'})
True
>>> f.match(attr={'standard_name': ['air_temperature']})
True
>>> f.match(attr={'standard_name': ['air_temperature', 'air_pressure']})
True
>>> f.match(attr={'standard_name': '.*temperature.*'})
True
>>> f.match(attr={'standard_name': ['.*temperature.*', 'air_pressure']})
True
>>> f.match(attr={'standard_name': '.*pressure.*'})
False

>>> f.match(attr={'Units': 'K'})
True
>>> f.match(attr={'Units': cf.Units('1.8 K @ 459.67')})
True
>>> f.match(attr={'Units': [cf.Units('Pa'), 'K']})
True
>>> f.match(attr={'Units': cf.Units('Pa')})
False

>>> f.match(attr={'cell_methods': 'time: mean'})
True
>>> f.match(attr={'cell_methods': cf.CellMethods('time: mean')})
True
>>> f.match(attr={'cell_methods': ['time: mean', 'time: max']})
True
>>> f.match(attr={'cell_methods': cf.CellMethods('time: max')})
False
>>> f.match(attr={'cell_methods': 'time: mean time: min')})
False

>>> f.match(coord={'latitude': 0})
False
>>> f.match(coord={'latitude': [0, cf.gt(30)]})
True

>>> f.match(cellsize={'time': cf.inside(28, 31, 'days')})
True

'''
        attr = attr.copy()

        # ------------------------------------------------------------
        # Try to match other cell methods
        # ------------------------------------------------------------
        cell_methods = attr.pop('cell_methods', None)

        if cell_methods is not None:
            if not hasattr(self, 'cell_methods'):
                return False

            if not isinstance(cell_methods, (tuple, list)):
                cell_methods = (cell_methods,)

            match = False
            for v in cell_methods:
                if isinstance(v, basestring):
                    v = CellMethods(v)

                if self.cell_methods.has_cellmethods(v):
                    match = True
                    break
                else:
                    continue
            #--- End: for
            if not match:
                return False
        #--- End: if

        # ------------------------------------------------------------
        # Try to match other attributes
        # ------------------------------------------------------------
        if not super(Field, self).match(attr=attr, priv=priv):
            return False

        # Still here? Then the attributes have matched.

        # ------------------------------------------------------------
        # Try to match coordinate values
        # ------------------------------------------------------------
        if coord:
            # Some coordinate conditions have been been specified

            # Keep a note of found coordinates to save us possibly
            # having to look them up again if there are cellsize
            # conditions
            coords = {}

            # Loop round coordinate criteria
            for name, value in coord.iteritems():

                if value is None:
                    continue

                c = self.coord(name, exact=True)

                if not c:
                    # The field does not have a coordinate with this name,
                    # so it does not match.
                    return False

                coord_match = None

                if not isinstance(value, (tuple, list)):
                    value = (value,)

                for v in value:
                    c_match = (c.Data == v)
                    if not c_match.any():
                        continue

                    match = True
                    if coord_match is None:
                        coord_match = c_match
                    else:
                        coord_match = c_match | coord_match
                #--- End: for

                if coord_match is None:
                    return False

                coords[name] = (c, coord_match)
            #--- End: for
        #--- End: if

        # ------------------------------------------------------------
        # Try to match coordinate cell sizes
        # ------------------------------------------------------------
        if cellsize:
            # Some coordinate conditions have been been specified

            for name, value in cellsize.iteritems():
                if value is None:
                    continue

                if name in coords:
                    c, coord_match = coords[name]
                else:
                    c = self.coord(name, exact=True)
                    coord_match = None
                #--- End: if

                if not c:
                    # The field does not have a coordinate with this name,
                    # so it does not match.
                    return False

                if not hasattr(c, 'bounds'):
                    if value != 0:
                        # The coordinate does not have any bounds and the
                        # specified cell size is not zero, so it does not
                        # match.
                        return False
                    else:
                        continue
                #--- End: if

                b  = c.bounds.Data
                cs = abs(b[...,1] - b[...,0])
                cs.squeeze(cs.ndim-1)

                bounds_match = (value == cs)

                if coord_match is None:
                    if not bounds_match.all():
                        return False
                else:
                    if not ((bounds_match & coord_match) == coord_match).all():
                        return False
            #--- End: for
        #--- End: if

        return True
    #--- End: def

    def squeeze(self, axes=None):
        '''

Remove size 1 dimensions from the field's data array in place.

:Parameters:

    axes : optional
        The size 1 axes to remove. By default, all size 1 axes are
        removed. Size 1 axes for removal may be identified with one
        of, or a sequence of any combination of:

            * A dimension's standard name.
            * The integer position of a dimension in the data array.
            * The field's space's internal name of a dimension
:Returns:

    None

**Examples**

>>> f.squeeze()
>>> f.squeeze('time')
>>> f.squeeze(1)
>>> f.squeeze('dim2')
>>> f.squeeze([1, 'time', 'dim2'])

'''
        space           = self.space
        dimensions      = space.dimensions
        data_dimensions = dimensions['data']

        # ------------------------------------------------------------
        # Squeeze everything if no axes have been given
        # ------------------------------------------------------------
        if axes is None:
            # Squeeze the data
            axes = super(Field, self).squeeze()

            # Remove the squeezed dimensions from the space's list of
            # data dimensions
            if axes:
                data_dimensions[:] = [dim
                                      for i, dim in enumerate(data_dimensions)
                                      if i not in axes]
            #--- End: if

            return
        #--- End: if

        # ------------------------------------------------------------
        # Still here? Then squeeze only the specified dimensions
        # ------------------------------------------------------------

        if isinstance(axes, (str, int, long)):
            axes = (axes,)

        for i, axis in enumerate(axes):
            if axis in data_dimensions:
                axes[i] = data_dimensions.index(dim)
            elif isinstance(axis, (int, long)):
                pass
            else:
                key = self.coord(axis, key=True, exact=True)
                
                if (key is None                               or
                    len(dimensions[key]) > 1                  or
                    dimensions[key][0] not in data_dimensions):
                    raise ValueError(
                        "Can't determine '%s' squeeze dimension" %
                        self.__class__.__name__)
                #--- End: if

                axes[i] = data_dimensions.index(dimensions[key][0])
        #--- End: for

        if axes:
            # Squeeze the data
            axes = super(Field, self).squeeze(axes)

            # Remove the squeezed dimensions from the space's list of
            # data dimensions
            data_dimensions[:] = [dim
                                  for i, dim in enumerate(data_dimensions)
                                  if i not in axes]
        #--- End: if

        return axes
    #--- End: def

    def unsqueeze(self):
        '''

Insert size 1 dimensions from the field's space into its data array in
place.

:Returns:

    None

**Examples**

>>> f.unsqueeze()

'''
        space           = self.space
        dimensions      = space.dimensions
        data_dimensions = dimensions['data']

        # --------------------------------------------------------
        # Unsqueeze everything
        # --------------------------------------------------------
        # Expand any the data
        for dim, size in space.dimension_sizes.iteritems():
            if size == 1 and dim not in data_dimensions:
                self.expand_dims(dim=dim)
        #--- End: for

        return
    #--- End: def

    def transpose(self, axes=None):
        '''

Permute the dimensions of the data array in place.

:Parameters:

    axes : sequence, optional
        The new order of the data array. By default, reverse the
        dimensions' order, otherwise the axes are permuted according
        to the values given. The values of the sequence may be any
        combination of:

            * A dimension's standard name.
            * The integer position of a dimension in the data array.
            * The field's space's internal name of a dimension

:Returns:

    None

**Examples**

>>> f.transpose()
>>> f.transpose(['latitude', 'time', 'longitude'])
>>> f.transpose([1, 0, 2])
>>> f.transpose((1, 'time', 'dim2'))

'''
        space           = self.space
        dimensions      = space.dimensions
        data_dimensions = dimensions['data']

        # Convert axes to a list of integer positions
        if isinstance(axes, (str, int, long)):
            axes = [axes]
        else:
            axes = list(axes)

        for i, axis in enumerate(axes):
            if axis in data_dimensions:
                axes[i] = data_dimensions.index(axis)
            elif not isinstance(axis, (int, long)):
                key = self.coord(axis, key=True, exact=True)

                if (key is None                               or
                    len(dimensions[key]) > 1                  or
                    dimensions[key][0] not in data_dimensions):
                    raise ValueError(
                        "Can't determine %s dimension from '%s'" %
                        (self.__class__.__name__, axis))

                axes[i] = data_dimensions.index(dimensions[key][0])
        #--- End: for

        # Transpose the field
        super(Field, self).transpose(axes)

        # Reorder the list of dimensions in the space
        dimensions    = dimensions['data']
        dimensions[:] = [dimensions[i] for i in axes]
    #--- End: def

    def expand_dims(self, arg=None, axis=0):
        '''

Expand the shape of the data array in place.

Insert a new size 1 axis, corresponding to a given position in the
data array shape.

:Parameters:

    arg : str, optional
        The dimension to insert. if specified, the dimension should
        exist in the field's space and is identified by its standard
        name or by the space's internal dimension name. By default,
        insert a new dimension which doesn't yet exist in the field's
        space.

    axis : int, optional
        Position (amongst axes) where new axis is to be inserted. By
        default, insert at position 0.

:Returns:

    None

**Examples**

>>> f.expand_dims()
>>> f.expand_dims(axis=1)
>>> f.expand_dims('height')
>>> f.expand_dims('height', 3)
>>> f.expand_dims('dim1', 3)

'''
        space           = self.space
        dimension_sizes = space.dimension_sizes
        dimensions      = space.dimensions
        data_dimensions = dimensions['data']

        if arg is None:
            dim = space.new_dimension()

        elif arg in dimension_sizes:
            if dimension_sizes[arg] != 1:
                raise ValueError(
                    "Can't insert a dimension with size > 1 into %s: '%s' (size %d)" %
                    (self.__class__.__name__, arg, dimension_sizes[arg]))

            if arg in data_dimensions:
                raise ValueError(
                    "Can't insert a duplicate dimension into %s: '%s'" %
                    (self.__class__.__name__, arg))

            dim = arg

        else:
            key = self.coord(arg, key=True, exact=True)

            if key is None or len(dimensions[key]) > 1:
                raise ValueError("9889asdasdasdaaaaaas 98123 gto dch axis")

            dim = dimensions[key][0]

            if dimension_sizes[dim] > 1:
                raise ValueError(
                    "Can't insert into %s Data a dimension with size > 1: '%s' ('%s') (size %d)" %
                    (self.__class__.__name__, dim, arg, dimension_sizes[dim]))

            if dim in data_dimensions:
                raise ValueError(
                    "Can't insert into %s Data a duplicate dimension: '%s' ('%s')" %
                    (self.__class__.__name__, dim, arg))
        #--- End: if

        super(Field, self).expand_dims(axis, dim, space.direction(dim))

        data_dimensions.insert(axis, dim)

        # Make sure that dimension_sizes is up to date
        if dim not in dimension_sizes:
            dimension_sizes[dim] = 1

        return dim
    #--- End: def

    def chunk(self, chunksize=None, extra_boundaries=None, chunk_dims=None):
        '''

Partition the field in place for LAMA functionality.

:Parameters:

    chunksize : int, optional

    extra_boundaries : sequence of lists or tuples, optional

    chunk_dims : sequence of lists or tuples, optional

:Returns:

    extra_boundaries, chunk_dims : {list, list}

'''

#### With the new aggregation algorithm we probabalby don't need to do
#### this - we can just inherit from Variable !!!!!!!!!!!!!!!!!!!!!!!!
#### because it ceases to matter if partitions are in different places
#### in coordinate and teh field (I think .... ?) ????

        if not chunksize:
            chunksize = CHUNKSIZE()

        extra_boundaries, chunk_dims = self.Data.chunk(chunksize,
                                                       extra_boundaries,
                                                       chunk_dims)

        if not extra_boundaries:
            return extra_boundaries, chunk_dims

        # Still here?
        space      = self.space
        dimensions = space.dimensions

        for key, variable in space.iteritems():
            variable_dimensions = dimensions[key]
            for extra_bounds, dim in zip(extra_boundaries, chunk_dims):
                if dim in variable_dimensions:
                    # This variable spans the chunking dimension, so
                    # chunk it with the same new partition locations
                    # as the field's data
                    variable.chunk(None, [extra_bounds], [dim])
        #--- End: for

        return extra_boundaries, chunk_dims
    #--- End: def

    def finalize(self, chunksize=None):
        '''

Finalize a newly created field.

It is essential that this is carried out on every new field to ensure
that methods and functions of fields work correctly. Finalization
entails:

* Expanding scalar coordinate and cell measures to 1-d.

* Setting the direction of the space's dimensions.

* Conforming the space's internal dimension names.

* Partitioning the data arrays for LAMA functionality.

:Parameters:

    chunksize : int, optional
        Set the maximum size in bytes of sub-arrays for data storage
        and processing. By default the size used by the `CHUNKSIZE`
        function is used.

:Returns:

    None

**Examples**

>>> f.finalize()
>>> f.finalize(2**30)

'''
        space           = self.space
        dimensions      = space.dimensions
        data_dimensions = dimensions['data']

        # ------------------------------------------------------------
        # Find the direction of each dimension
        # ------------------------------------------------------------
        directions = {}
        for dim in space.dimension_sizes:
            if dim in space:
                directions[dim] = space[dim]._get_direction()
            else:
                directions[dim] = True
        #--- End: for

        self._set_Data_attributes(data_dimensions, directions)

        for key, variable in space.iteritems():
            variable._set_Data_attributes(dimensions[key], directions)

        # ------------------------------------------------------------
        # Turn scalar coordinates and cell measures into size 1, 1-d
        # ------------------------------------------------------------
        for key, variable in space.iteritems():
            if variable.is_scalar:
                dim = dimensions[key][0]
                variable.expand_dims(0, dim, variable.Data.direction)
        #--- End: for

        # ------------------------------------------------------------
        # Make sure that each partition is no larger than the chunk
        # size
        # ------------------------------------------------------------
        self.chunk(chunksize)
    #--- End: def

    def dump(self, id=None):
        '''

Return a string containing a full description of the instance.

:Parameters:

    id : str, optional
        Set the common prefix of component names. By default the
        instance's class name is used.

:Returns:

    out : str
        A string containing the description.

**Examples**

>>> x = f.dump()
>>> print f.dump()
>>> print f.dump(id='pressure_field')

'''
        if id is None:
            id = self.__class__.__name__

        # Produce summary
        string = ['field summary\n-------------']
        string.append(str(self))
        string.append('')

        # Produce dump
        name = self.name(default='', long_name=True)
        string.append('%s field' % name)
        string.append(''.ljust(len(string[-1]), '-'))

        string.append(super(Field, self).dump(id=id, omit=('Flags',
                                                           'space',)))

        if hasattr(self, 'Flags'):
            string.append(self.Flags.dump(id='%(id)s' % locals()))

        if hasattr(self, 'space'):
            string.append(self.space.dump(id='%(id)s.space' % locals()))

        return '\n'.join(string)
    #--- End: def

    def coord(self, arg,
              role=None, key=False, exact=False, maximal_match=True):
        '''

Return a coordinate of the field.

Note that the returned coordinate is an object identity to the
coordinate stored in the space so, for example, a coordinate's
attributes may be changed in-place:

>>> f.coord('height').long_name
AttributeError: 'Coordinate' object has no attribute 'long_name'
>>> f.coord('hei').long_name = 'HEIGHT'
>>> f.coord('heigh').long_name
'HEIGHT'

Or a deep copy may be made:

>>> c = f.coord('height').copy()

:Parameters:

    arg : str or dict or int
        The identify of the coordinate. One of:

        * A string containing (an abbreviation) of its standard name
          (such as 'time').

        * A string containing its identifier in the field's space
          (such as 'dim2').

        * A dictionary containing one or more (abbreviations of )
          attributes names and their values as its key/value pairs
          (such as {'long_name': 'something'}). If two or more
          attributes are specified then the returned coordinate must
          satisfy all criteria, unless `maximal_match` is False in
          which case the returned coordinate will satisfy all criteria
          at least one of the criteria.

        * An integer given the position of a dimension in the field's
          data array.

    exact : bool, optional
        If True then assume that the value of a name given by `arg` is
        exact. By default it is considered to be an abbreviation.

    key : bool, optional
        Return the field's space's identifier for the coordinate.

    role : bool, optional
        Restrict the search to coordinates of the given role. Valid
        values are 'dim' and 'aux', for dimension and auxiliary
        coordinate constructs. By default both types are considered.

:Returns:

    out : Coordinate or str or None
        The requested coordinate, or the field's space's identifier
        for the coordinate, or `None` if no coordinate could be found.

**Examples**

>>> f.coord('lon')
<CF Coordinate: longitude(128)>
>>> f.coord('lon', key=True)
'dim2'
>>> f.coord('lonX', key=True)
None
>>> f.coord('lon', exact=True)
None
>>> f.coord('longitude', exact=True)
<CF Coordinate: longitude(128)>

'''
        if hasattr(self, 'space'):
            return self.space.coord(arg,
                                    role=role, key=key, exact=exact,
                                    maximal_match=maximal_match)
        #--- End: if

        return None
    #--- End: def

    def copy(self):
        '''

Return a deep copy.

Equivalent to ``copy.deepcopy(f)``

:Returns:

    out :
        The deep copy.

**Examples**

>>> g = f.copy()

'''
        new = super(Field, self).copy(_omit_special=('space',))

        if hasattr(self, 'space'):
            new.space = self.space.copy()

        return new
    #--- End: def

    @property
    def subset(self):
        '''

Return a new field whose data and space are subsetted in a consistent
manner.

This attribute may be indexed to select a subset from dimension index
values (square brackets: ``f.subset[indices]``) or called to select a
subset from dimension coordinate array values (round brackets:
``f.subset(coordinate_values)``)

**Subsetting by indexing**

Subsetting by dimension indices uses an extended Python slicing
syntax, which is similar numpy array indexing. There are two
extensions to the numpy indexing functionality:

* Size 1 dimensions are never removed.

  An integer index i takes the i-th element but does not reduce the
  rank of the output array by one.

* When advanced indexing is used on more than one dimension, the
  advanced indices work independently.

  When more than one dimension's slice is a 1-d boolean array or 1-d
  sequence of integers, then these indices work independently along
  each dimension (similar to the way vector subscripts work in
  Fortran), rather than by their elements.

**Subsetting by coordinate values**

Subsetting by coordinate values allows a subsetted field to be defined
by particular coordinate values of its space.

Subsetting by coordinate values is functionally equivalent to
subsetting by indexing - internally, the selected coordinate values
are in fact converted to dimension indices.

Coordinate values are provided as arguments to a call to the subset
method.

The benefits to subsetting in this fashion are:

* The dimensions to be subsetted are identified by name.

* The position in the data array of each dimension need not be
  known.

* Dimensions for which no subsetting is required need not be
  specified.

* Size 1 dimensions of the space which are not spanned by the data
  array may be specified.

**Examples**

'''
        return SubsetField(self)
    #--- End: def

#--- End: class


# ====================================================================
#
# Field slice object
#
# ====================================================================

class SubsetField(SubsetVariable):
    '''

Return a new field whose data and space are subsetted in a consistent
manner.

Return a new Field instance whose data and grid are sliced according
to indices or call arguments to its `slice` attribute. The latter
converts criteria on coordinate values into indices and so the two
methods a formally equivalent.

**Examples**

>>> print s
Data            : air_temperature(time, latitude, longitude)
Cell methods    : time: mean (interval: 1.0 month)
Dimensions      : time(7070) -> 450-11-16 00:00:00 to 1049-12-16 12:00:00
                : latitude(64) -> -87.8638 to 87.8638 degrees_north
                : longitude(128) -> 0 to 357.1875 degrees_east
                : height(1) -> 2 m
Auxiliary coords:
>>> s.subset[:,32, 64]
<CF Field: air_temperature(12, 1, 1)
>>> s.slice(longitude=0.0, latitude=0.0)
<CF Field: air_temperature(12, 1, 1)

Slicing by index
----------------
Refer to the 'Copying and resizing' of the `cf` help.

Slicing by coordinate criteria
------------------------------

Coordinates are identified by keyword parameters of their standard
names. If the optional positional argument `exact` evaluates to `False`
(the default) then unambiguous abbreviations of the standard names are
accepted, otherwise exact standard names are required:

>>> s.slice(longitude=0.0)
<CF Field: air_temperature(12, 64, 1)
>>> s.slice(lon=0.0)
<CF Field: air_temperature(12, 64, 1)
>>> s.slice(l=0.0)
ValueError: Can't identify coordinate from keyword 'l'
>>> s.slice(False, lon=0.0)
<CF Field: air_temperature(12, 64, 1)
>>> s.slice(True, lon=0.0)
ValueError: Can't identify coordinate from keyword 'lon'

The coordinate criteria are given by the keyword parameters
values. Four different methods are available:

1. A scalar. The dimension is sliced where coordinate equals (using
   the coordinate's == operator) the scalar value.

   **Example**
   >>> s.slice(lat=0.0)

2. A sequence. The dimension is sliced where the coordinate equals
   (using the coordinate's == operator) the sequence's elements. All
   elements of the sequence must be found in the coordinate, but the
   sequence may contain fewer elements than the coordinate's data.

   **Example**
   >>> s.slice(lon=[0.0, 2.8125])

3. A :class:`~cf.Comparison` object (as, for example, returned by the
   :func:`~cf.lt` function). The dimension is sliced where the
   comparison for the coordinate evaluates to True.

   **Example**
   >>> s.slice(lat=cf.lt(0.0))

4. An numpy array of booleans with the same shape as the
   coordinate. The dimension is sliced where elements of this array
   are True.

   **Example**
   >>> s.slice(lat=s.coord('lat')<0.0)

If a coordinate is not identified with a keyword parameter then its
dimension is returned unchanged.

The order in which coordinates are given in the parameter list is
irrelevant.

A call with no keyword parameters is equivalent to indexing with an
Ellipsis (...).

**Examples**

>>> f.slice(lat=0.0, lon=0.0)
<CF Field: air_temperature(12, 1, 1)
>>> f.slice(lon=cf.lt(0.0), lon=[0, 2.8125])
<CF Field: air_temperature(12, 32, 2)
>>> f.slice(lon=[0, 1])
ValueError: 'longitude' coordinate doesn't contain the value 1
>>> f.slice(lon=1)
ValueError: No indices found for 'longitude' coordinate

'''
    __slots__ = []

    def __getitem__(self, indices):
        '''
x.__getitem__(indices) <==> x[indices]

'''

        field = self.variable

        # Copy the field, omitting its data
        new = field.copy()

        # Parse the index
        indices = parse_indices(field, indices)

        # Slice the field's data
        new.Data = field.Data[tuple(indices)]

        # Break ancillary variables
        full_slice = (slice(None),) * field.ndim
        if indices != full_slice:
            try:
                del new.ancillary_variables
            except AttributeError:
                pass

        if not hasattr(new, 'space'):
            # Return now if the field doesn't have a space
            return new

        # -----------------------------------------------------------
        # Still here? Then slice the space by slicing every variable
        # it contains which spans a sliced dimension
        # -----------------------------------------------------------
        space = new.space
        i = -1
        for dim in space.dimensions['data'][:]:

            i += 1

            # Continue if no slicing to be done on this dimension
            if indices[i] == slice(None, None, None): # dch ??
                continue

            for key in space:

                if dim not in space.dimensions[key]:
                    # This space component does not span the
                    # current dimension so continue
                    continue

                dice       = [slice(None)] * space[key].ndim
                pos        = space.dimensions[key].index(dim)
                dice[pos]  = indices[i]
                space[key] = space[key].subset[tuple(dice)]

                # Update the space's dimension size
                space.dimension_sizes[dim] = space[key].shape[pos]
            #--- End: for
        #--- End: for

        if indices != full_slice:  # DCH indices will *never* equal full_slice !!!
            # Break formula_terms
            for key, transform in new.space.transforms.items():
                if transform.is_grid_mapping:
                    continue
                
                for variable in new.space.itervalues():
                    if getattr(variable, 'transform', None) == key:
                        del variable.transform

                del new.space.transforms[key]
            #--- End: for
        #--- End: if

        return new
    #--- End: def

    def __call__(self, *args, **kwargs):
        '''

Slice by coordinate value. refer to :meth:`cf.Field.subset` for
details.

'''

        field = self.variable
        space = field.space

        if not (args or kwargs):
            return field.copy()

        dimensions = []

        # Initialize index
        indices = [slice(None)] * field.ndim

        # Loop round slice criteria
        for name, value in kwargs.iteritems():

            key = field.coord(name, key=True, exact=True)

            if key is None:
                raise ValueError(
                    "Can't identify coordinate '%(name)s'" % locals())

            c = space[key]

            if key.startswith('dim'):
                # Dimension coordinate
                dim = key
            else:
                # Auxiliary coordinate
                dim = space.dimensions[key][0]

                if c.ndim > 1:
                    raise ValueError(
                        "Can't derive an index from multidimensional coordinate '%s'" %
                        c.name())
            #--- End: if

            if dim in dimensions:
                raise ValueError("Two coordinates share the same dimension")

            dimensions.append(dim)

            coord_match = None

            if not isinstance(value, (tuple, list)):
                value = (value,)

            for v in value:
                c_match = (c.Data == v)
                if not c_match.any():
                    continue

                if coord_match is None:
                    coord_match = c_match
                else:
                    coord_match = c_match | coord_match
            #--- End: for

            if coord_match is None:
                raise IndexError(
                    "No indices found for '%s' values %s" %
                    (c.name(), v))

            if dim not in space.dimensions['data']:
                continue

            # Add slice_ to the right place in the data slice
            pos          = space.dimensions['data'].index(dim)
            indices[pos] = coord_match.array
        #--- End: for

        if not args:
            return field.subset[tuple(indices)]

        value = args[0]
        if not isinstance(value, field.__class__):
            value = deepcopy(value)

        field.subset[tuple(indices)] = value
    #--- End: def

#--- End: class
