import numpy
from tempfile import mkstemp
from os       import close
from copy     import deepcopy
from .utils import parse_indices, subset_array, CONSTANTS

# ====================================================================
#
# FileArray object
#
# ====================================================================

class FileArray(object):
    ''' 

A numpy array stored on disk in a temporary file.

'''
    __slots__ = ('_masked_as_record',
                 '_partition_file',
                 'dtype',
                 'size',
                 'ndim',
                 'shape',
                 )

    def __init__(self, array):
        '''

**Initialization**

:Parameters:

    array : numpy array
        The array to be stored on disk in a temporary file.        

**Examples**

>>> f = FileArray(numpy.ma.array([1, 2, 3, 4, 5]))

'''
        # ------------------------------------------------------------
        # Use mkstemp because we want to be responsible for deleting
        # the temporary file when done with it.
        # ------------------------------------------------------------
        fd, _partition_file = mkstemp(prefix='cf_array_', suffix='.npy', 
                                      dir=CONSTANTS['TEMPDIR'])

        close(fd)

        # The name of the temporary file
        self._partition_file = _partition_file
        '''

The name of the temporary file storing the array.

**Examples**

>>> f._partition_file
'/tmp/cf_array_B8SSw2.npy'

'''


        self.dtype = array.dtype
        '''

Numpy data-type of the array.

**Examples**

>>> f.dtype
dtype('float64')

'''

        self.shape = array.shape
        '''

Tuple of the array's dimension sizes.

**Examples**

>>> f.shape
(73, 96)

'''

        self.size = long(array.size)
        '''

Number of elements in the array.

**Examples**

>>> f.shape
(73, 96)
>>> f.size
7008

'''

        self.ndim = array.ndim
        '''

Number of dimensions in the array.

**Examples**

>>> f.shape
(73, 96)
>>> f.ndim
2

'''

        if isinstance(array, numpy.ma.core.MaskedArray):
            # Array is a masked array.
            #
            # Note that numpy.save doesn't work yet for masked arrays
            # (numpy vn1.7). But we can work around this by converting
            # to a record array and then creating a masked array from
            # the data and mask elements.
            numpy.save(_partition_file, array.toflex())
            self._masked_as_record = True
        else:
            # Array is not a masked array.
            numpy.save(_partition_file, array)
            self._masked_as_record = False
    #--- End: def

    def __getitem__(self, indices):
        '''
x.__getitem__(indices) <==> x[indices]

'''
        array = numpy.load(self._partition_file)

        indices = parse_indices(array, indices)

        array = subset_array(array, indices)

        if self._masked_as_record:
            # Convert a record array to a masked array
            array = numpy.ma.array(array['_data'], mask=array['_mask'],
                                   copy=False)
            array.shrink_mask()
        #--- End: if

        # Return the array
        return array
    #--- End: def

    def copy(self):
        '''

Return a deep copy.

Equivalent to ``copy.deepcopy(f)``.

:Returns:

    out :
        A deep copy.
    
**Examples**

>>> f.copy()

'''  
        return deepcopy(self)
    #--- End: def

#--- End: class
