import netCDF4
import numpy
import re
from operator import mul
from .coordinate       import Coordinate
from .coordinatebounds import CoordinateBounds
from .cellmeasures     import CellMeasures
from .space            import Space
from .transform        import Transform
from .field            import Field
from .cellmethods      import CellMethods
from .fieldlist        import FieldList
from .data             import Data
from .units            import Units


def read(filename, close=False):
    ''' 

Read fields from an input netCDF file on disk or from an OPeNDAP
server location.

The file may be big or little endian.

NetCDF dimension names are stored in the `nc_dimensions` attribute of
a field's space and netCDF variable names are stored in the `ncvar`
attributes of the field and its space components (coordinates,
coordinate bounds, cell measures and transformations).

:Parameters:

    filename : str or file
        A string giving the file name or OPenDAP URL, or an open file
        object, from which to read fields. Note that if a file object
        is given it will be closed.

    close : bool, optional
        If True then close the file after it has been read.

:Returns:

    out : FieldList
        The fields in the file.

**Examples**

>>> f = cf.netcdf.read('file.nc')
>>> type(f)
<class 'cf.field.FieldList'>
>>> f
[<CF Field: pmsl(30, 24)>,
 <CF Field: z-squared(17, 30, 24)>,
 <CF Field: temperature(17, 30, 24)>,
 <CF Field: temperature_wind(17, 29, 24)>]

>>> cf.netcdf.read('file.nc')[0:2]
[<CF Field: pmsl(30, 24)>,
 <CF Field: z-squared(17, 30, 24)>]

>>> cf.netcdf.read('file.nc', units='K')
[<CF Field: temperature(17, 30, 24)>,
 <CF Field: temperature_wind(17, 29, 24)>]

>>> cf.netcdf.read('file.nc').extract(units='K')
[<CF Field: temperature(17, 30, 24)>,
 <CF Field: temperature_wind(17, 29, 24)>]

>>> cf.netcdf.read('file.nc')[0]
<CF Field: pmsl(30, 24)>

'''

    if isinstance(filename, file):
        name = filename.name
        filename.close()
        filename = name
    #--- End: if
    
    # Read the netCDF file
    nc = netCDF4.Dataset(filename, 'r')

    # Set of all of the netCDF variable names in the file.
    #
    # For example:
    # >>> variables
    # set(['lon','lat','tas'])
    variables = set(map(str, nc.variables))

    # ----------------------------------------------------------------
    # Put the file's global attributes into the global
    # 'global_attributes' dictionary
    # ----------------------------------------------------------------
    global_attributes = {}
    for attr in map(str, nc.ncattrs()):
        try:
            value = nc.getncattr(attr)
            if isinstance(value, basestring):
                try:
                    global_attributes[attr] = str(value)
                except UnicodeEncodeError:
                    global_attributes[attr] = value           
            else:
                global_attributes[attr] = value     
        except UnicodeDecodeError:
            pass
    #--- End: for
    
    # ----------------------------------------------------------------
    # Create a dictionary keyed by nc variable names where each key's
    # value is a dictionary of that variable's nc
    # attributes. E.g. attributes['tas']['units']='K'
    # ----------------------------------------------------------------
    attributes = {}
    for ncvar in variables:
        attributes[ncvar] = {}
        for attr in map(str, nc.variables[ncvar].ncattrs()):
            try:
                attributes[ncvar][attr] = nc.variables[ncvar].getncattr(attr)
                if isinstance(attributes[ncvar][attr], basestring):
                    try:
                        attributes[ncvar][attr] = str(attributes[ncvar][attr])
                    except UnicodeEncodeError:
                        attributes[ncvar][attr] = attributes[ncvar][attr]                   
            except UnicodeDecodeError:
                pass
        #--- End: for

        # Set default units and check for bad units
        if 'units' in attributes[ncvar]:
            # Check for bad units
            try:
                Units(units=attributes[ncvar]['units'])
            except TypeError:
                # Units have been set to unknown units so 1) give a
                # warning 2) set the 'nonCF_units' attribute to the
                # bad units 3) set the 'units' attribute to
                # None. Setting the units to None may cause exceptions
                # when trying to do stuff on this variable (like
                # arithmetic operations).
                print \
"WARNING: Moving unsupported units to 'nonCF_units': '%s'" % \
attributes[ncvar]['units']
                attributes[ncvar]['nonCF_units'] = attributes[ncvar].pop('units')
        #--- End: if
    
        # Check for bad calendar
        if 'calendar' in attributes[ncvar]:
            try:
                Units(calendar=attributes[ncvar]['calendar'])
            except TypeError:
                print \
"WARNING: Moving unsupported calendar to 'nonCF_calendar': '%s'" % \
attributes[ncvar]['calendar']
                attributes[ncvar]['nonCF_calendar'] = attributes[ncvar].pop('calendar')
    #--- End: for

    # ----------------------------------------------------------------
    # Remove everything bar data variables from the list of
    # variables. I.e. remove dimension and auxiliary coordinates,
    # their bounds and grid_mapping variables
    # ----------------------------------------------------------------
    for ncvar in variables.copy():

        # Remove dimension coordinates and their bounds
        if ncvar in map(str, nc.dimensions):

            if ncvar in variables:
                variables.discard(ncvar)
                for attr in ('bounds', 'climatology'):
                    if attr not in attributes[ncvar]:
                        continue
                
                    # Check the dimensionality of the coordinate's
                    # bounds. If it is not right, then it can't be a
                    # bounds variable and so promote to an independent
                    # data variable
                    bounds = attributes[ncvar][attr]
                    if bounds in nc.variables:
                        if nc.variables[bounds].ndim == nc.variables[ncvar].ndim+1:
                            variables.discard(bounds)
                        else:
                            del attributes[ncvar][attr]

                        break
                    else:
                        print \
"WARNING: Missing bounds variable '%(bounds)s' in %(filename)s" % \
locals()
                        del attributes[ncvar][attr]
                #--- End: for
            #--- End: if

            continue
        #--- End: if
    
        # Still here? Then remove auxiliary coordinates and their
        # bounds
        if 'coordinates' in attributes[ncvar]:
            # Allow for (incorrect) comma separated lists
            for aux in re.split('\s+|\s*,\s*', attributes[ncvar]['coordinates']):
                if aux in variables:
                    variables.discard(aux)
                    for attr in ('bounds', 'climatology'):
                        if attr not in attributes[aux]:
                            continue

                        # Check the dimensionality of the coordinate's
                        # bounds. If it is not right, then it can't be
                        # a bounds variable and so promote to an
                        # independent data variable.
                        bounds = attributes[aux][attr]
                        if bounds in nc.variables:
                            if nc.variables[bounds].ndim == nc.variables[aux].ndim+1:
                                variables.discard(bounds)
                            else:
                                del attributes[aux][attr]

                            break
                        else:
                            print \
"WARNING: Missing bounds variable '%(bounds)s' in %(filename)s" % \
locals()
                            del attributes[aux][attr]
                    #--- End: for
                #--- End: if
            #--- End: for
        #--- End: if
    
        # Remove grid mapping variables
        if 'grid_mapping' in attributes[ncvar]:
            variables.discard(attributes[ncvar]['grid_mapping'])

        # Remove cell measure variables
        if 'cell_measures' in attributes[ncvar]:
            cell_measures = re.split('\s*(\w+):\s*', attributes[ncvar]['cell_measures'])
            for ncvar in cell_measures[2::2]:
                variables.discard(ncvar)
        #--- End: if

    #--- End: for

    # ----------------------------------------------------------------
    # Everything left in the variables set is now a proper data
    # variable, so make a list of fields, each of which contains one
    # data variable and the relevant shared metadata.
    # ----------------------------------------------------------------

    # Dictionary mapping netCDF variable names of space components to
    # their cf Variables.
    #
    # For example:
    # >>> seen_in_file
    # {'lat': <CF Coordinate: (73)>}
    seen_in_file = {}

    # Dictionary of ?
    #
    # For example:
    # >>> formula_terms
    #
    formula_terms = {}

    ancillary_variables = set()
    fields_in_file      = FieldList()

    for data_var in variables:

        f = _create_Field(nc,
                          data_var,
                          attributes,
                          seen_in_file,
                          formula_terms, 
                          ancillary_variables, 
                          global_attributes)
        
        f.file = filename
        
        fields_in_file.append(f)
    #--- End: for

    # ----------------------------------------------------------------
    # If any transform equation terms are fields then find out which
    # fields they are
    # ----------------------------------------------------------------
    if formula_terms:

        # Find the formula terms fields, identified by their netCDF
        # variable names.
        ft_ncvars = []
        for transform in formula_terms.itervalues():
            for value in transform.itervalues():
                try:
                    ft_ncvars.append(value['ncvar'])
                except (TypeError, KeyError):
                    continue
        #--- End: for

        # 
        ncvar_to_field = {}
        index = 0
        while ft_ncvars:
            try:
                f = fields_in_file[index]
            except IndexError:
                break

            # Still here?
            ncvar = f.ncvar            
            if ncvar in ft_ncvars:
                # Remove formula_terms transforms from fields
                # contained within formula_terms transforms
                transforms = f.space.transforms
                for key in transforms.keys():
                    if transforms[key].is_formula_terms:
                        del transforms[key]
                        for k in f.space:
                            if hasattr(f.space[k], 'transform'):
                                del f.space[k].transform
                #--- End: if
                ft_ncvars.remove(ncvar)
                ncvar_to_field[ncvar] = f
                fields_in_file.pop(index)
            else:
                index += 1
        #--- End: while

        for f in fields_in_file:

            # Loop round this field's transforms (if any)
            transforms = f.space.transforms.values()
            for transform in transforms:

                if transform.is_grid_mapping:
                    continue

                # Still here? Then loop round this formula_terms
                # transform's terms and values
                for term, value in transform.iteritems():
                    try:
                        transform[term] = ncvar_to_field[value['ncvar']].copy()
                    except (TypeError, KeyError):
                        continue
                #--- End: for
            #--- End: for
        #--- End: for
    #--- End: if
        
    # ----------------------------------------------------------------
    # For each field that has ancillary variables, replace its list of
    # netCDF variable names with a FieldList.
    # ----------------------------------------------------------------
    if ancillary_variables:
        ncvar_to_field = {}
        index          = 0
        while ancillary_variables:
            try:
                f = fields_in_file[index]
            except IndexError:
                break
            
            # Still here?
            ncvar = f.ncvar            
            if ncvar in ancillary_variables:
                ancillary_variables.discard(ncvar)
                ncvar_to_field[ncvar] = f
                fields_in_file.pop(index)
            else:
                index += 1
        #--- End: while
        
        for f in fields_in_file:
            if not hasattr(f, 'ancillary_variables'):
                continue        
            av_fields = FieldList()
            for ncvar in f.ancillary_variables:
                av_fields.append(ncvar_to_field[ncvar].copy())
            f.ancillary_variables = av_fields
        #--- End: for
    #--- End: if

    # ----------------------------------------------------------------
    # Close the file
    # ----------------------------------------------------------------
    if close:
        nc.close()
    
    return fields_in_file
#--- End: def

def _create_Field(nc,
                  data_var,
                  attributes,
                  seen_in_file,
                  formula_terms, 
                  ancillary_variables,
                  global_attributes):
    '''

Create a field for a given netCDF variable.

:Parameters:

    nc : netCDF4.Dataset
        The entire netCDF file in a `netCDF4.Dataset` instance.

    data_var : str
        The netCDF name of the variable to be turned into a field.

    attributes : dict
        Dictionary of the data variable's netCDF attributes.

    seen_in_file : dict

    formula_terms : dict

    ancillary_variables : set

    global_attributes : dict

:Returns:

    out : Field
        The new field.

'''    
    # Add global attributes to the data variable's properties, unless
    # the data variables already has a property with the same name.
    for attr, value in global_attributes.iteritems():
        if attr not in attributes[data_var]:
            attributes[data_var][attr] = value

    # Take cell_methods out of the data variable's properties since it
    # will need special processing once the space has been defined
    if 'cell_methods' in attributes[data_var]:
        cell_methods = CellMethods(attributes[data_var].pop('cell_methods'))
    else:
        cell_methods = None

    # Take add_offset and scale_factor out of the data variable's
    # properties since they will be dealt with by the variable's Data
    # object
    for attr in ('add_offset', 'scale_factor'):
        if attr in attributes[data_var]:
            del attributes[data_var][attr]

    # Change numpy arrays to tuples for selected attributes
    for attr in ('valid_range',):
        if attr in attributes[data_var]:
            attributes[data_var][attr] = tuple(attributes[data_var][attr])

    # ----------------------------------------------------------------
    # Initialize the field with the data variable and its attributes
    # ----------------------------------------------------------------
    f = Field(attributes[data_var])

    f.ncvar = data_var
    f.space = Space()

    if 'units' not in attributes[data_var]:
        # Units have not been set on this variable so, as per the CF
        # conventions, we can assume that the variable is
        # dimensionless and set the units accordingly.
        f.Units = Units('')

    f.Data = Data(nc.variables[data_var],
                  units      = f.Units,
                  _FillValue = f._FillValue,
                  chunk      = False)
    
    # Initialize mappings of nc variable and dimension names to
    # internal field names. E.g. ncvar_to_var['geo_region']='aux0',
    # ncdim_to_dim['depth']='dim1'
    ncvar_to_var = {}
    ncdim_to_dim = {}
        
    f.space.dimensions['data'] = []
    f.space.nc_dimensions      = {}
    
    # ----------------------------------------------------------------
    # Add dimensions and dimension coordinates to the field
    # ----------------------------------------------------------------
    dimN = 0
#    uncN = 0

    for ncdim in map(str, nc.variables[data_var].dimensions):
        
        # Set the internal field name for this dimension and map the
        # nc name to it
        dim = 'dim%(dimN)d' % locals()
        dimN += 1         # Increment dimension number
        ncvar_to_var[ncdim] = dim
        ncdim_to_dim[ncdim] = dim
        
        # Update data dimension name and set dimension size
        f.space.nc_dimensions[dim]   = ncdim
        f.space.dimension_sizes[dim] = len(nc.dimensions[ncdim])
        f.space.dimensions['data'].append(dim)
        
        # If a dimension coordinate exists then create it in the field
        if ncdim in nc.variables:

            f.space.dimensions[dim] = [dim]
            
            if ncdim in seen_in_file:
                f.space[dim] = seen_in_file[ncdim].copy()
            else:
                coord = _create_Coordinate(nc, ncdim, attributes)
                seen_in_file[ncdim] = coord                
                f.space[dim]        = coord
            #--- End: if

#            # --------------------------------------------------------
#            # Compression by gathering
#            # --------------------------------------------------------
#            try:
#                t.space[dim].compress = CoordinateList(t.space[dim].compress.split())
#            except AttributeError:
#                pass
#            else:
#                # Create the uncompressed coordinates
#                for i in xrange(len(t.space[dim].compress)):
#                    try:
#                        t.space[dim].compress[i] = seen_in_file[ncvar].copy()
#                    except KeyError:
#                        pass
#                        t.space[dim].compress[i] = _create_Coordinate(dim, 
#                                                                     nc, ncvar, attributes)
#                        seen_in_file[ncvar] = t.space[dim].compress[i]
#                #--- End: for
#
#                t.space[dim].compress = t.space[dim].compress.split()
#                for i in xrange(len(t.space[dim].compress)):
#                    # Create the coordinate without updating the
#                    # dimension_sizes or dimensions dictionaries
#                    unc = 'unc'+str(uncN)
#                    ncvar = t.space[dim].compress[i]
#                    t.space[dim].compress[i] = unc
#                    t.space.nc_dimensions[unc] = ncvar
#                    try:
#                        t.space[unc] = seen_in_file[ncvar].copy()
#                    except KeyError:
#                        t.space[unc] = _create_Coordinate(nc, ncvar, attributes)
#                        seen_in_file[ncvar] = t.space[unc] 
#
#                    # Increment ?? number
#                    uncN += 1
#                #--- End: for
#
#        else:
#            # Dimension coordinate for this dimension does not exist
#            increasing[ncdim]       = True
#            t.space.increasing[dim] = True

        #--- End: if

    #--- End: for

    # ----------------------------------------------------------------
    # Add scalar dimension coordinates and auxiliary coordinates to
    # the field
    # ----------------------------------------------------------------
    auxN = 0
    try:
        coordinates = f.getattr('coordinates')
    except AttributeError:
        pass
    else:
        # Allow for (incorrect) comma separated lists
        for ncvar in re.split('\s+|\s*,\s*', coordinates):

            # Skip dimension coordinates which are in the list
            if ncvar in nc.variables[data_var].dimensions:
                continue

            # Set the internal field name for this auxiliary
            # coordinate and map the nc name to it
            aux = 'aux%(auxN)d' % locals()
            auxN += 1       # Increment auxiliary coordinate number
            ncvar_to_var[ncvar] = aux

            # If the auxiliary coordinate variable exists then create
            # a new coordinate variable for it in the field dictionary
            if ncvar in nc.variables:

                # Set auxiliary coordinate dimension names
                f.space.dimensions[aux] = []
                for ncdim in nc.variables[ncvar].dimensions:
                    try:
                        f.space.dimensions[aux].append(ncdim_to_dim[ncdim])
                    except KeyError:
                        # Why try here? because of things like bounds, strlen
                        pass
                #--- End: for

                key = aux

                # ----------------------------------------------------
                # Create a dimension for a numeric scalar coordinate
                # and turn it into a dimension coordinate.
                # ----------------------------------------------------
                if not f.space.dimensions[aux]:
                    dim = 'dim%(dimN)d' % locals()
                    dimN += 1         # Increment dimension number
                    
                    f.space.dimension_sizes[dim] = 1
                    f.space.dimensions[dim]      = [dim]

                    if nc.variables[ncvar].dtype.kind is 'S':
                        # String valued scalar coordinate                        
                        f.space.dimensions['aux'] = [dim]
                    else:  
                        # Numeric valued scalar coordinate                           
                        del f.space.dimensions[aux]
                        key = dim
                #--- End: if

                if ncvar in seen_in_file:
                    f.space[key] = seen_in_file[ncvar].copy()
                else:
                    coord = _create_Coordinate(nc, ncvar, attributes)

                    seen_in_file[ncvar] = coord
                    f.space[key]        = coord
                #--- End: if
            #--- End: if

        #--- End: for

        f.delattr('coordinates')
    #--- End: try

    # ----------------------------------------------------------------
    # Add formula_terms to coordinates' Transform attributes
    # ----------------------------------------------------------------
    transN = 0
    for coord in f.space.keys():
 
        coord_ncvar = f.space[coord].ncvar

        if coord_ncvar in formula_terms:
            # This coordinate's formula_terms transform has already
            # been created, so just copy it. (We have to do this
            # because the 'formula_terms' public attribute on the
            # coordinate has been deleted.)
            transform_id = 'trans%(transN)d' % locals()
            transN += 1       # Increment auxiliary coordinate number
            f.space[coord].transform = transform_id
            f.space.transforms[transform_id] = formula_terms[coord_ncvar].copy()
            continue
        #-- End: if

        # Still here?
        form_terms = f.space[coord].getattr('formula_terms', None)
        if form_terms is None:
            # This coordinate doesn't have a formula_terms attribute
            continue

        # Still here? Then this coordinate has a formula_terms
        # transform that needs creating
        transform      = Transform()
        transform.name = f.space[coord].standard_name
        
        # Add the equation terms and references to their values to to
        # new auxiliary coordinate's transform.
        ft = re.split('\s+|\s*:\s+', form_terms)
        for term, ncvar in zip(ft[0::2], ft[1::2]):
            if ncvar in ncvar_to_var:
                # A coordinate variable from this field
                value = ncvar_to_var[ncvar]
            else:
                # Assume it's another field. We don't know what this
                # field is yet, so we'll give it its ncvar name and
                # work out which field it is later, if we
                # can. E.g. value={'ncvar':'pstar'}
                value = {'ncvar': ncvar}
    
            transform[term] = value
        #--- End: for 

        transform_id = 'trans%(transN)d' % locals()
        transN += 1       # Increment transform number

        f.space[coord].transform = transform_id

        f.space.transforms[transform_id] = transform

        coord_ncvar = f.space[coord].ncvar
        formula_terms[coord_ncvar] = transform

        f.space[coord].delattr('formula_terms')
    #--- End: for

    # ----------------------------------------------------------------
    # Add grid mapping to coordinates' Transform attributes
    # ----------------------------------------------------------------
    grid_mapping = f.getattr('grid_mapping', None)

    if grid_mapping is not None:

        for ncvar in grid_mapping.split():

            if ncvar not in attributes:
                continue

            if 'grid_mapping_name' not in attributes[ncvar]:
                attributes[ncvar]['grid_mapping_name'] = ncvar

            grid_mapping_name = attributes[ncvar]['grid_mapping_name']

            transform_id = 'trans%(transN)d' % locals()
            transN += 1     

            f.space.transforms[transform_id] = Transform(**attributes[ncvar])
            f.space.transforms[transform_id].ncvar = ncvar
            f.space.transforms[transform_id].name  = grid_mapping_name

            # Work out which coordinates need the transform attached
            # to them
            for key in f.space.get_keys('^dim|^aux'):
                coord = f.space[key]
                coord_name = coord.name()

                if coord_name in ('longitude', 'latitude'):
                    coord.transform = transform_id
                    continue
                #--- End: if

                if grid_mapping_name == 'rotated_latitude_longitude':
                    if coord_name in ('grid_longitude', 'grid_latitude'):
                        coord.transform = transform_id
                elif coord_name in ('projection_x_coordinate', 
                                    'projection_y_coordinate'):
                    coord.transform = transform_id
            #--- End: for
        #--- End: for

        f.delattr('grid_mapping')
    #--- End: try

    # ----------------------------------------------------------------
    # Add cell measures to the field
    # ----------------------------------------------------------------
    cell_measures = f.getattr('cell_measures', None)

    if cell_measures is not None:

        # Parse the cell measures attribute
        cell_measures = re.split('\s*(\w+):\s*', cell_measures)

        cmN = 0
        for measure, ncvar in zip(cell_measures[1::2], 
                                  cell_measures[2::2]):

            if ncvar not in attributes:
                continue

            cm = 'cm%(cmN)d' % locals()
            cmN += 1                # Increment cell measures number

            # Set cell measures' dimension names
            f.space.dimensions[cm] = [ncdim_to_dim[ncdim]
                                      for ncdim in nc.variables[ncvar].dimensions]

            # Still here? Then put the cell measure into the space.
            if ncvar in seen_in_file:
                # Copy the cell measure as it already exists
                f.space[cm] = seen_in_file[ncvar].copy()
            else:
                cell = _create_CellMeasures(nc, ncvar, attributes)
                
                cell.measure = measure
                    
                seen_in_file[ncvar] = cell
                f.space[cm]         = cell
            #--- End: if
        #--- End: for

        f.delattr('cell_measures')
    #--- End: if

    # -----------------------------
    # Add cell methods to the field
    # -----------------------------
    if cell_methods is not None:
        f.cell_methods = cell_methods.netCDF_translation(f)

    # ----------------------------------------------------------------
    # Parse ancillary variables to a list of netCDF variable names,
    # which will get converted to a FieldList later.
    # ----------------------------------------------------------------
    if hasattr(f, 'ancillary_variables'):
        f.ancillary_variables = f.ancillary_variables.split()
        ancillary_variables.union(set(f.ancillary_variables))

    # sdfsdfsdfsdfsdf
    f.finalize()

    # Return the finished field
    return f
#--- End: def

def _create_Coordinate(nc, ncvar, attributes):
    '''

Create a coordinate variable, including any bounds.

:Parameters:

    nc : netCDF4.Dataset
        The entire netCDF file in a `netCDF4.Dataset` instance.

    ncvar : str
        The netCDF name of the coordinate variable.

    attributes : dict
        Dictionary of the coordinate variable's netCDF attributes.

:Returns:

    out : Coordinate
        The new coordinate.

'''
    c = Coordinate(attributes[ncvar])

    c.ncvar = ncvar
    
    data = nc.variables[ncvar]

    # ------------------------------------------------------------
    # Collapse a string valued auxiliary
    # coordinate. E.g. ['a','b','c'] becomes ['abc']
    # ------------------------------------------------------------
    if data.dtype.kind is 'S':
        new_shape = data.shape[0:-1]
        new_size  = long(reduce(mul, new_shape, 1))
        strlen    = data.shape[-1]

        data = numpy.ma.resize(data[...], (new_size, strlen))
        
        # Make sure everything is ok if data is a masked array
        try:
            data.set_fill_value(' ')
            data = data.filled()
        except AttributeError:
            pass

        temp = [''.join(x).rstrip() for x in data]
        data = numpy.ma.array(temp, dtype='|S%(strlen)d' % locals())

        data = numpy.ma.resize(data, new_shape)

        del temp
    #--- End: if

    if 'units' not in attributes[ncvar]:
        # Units have not been set on this variable so, as per the CF
        # conventions, we can assume that the variable is
        # dimensionless and set the units accordingly.
        c.Units = Units('')

    c.Data = Data(data,
                  units      = c.Units,
                  _FillValue = c._FillValue,
                  chunk      = False)
    
    # ------------------------------------------------------------
    # Add any bounds
    # ------------------------------------------------------------
    if hasattr(c, 'climatology'):
        c.bounds      = c.climatology
        c.climatology = True
    #--- End: if

    if hasattr(c, 'bounds'):
        ncbounds = c.bounds

        b = CoordinateBounds(attributes[ncbounds])

        b.ncvar = ncbounds
        
        if 'units' in attributes[ncbounds]:
            units = Units(attributes[ncbounds]['units'])
            if not units.equivalent(c.Units):
                print \
"WARNING: Overriding incompatible bounds units: %s (%s)" % \
(repr(units), repr(c.Units))
                units = c.Units
        else:
            units = c.Units

        b.Data = Data(nc.variables[ncbounds],
                      units      = units,
                      _FillValue = b._FillValue,
                      chunk      = False)

        # Ignore formula_terms attributes on coordinate bounds
        # variables
        if b.hasattr('formula_terms'):
            b.delattr('formula_terms')

        c.bounds = b
    #--- End: if
    
    # ---------------------------------------------------------
    # Return the coordinate
    # ---------------------------------------------------------
    return c
#--- End: def

def _create_CellMeasures(nc, ncvar, attributes):
    '''

Create a cell measures variable.

:Parameters:

    nc : netCDF4.Dataset
        The entire netCDF file in a `netCDF4.Dataset` instance.

    ncvar : str
        The netCDF name of the cell measures variable.

    attributes : dict
        Dictionary of the cell measures variable's netCDF attributes.

:Returns:

    out : CellMeasures
        The new cell measures.

'''
    cm       = CellMeasures(attributes[ncvar])
    cm.ncvar = ncvar

    if 'units' not in attributes[ncvar]:
        # Units have not been set on this variable so, as per the CF
        # conventions, we can assume that the variable is
        # dimensionless and set the units accordingly.
        cm.Units = Units('')
       
    cm.Data = Data(nc.variables[ncvar],
                   units      = cm.Units,
                   _FillValue = cm._FillValue,
                   chunk      = False)

    return cm
#--- End: def
