import numpy
from copy import copy
from .utils     import CfList, CHUNKSIZE, FM_THRESHOLD
from .partition import Partition

_meminfo_fields = ['SReclaimable:', 'Cached:', 'Buffers:', 'MemFree:']

def free_memory():
    '''

The amount of available physical memory in kibibytes.

This amount includes any memory which is still allocated but is no
longer required.

:Returns:

    out : float
        The amount of available physical memory in kibibytes.

.. warning::
   
   This has only been tested on Linux, and will need other cases for
   Unix, Mac, Windows, etc.

'''
    # ----------------------------------------------------------------
    #
    # LINUX: The available physical memory is the sum of the values of
    #        the 'SReclaimable', 'Cached', 'Buffers' and 'MemFree'
    #        entries in the /proc/meminfo file.
    #
    # ----------------------------------------------------------------
    meminfo_fields = _meminfo_fields[:]

    free_kB = 0.0

    meminfo_file = open('/proc/meminfo', 'r', 1)
    for line in meminfo_file:
        field, size, kB = line.split()
        if field == meminfo_fields[-1]:
            free_kB += float(size)
            meminfo_fields.pop()
            if not meminfo_fields:
                break
    #--- End: for
    meminfo_file.close()

    return free_kB
#--- End: def


# ====================================================================
#
# PartitionArray object
#
# ====================================================================

class PartitionArray(CfList):
    '''

An N-dimensional partition array.

'''

    def __init__(self, array, **kwargs):
        '''

**Initialization**

:Parameters:

    array : list 
        A list of Partition objects.

    kwargs :
        Set attributes specified by the keywords and their
        values. Values are not deep copied.

**Examples**

>>> pa = PartitionArray(
        [Partition(location  = [(0,n) for n in shape],
                   shape     = shape[:],
                   order     = order[:],
                   direction = copy(direction),
                   Units     = units.copy(),
                   part      = [],
                   data      = data)
         ],
        order      = order,
        shape      = shape,
        size       = data.size,
        Units      = units.copy(),
        ndim       = ndim,
        direction  = direction,
        adims      = [],
        _FillValue = _FillValue)

'''
        super(PartitionArray, self).__init__()

        self._list = array

        for attr, value in kwargs.iteritems():
            setattr(self, attr, value)
    #--- End: def
    
    def __repr__(self):
        '''
x.__repr__() <==> repr(x)

'''
        return '<CF %s: %d aggregating dimensions>' % (self.__class__.__name__, 
                                                       len(self.adims))
    #--- End: def

    def __nonzero__(self):
        '''
x.__nonzero__() <==> bool(x)

'''
        return True
    #--- End: def

    def __getitem__(self, index):
        '''
x.__getitem__(index) <==> x[index]

'''
        out = self._list[index]
        if isinstance(out, Partition):
            return out

        # out is a list so return a PartitionArray
        return type(self)(out,                         
                          order      = self.order,   
                          direction  = self.direction,
                          ndim       = self.ndim,
                          size       = self.size,
                          _FillValue = self._FillValue,
                          Units      = self.Units,    
                          adims      = self.adims,    
                          shape      = self.shape,                         
                          )
    #--- End: def

    def copy(self, _copy_attr=True):
        '''

Return a deep copy.

Do not set the `_copy_attr` parameter. It is for internal use only.

Equivalent to ``copy.deepcopy(pa)``.

:Returns:

    out :
        The deep copy.

**Examples**

>>> pa.copy()

''' 
        new = type(self)([])
        
        if _copy_attr:
            # Only copy the attributes once.
            new.adims      = self.adims[:]
            new.order      = self.order[:] 
            new.shape      = self.shape[:]
            new.direction  = copy(self.direction)
            new.Units      = self.Units.copy()
            new.ndim       = self.ndim
            new.size       = self.size
            new._FillValue = self._FillValue
        #--- End: if

        if not self:
            return new

        if isinstance(self[0], Partition):
            for partition in self:
                new.append(partition.copy())
        else:
            for nested_array in self:
                # Recursive call
                om = nested_array.copy(_copy_attr=False)
                new.append(om)
        #--- End: if

        return new
    #--- End: def

    def info(self, attr):
        '''

'''
        if not self:
            return []

        out = []
        if isinstance(self[0], Partition):
            for partition in self:
                out.append(getattr(partition, attr, None))
        else:
            for nested_partition in self:
                # Recursive call
                out.append(nested_partition.info(attr))
        #--- End: if

        print out
    #--- End: def

    def __str__(self):
        '''
x.__str__() <==> str(x)

'''
        return str(self.info('location'))
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: dtype
    # ----------------------------------------------------------------
    @property
    def dtype(self):
        '''

Numpy data-type of the master array.

When there is more than one partition in the partition array, the
normal data-type coercion rules apply. For example, if the partitions
have data-types 'int32' and 'float32' then the master array's
data-type will be 'float32'.

**Examples**

>>> pa.dtype
dtype('float64')

'''
        return max((partition.data.dtype for partition in self.flat()))
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: is_scalar (can't set or delete)
    # ----------------------------------------------------------------
    @property
    def is_scalar(self):
        '''

True if the master array is a 0-d scalar array.

**Examples**

>>> pa.ndim
0
>>> pa.is_scalar
True

>>> pa.ndim >= 1
True
>>> pa.is_scalar
False

'''
        return not self.ndim
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: npartitions (can't set or delete)
    # ----------------------------------------------------------------
    @property
    def npartitions(self):
        '''

The number of partitions in the partition array.

Logically equivalent to ``len(pa.ravel())``.

**Examples**

>>> pa.npartitions
7

'''
        for i in enumerate(self.flat()):
            pass

        return i+1
    #--- End: def

    def flat(self):
        '''

Return a flat iterator over the Partition objects in the partition
array.

:Returns:

    out : generator
        An iterator over the elements of the partition array.

**Examples**

>>> type(pa.flat())
<generator object flat at 0x145a0f0>
>>> for partition in pa.flat():
...     print partition.Units

'''
        if not self:
            return

        if isinstance(self[0], Partition):
            for partition in self:
                yield partition
        else:        
            for nested_partitions in self:
                for x in nested_partitions.flat():
                    yield x
    #--- End: def

    def ravel(self):
        '''

Return a flattened partition array as a built-in list.

:Returns:

    out : list of Partitions
        A list containing the flattened partition array.

**Examples**

>>> x = partitions.ravel()
>>> type(x)
list

'''
        if not self:
            return []

        out = []
        if isinstance(self[0], Partition):
            out.extend(self)
        else:        
            for nested_partitions in self:
                out.append(nested_partitions.ravel())
        #--- End: if

        return out
    #--- End: def


    def expand_dims(self, adim):
        '''

'''
        if not self.adims:
            self.adims = [adim]
        else:
            self._list = [self._list]
            self.adims.insert(0, adim)        
    #--- End: def
      
    def change_dimension_names(self, dim_name_map):
        '''

dim_name_map should be a dictionary which maps each dimension names in
self.order to its new dimension
name. E.g. {'dim0':'dim1', 'dim1':'dim0'}

'''
        # Check for a null dimension name mapping
        # (e.g. 'dim0'->'dim0', etc.) and return if there is nothing
        # to change. I think that this is worth it, on the grounds
        # that this is often be the case.
        map_is_null = True
        for dim0, dim1 in dim_name_map.iteritems():
            if dim0 != dim1:
                map_is_null = False
                break
        #--- End: for       
        if map_is_null:
            return
            
        # Still here? Then some dimension names need changing.

        # Dimension order
        self.order = [dim_name_map[dim] for dim in self.order]
        
        # Aggregating dimensions
        self.adims = [dim_name_map[dim] for dim in self.adims]

        # Dimension directions
        if not self.is_scalar:
            self.direction = dict(
                ((dim_name_map[dim], value)
                 for dim, value in self.direction.iteritems())
                )

        # Partitions (Note that a partition may have dimensions which
        # are not in self.order and that these must also be in
        # dim_name_map).
        for partition in self.flat():

            # Partition data dimension order
            partition.order = [dim_name_map[dim] for dim in partition.order]
            
            # Partition data dimension directions
            direction = partition.direction
            if not partition.is_scalar:
                partition.direction = dict(
                    ((dim_name_map[dim], value)
                     for dim, value in direction.iteritems())
                    )
        #--- End: for
    #--- End: def

    def equals(self, *args, **kwargs):
        # No doc string
        raise NotImplementedError("%s has no 'equals' method" %
                                  self.__class__.__name__)
    #--- End: def

    def save_to_disk(self, itemsize=None):
        '''

:Parameters:

    itemsize : int, optional

:Returns:

    out : bool

'''
        if not itemsize:
            if hasattr(self, 'dtype'):
                itemsize = self.dtype.itemsize
            else:
                raise TypeError, "Must set itemsize if there is no dtype"
        #--- End: if

        if self.size > CHUNKSIZE()/(itemsize + 1):
            # The size of the data array (including a mask) is greater
            # than the chunk size, so return True
            return True

        if free_memory() < FM_THRESHOLD():
            # The available memory is less than the minimum amount
            # that we want to try to keep free, so return True
            return True

        # Still here? Then this array is small enough to remain in
        # memory
        return
    #--- End: def

    def to_disk(self):
        '''

Store each partition's data on disk in place.

There is no change to partitions with data that are already on disk.

:Returns:

    None

**Examples**

>>> pa.to_disk()

'''
        for partition in self.flat():
            if partition.in_memory:
                partition.to_disk()
    #--- End: def

    def to_memory(self, regardless=False):
        '''

Store each partition's data in memory in place if the master array is
smaller than the chunk size.

There is no change to partitions with data that are already in memory.

:Parameters:
    
    regardless : bool, optional
        If True then store all partitions' data in memory regardless
        of the size of the master array. By default only store all
        partitions' data in memory if the master array is smaller than
        the chunk size.

:Returns:

    None

**Examples**

>>> pa.to_memory()
>>> pa.to_memory(True)

'''
        if not self.save_to_disk() or regardless:
            order     = self.order
            direction = self.direction
            units     = self.Units   
            
            save           = False
            revert_to_file = False
            
            for partition in self.flat():
                if partition.on_disk:
                    partition.conform(order, direction, units,
                                      save, revert_to_file)
                    # Note: No need to close the partition
    #--- End: def

    def add_partitions(self,
                       extra_boundaries, 
                       adim,
                       existing_boundaries=None):
        '''

'''     

        def _add_partitions(partitions, existing_boundaries, extra_boundaries, 
                            adims, adim, index, direction, dim2index):
            '''
            '''
            def _update_p(partitions, location, index, 
                          part, dim2index, direction):
                '''
'''                
                if isinstance(partitions, Partition):
                    partitions = [partitions]                        
                    
                elif not isinstance(partitions[0], Partition):
                    for nested_partitions in partitions:
                        # Recursive call
                        _update_p(nested_partitions,
                                  location, index,
                                  part, dim2index, direction)
                    #--- End: for
                    return
                #--- End: if

                for partition in partitions:
                    partition.location[index] = location
                    partition.shape[index]    = shape                       
                    partition.part = partition.new_part(part, 
                                                        dim2index, 
                                                        direction)
                #--- End: for
            #--- End: def

            full = [slice(None)] * self.ndim

            n_adim = len(adims)

            if adim == adims[0]:

                new_list = []
                extra_boundaries = extra_boundaries[:]
                x = extra_boundaries.pop(0)

                for p, r0, r1 in zip(partitions,
                                     existing_boundaries[:-1], 
                                     existing_boundaries[1:]):

                    if not r0 < x < r1:
                        if n_adim == 1:
                            # p is a partition
                            new_list.append(p)
                        else:
                            # p is a partition array
                            new_list.append(p._list)

                        continue
                    #--- End: if

                    # Still here?

                    # Find the new extent of the original partition(s)
                    location    = (r0, x)
                    shape       = x - r0
                    part        = full[:]
                    part[index] = slice(0, shape)

                    # Create new partition(s) in place of the original
                    # ones(s) and set the location, shape and part
                    # attributes
                    new_p = p.copy()
                    _update_p(new_p,
                              location, index,
                              part, dim2index, direction)

                    # Append the new partition(s) to the new list
                    if n_adim == 1:
                        new_list.append(new_p)
                    else:
                        new_list.append(new_p._list)

                    while x < r1:

                        # Find the extent of the new partition(s)
                        if not extra_boundaries:
                            # No more new boundaries, so the new
                            # partition(s) run to the end of the
                            # original partition(s) in which they lie.
                            location = (x, r1)
                        else:
                            # There are more new boundaries, so this
                            # new partition runs either to the next
                            # new boundary or to the end of the
                            # original partition, which comes first.
                            location = (x, min(extra_boundaries[0], r1))

                        shape       = location[1] - location[0]
                        offset      = x - r0
                        part        = full[:]
                        part[index] = slice(offset, offset+shape)

                        # Create the new partition(s) and set the
                        # location, shape and part attributes
                        new_p = p.copy()
                        _update_p(new_p,
                                  location, index,
                                  part, dim2index, direction)
                        
                        # Append the new partition(s) to the new list
                        if n_adim == 1:
                            new_list.append(new_p)
                        else:
                            new_list.append(new_p._list)

                        # Move on to the next new boundary, if there
                        # is one
                        if not extra_boundaries:
                            break

                        x = extra_boundaries.pop(0)                        
                    #--- End: while                        
                #--- End: for 
                        
                # Update partitions array in place
                partitions[:] = new_list
                
            else:
                adims = adims[1:]
                if not adims:
                    return
                for nested_partitions in partitions:
                    # Recursive call
                    _add_partitions(nested_partitions,
                                    existing_boundaries, extra_boundaries,
                                    adims, adim, index,
                                    direction, dim2index)
            #--- End: for
        #--- End: def
                    
        # If no extra boundaries have been provided, just return
        # without doing anything
        if not extra_boundaries:
            return

        # Find the existing boundaries if they haven't been provided
        if existing_boundaries is None:
            existing_boundaries = self.partition_boundaries()[adim]
            
        dim2index = dict([(dim, i) for i, dim in enumerate(self.order)])

        _add_partitions(self, existing_boundaries, extra_boundaries, 
                        self.adims, adim, self.order.index(adim), 
                        self.direction,
                        dim2index)        
    #--- End: def
  
    def partition_boundaries(self):
        '''

'''            
        boundaries = {}

        zeros = [0] * len(self.adims)

        for j, adim in enumerate(self.adims):

            first = self

            indices    = zeros[:]
            indices[j] = True

            while indices:
                index = indices.pop(0)
                if not index:
                    first = first[index]      # index is 0
                else:
                    break
            #--- End: while

            temp = []
            for element in first:
                for index in indices:
                    element = element[index]  # index is 0
                temp.append(element)
            #--- End: for
            first = temp
            # 'first' should now be list of Partition objects

            i = self.order.index(adim)

            b = [partition.location[i][1] for partition in first]
            b.insert(0, 0)

            boundaries[adim] = b
        #--- End: for

        return boundaries
    #--- End: def

    def set_location_map(self):
        '''

Recalculate the `location` attribute of each Partition object in the
partition array in place.

**Examples**

>>> pa.set_location_map()

'''
        def _set_location_map(partitions, starts, stops, adims, 
                              index, level):
            '''
    '''
            if isinstance(partitions[0], Partition):
                if not adims:
                    inner_adim = None
                else:
                    inner_adim = adims[-1]

                for dim, start in starts.iteritems():
                    i = index[dim]
                    if dim == inner_adim:
                        for partition in partitions:
                            stop                  = start + partition.shape[i]
                            partition.location[i] = (start, stop)
                            start                 = stop
                        #--- End: for
                        stops[dim] = 0
                    else:
                        stop     = start + partitions[0].shape[i]
                        location = (start, stop)
                        for partition in partitions:
                            partition.location[i] = location

                        if dim in adims:
                            stops[dim] = stop
                #--- End: for
            else:
                level += 1
                for nested_partitions in partitions:
                    starts[adims[level]] = stops[adims[level]]
                    for dim in adims[level+1:]:
                        starts[dim] = 0
                        stops[dim]  = 0
                    #--- End: for

                    # Recursive call
                    _set_location_map(nested_partitions,
                                      starts,
                                      stops,
                                      adims,
                                      index,
                                      level)
                #--- End: for
            #--- End: if
        #--- End: def

        order = self.order

        starts = dict((dim, 0) for dim in order)
        stops  = starts.copy()

        index  = dict((dim, i) for i, dim in enumerate(order))

        level  = -1

        _set_location_map(self, starts, stops, self.adims, index, level)
    #--- End: def

    def rollaxis(self, axis, start=0):
        '''

Roll the specified aggregating dimension backwards,in place until it
lies in a given position.

This does not change the master array.

:Parameters:

    axis : int 
        The axis to roll backwards. The positions of the other axes do
        not change relative to one another.

    start : int, optional 
        The axis is rolled until it lies before this position.  The
        default, 0, results in a "complete" roll.

:Returns:

    None

**Examples**

>>> pa.rollaxis(2)
>>> pa.rollaxis(2, start=1)

'''
        if axis != start:
            self._list = numpy.rollaxis(numpy.array(self._list), 
                                        axis, start=start
                                        ).tolist()
            self.adims.insert(start, self.adims.pop(axis))
    #--- End: def
             
    def transpose(self, axes):
        '''

Permute the aggregating dimensions of the partition array in-place.

This does not change the master array.

:Parameters:

    axes : sequence of ints 
        Permute the axes according to the values given.

:Returns:

    None

**Examples**

>>> pa.transpose((2,0,1))

'''
        adims = self.adims
        if axes != range(len(adims)):
            self._list = numpy.transpose(numpy.array(self._list), 
                                         axes=axes
                                         ).tolist()
            self.adims = [adims[i] for i in axes]
    #--- End: def
             
#--- End: class
