import os
import numpy
import netCDF4
import csv
import re
import textwrap
import time
from .                 import __version__, __Conventions__, __file__
from .space            import Space
from .transform        import Transform
from .field            import Field
from .fieldlist        import FieldList
from .cellmethods      import CellMethods
from .coordinate       import Coordinate
from .coordinatebounds import CoordinateBounds
from .utils            import (RTOL, ATOL, equals, 
                               parse_indices, subset_array)
from .data             import Data
from .units            import Units

# Number of bytes per word in the input PP files
_word_size = 4
_pp_rmdi   = -1e+30 

# --------------------------------------------------------------------
# Set default UM version from the environment
# --------------------------------------------------------------------
try:
    _UMVERSION = os.environ['UMVERSION']
    _UMVERSION = _UMVERSION.replace('.', '0', 1)
except KeyError:
    pass

# --------------------------------------------------------------------
# Constants for converting creating unrotated lat/lon auxiliary
# coordinates
# --------------------------------------------------------------------
_PI          = numpy.pi
_PI_OVER_180 = _PI/180.0

# --------------------------------------------------------------------
# Number matching regular expression
# --------------------------------------------------------------------
_number_regex = '([-+]?\d*\.?\d+(e[-+]?\d+)?)'

# --------------------------------------------------------------------
# Datetime object that copes with non-standard calendars
# --------------------------------------------------------------------
datetime = netCDF4.netcdftime.datetime

# --------------------------------------------------------------------
# Cache for derived values
# --------------------------------------------------------------------
_cache = {'runid' : {},
          'latlon': {},
          'time'  : {},
          }

# --------------------------------------------------------------------
# Names of PP integer and real header items
# --------------------------------------------------------------------
_lhd_items = ('lbyr', 'lbmon', 'lbdat', 'lbhr', 'lbmin', 'lbday',
              'lbyrd', 'lbmond', 'lbdatd', 'lbhrd', 'lbmind',
              'lbdayd', 'lbtim', 'lbft', 'lblrec', 'lbcode', 'lbhem',
              'lbrow', 'lbnpt', 'lbext', 'lbpack', 'lbrel', 'lbfc',
              'lbcfc', 'lbproc', 'lbvc', 'lbrvc', 'lbexp', 'lbegin', 
              'lbnrec', 'lbproj', 'lbtyp', 'lblev', 'lbrsvd1',
              'lbrsvd2', 'lbrsvd3', 'lbrsvd4', 'lbsrce', 'lbuser1',
              'lbuser2', 'lbuser3', 'lbuser4', 'lbuser5', 'lbuser6',
              'lbuser7')

_bhd_items = ('brsvd1', 'brsvd2', 'brsvd3', 'brsvd4', 
              'bdatum', 'bacc', 'blev', 'brlev', 'bhlev', 'bhrlev',
              'bplat', 'bplon', 'bgor',
              'bzy', 'bdy', 'bzx', 'bdx', 'bmdi', 'bmks')

# --------------------------------------------------------------------
# Positions of PP header items in their arrays
# --------------------------------------------------------------------
(lbyr, lbmon, lbdat, lbhr, lbmin, lbday,
 lbyrd, lbmond, lbdatd, lbhrd, lbmind,
 lbdayd, lbtim, lbft, lblrec, lbcode, lbhem,
 lbrow, lbnpt, lbext, lbpack, lbrel, lbfc,
 lbcfc, lbproc, lbvc, lbrvc, lbexp, lbegin, 
 lbnrec, lbproj, lbtyp, lblev, lbrsvd1,
 lbrsvd2, lbrsvd3, lbrsvd4, lbsrce, lbuser1,
 lbuser2, lbuser3, lbuser4, lbuser5, lbuser6,
 lbuser7
 ) = range(45)

(brsvd1, brsvd2, brsvd3, brsvd4, 
 bdatum, bacc, blev, brlev, bhlev, bhrlev,
 bplat, bplon, bgor,
 bzy, bdy, bzx, bdx, bmdi, bmks
 ) = range(19)

# --------------------------------------------------------------------
# Assign CF standard name attributes to PP axis codes. (The full list
# of field code keys may be found at
# http://cms.ncas.ac.uk/html_umdocs/wave/@header.)
# --------------------------------------------------------------------
_coord_standard_name = {
    1   : 'air_pressure', 
    2   : 'altitude',
    3   : 'atmosphere_hybrid_sigma_pressure_coordinate',
    4   : 'depth',
    5   : 'model_level_number',        
    6   : 'air_potential_temperature',
    7   : 'atmosphere_sigma_coordinate',
    10  : 'latitude',
    11  : 'longitude',
    13  : 'region',
    14  : 'atmosphere_hybrid_height_coordinate',
    15  : 'height',
    20  : 'time',      # time (gregorian)
    23  : 'time',      # time (360_day)
    40  : 'pseudolevel',        # pseudolevel. THIS IS NOT A STANDARD NAME!!    #dch should this be None?
    }

# --------------------------------------------------------------------
# Assign CF long names attributes to PP axis codes.
# --------------------------------------------------------------------
_coord_long_name = {
    40 : 'pseudolevel',
    }

# --------------------------------------------------------------------
# Assign CF units attributes to PP axis codes.
# --------------------------------------------------------------------
_coord_units = {
    1   : 'hPa',           # air_pressure                      
    2   : 'm',             # altitude         
    3   : '1',             # atmosphere_hybrid_sigma_pressure_coordinate
    4   : 'm',             # depth                                  
    5   : '1',             # model_level_number                         
    6   : 'K',             # air_potential_temperature
    7   : '1',             # atmosphere_sigma_coordinate               
    10  : 'degrees_north', # latitude                               
    11  : 'degrees_east',  # longitude                                
    13  : 'string',        # region                                     
    14  : '1',             # atmosphere_hybrid_height_coordinate          
    15  : 'm',             # height                                      
    20  : 'days',          # time (gregorian)                    
    23  : 'days',          # time (360_day)
    40  : '1',             # pseudolevel
    }

# --------------------------------------------------------------------
# Assign CF axis attributes to PP axis codes.
# --------------------------------------------------------------------
_coord_axis = {
    1   : 'Z',   # air_pressure                       
    2   : 'Z',   # altitude                                     
    3   : 'Z',   # atmosphere_hybrid_sigma_pressure_coordinate  
    4   : 'Z',   # depth                                        
    5   : 'Z',   # model_level_number                          
    6   : 'Z',   # air_potential_temperature
    7   : 'Z',   # atmosphere_sigma_coordinate                
    10  : 'Y',   # latitude                                     
    11  : 'X',   # longitude                                    
    13  : None,  # region                                       
    14  : 'Z',   # atmosphere_hybrid_height_coordinate          
    15  : 'Z',   # height                                       
    20  : 'T',   # time (gregorian)                                         
    23  : 'T',   # time (360_day)                                         
    40  : None,  # pseudolevel                                    
    }

# --------------------------------------------------------------------
# Assign CF positive attributes to PP axis codes.
# --------------------------------------------------------------------
_coord_positive = {
    1   : 'down',  # air_pressure                     
    2   : 'up',    # altitude                                  
    3   : 'down',  # atmosphere_hybrid_sigma_pressure_coordinate 
    4   : 'down',  # depth                                     
    5   : None,    # model_level_number                         
    6   : 'up',    # air_potential_temperature
    7   : 'down',  # atmosphere_sigma_coordinate               
    10  : None,    # latitude                                   
    11  : None,    # longitude                                   
    13  : None,    # region                                     
    14  : 'up',    # atmosphere_hybrid_height_coordinate         
    15  : 'up',    # height                                      
    20  : None,    # time (gregorian)                                          
    23  : None,    # time (360_day)                                        
    40  : None,    # pseudolevel                                    
    }

# --------------------------------------------------------------------
# Translate LBVC codes to PP axis codes. (The full list of field code
# keys may be found at
# http://cms.ncas.ac.uk/html_umdocs/wave/@fcodes.)
# --------------------------------------------------------------------
_lbvc_codes = {
    1   :  2,   # altitude (Height) 
    2   :  4,   # depth (Depth)
    3   : None, # (Geopotential (= g*height))
    4   : None, # (ICAO height)
    6   :  5,   # model_level_number  
    7   : None, # (Exner pressure)
    8   :  1,   # air_pressure  (Pressure)
    9   :  3,   # atmosphere_hybrid_sigma_pressure_coordinate (Hybrid pressure)
    10  :  7,   # atmosphere_sigma_coordinate (Sigma (= p/surface p))   ## dch check
    16  : None, # (Temperature T)
    19  :  6,   # air_potential_temperature (Potential temperature)
    27  : None, # (Atmospheric) density
    28  : None, # (d(p*)/dt .  p* = surface pressure)
    44  : None, # (Time in seconds)
    65  : 14,   # atmosphere_hybrid_height_coordinate (Hybrid height)
    129 : None, # Surface
    176 : 10,   # latitude    (Latitude)
    177 : 11,   # longitude   (Longitude)
    }

# --------------------------------------------------------------------
# Characters used in decoding LBEXP into a runid
# --------------------------------------------------------------------
_characters = ('a','b','c','d','e','f','g','h','i','j','k','l','m',
               'n','o','p','q','r','s','t','u','v','w','x','y','z',               
               '0','1','2','3','4','5','6','7','8','9'
               )   

# --------------------------------------------------------------------
# Names of PP extra data codes
# --------------------------------------------------------------------
_extra_data_name = {
    1  : 'x',
    2  : 'y',
    3  : 'y_domain_lower_bound',
    4  : 'x_domain_lower_bound',
    5  : 'y_domain_upper_bound',
    6  : 'x_domain_upper_bound',
    7  : 'z_domain_lower_bound',
    8  : 'x_domain_upper_bound',
    10 : 'title',
    11 : 'domain_title',
    12 : 'x_lower_bound',
    13 : 'x_upper_bound',
    14 : 'y_lower_bound',
    15 : 'y_upper_bound',
    }

# ====================================================================
#
# PPFileArray object
#
# ====================================================================
class PPFileArray(object):
    ''' 
    
'''
    __slots__ = ('_binary_mask',
                 '_file'       ,
                 '_file_offset',
                 'dtype'       ,
                 'shape'       ,
                 'size'        ,
                 'ndim'        ,
                 '_FillValue'  ,     
                 'scale_factor',
                 'add_offset'  ,
                 )


    def __init__(self, **kwargs):
        '''
Set attributes from keyword arguments.
'''
        for attr, value in kwargs.iteritems():
            setattr(self, attr, value)
    #--- End: def
 
    def __getitem__(self, indices):
        '''
x.__getitem__(indices) <==> x[indices]
''' 
        # ------------------------------------------------------------
        # Read the array from the PP file
        # ------------------------------------------------------------
        array = numpy.memmap(self._file,
                             mode   = 'r',
                             offset = self._file_offset,
                             dtype  = self.dtype,
                             shape  = self.shape)

        indices = parse_indices(array, indices)

        array = subset_array(array, indices)

        # ------------------------------------------------------------
        # Convert to a masked array
        # ------------------------------------------------------------
        if hasattr(self, '_FillValue'):
            # _FillValue is set so mask any missing values
            _FillValue = self._FillValue
            array = numpy.ma.array(array, 
                                   mask=(array == _FillValue),
                                   fill_value=_FillValue)
            array.shrink_mask()
        else:
            # _FillValue is not set
            array = numpy.ma.asarray(array)

        # ------------------------------------------------------------
        # Unpack the array using the scale_factor and add_offset, if
        # either is available
        # ------------------------------------------------------------
        try:
            if self.scale_factor != 1.0:
                array *= self.scale_factor
        except AttributeError:
            pass
        
        try:
            if self.add_offset != 0.0:
                array += self.add_offset
        except AttributeError:
            pass
    
        if self._binary_mask:
            array = numpy.ma.where(array!=0, 1, 0)

        # ------------------------------------------------------------
        # Return the array
        # ------------------------------------------------------------
        return array
    #--- End: def

#--- End: class

# ====================================================================
#
# PPFileArrayBounds object
#
# ====================================================================

class PPFileArrayBounds(object):
    '''  
'''
    __slots__ = ('_lower'    ,
                 '_upper'    ,
                 'dtype'     ,
                 'shape'     ,
                 'size'      ,
                 'ndim'      ,
                 )

    def __init__(self, _lower, _upper):
        '''
'''
        self._lower = _lower
        self._upper = _upper

        self.dtype = max(self._lower.dtype, self._upper.dtype)
        self.shape = self._lower.shape + (2,)       
        self.size  = self._lower.size * 2
        self.ndim  = self._lower.ndim + 1
    #--- End: def
   
    def __getitem__(self, indices):
        '''

x.__getitem__(indices) <==> x[indices]
'''
        # ------------------------------------------------------------
        # Read the upper and lower bounds from the PP file and stick
        # them together
        # ------------------------------------------------------------
        array = numpy.column_stack((self._lower[...],
                                    self._upper[...],))

        indices = parse_indices(array, indices)

        return subset_array(array, indices)
    #--- End: def

#--- End: class


class PPMetadata(object):

    def __init__(self):
        '''

Set attributes giving the default data type for reals and
integers. The basic assumption is that data are native endian.
'''
        self.float32      = numpy.dtype('<f4').newbyteorder('=')
        self.int32        = numpy.dtype('<i4').newbyteorder('=')
        self.str32        = numpy.dtype('<S4').newbyteorder('=')
        self.file_float32 = self.float32
        self.file_int32   = self.int32
        self.file_str32   = self.str32
    #--- End: def

    @property
    def printfdr(self):
        '''
        '''
        fdr = []
        for i in range(45):
            fdr.append('%s:_%s' % (_lhd_items[i].upper(), self.lhd[i]))
        for i in range(19):
            fdr.append('%s:_%s' % (_bhd_items[i].upper(), self.bhd[i]))
        fdr = textwrap.fill( ' '.join(fdr), width=79)
        fdr = [fdr.replace(':_', ': ')]

        if self.extra:
            fdr.append('EXTRA DATA:')
            for key in self.extra:
                fdr.append('%s: %s' % (key, repr(self.extra[key])))
        
        print '\n'.join(fdr)
    #--- End: def

#--- End: class


def unrotated_latlon(rotated_lat, rotated_lon, pole_lat, pole_lon):
    '''

Create 2-d arrays of unrotated latitudes and longitudes.
'''
    # Make sure rotated_lon and pole_lon is in [0, 360)
    pole_lon = pole_lon % 360.0

    # Convert everything to radians
    pole_lon *= _PI_OVER_180
    pole_lat *= _PI_OVER_180

    cos_pole_lat = numpy.cos(pole_lat)
    sin_pole_lat = numpy.sin(pole_lat)

    # Create appropriate copies of the input rotated arrays
    rot_lon = rotated_lon.copy()
    rot_lat = rotated_lat.view()

    # Make sure rotated longitudes are between -180 and 180
    rot_lon %= 360.0
    rot_lon = numpy.where(rot_lon < 180.0,
                          rot_lon, rot_lon-360)

    # Create 2-d arrays of rotated latitudes and longitudes in radians
    nlat = rot_lat.size
    nlon = rot_lon.size
    rot_lon = numpy.resize(numpy.deg2rad(rot_lon), (nlat, nlon))    
    rot_lat = numpy.resize(numpy.deg2rad(rot_lat), (nlon, nlat))
    rot_lat = numpy.transpose(rot_lat, axes=(1,0))

    # Find unrotated latitudes
    CPART = numpy.cos(rot_lon) * numpy.cos(rot_lat)
    sin_rot_lat = numpy.sin(rot_lat)
    x = cos_pole_lat * CPART + sin_pole_lat * sin_rot_lat
    x = numpy.clip(x, -1.0, 1.0)
    unrotated_lat = numpy.arcsin(x)
    
    # Find unrotated longitudes
    x = -cos_pole_lat*sin_rot_lat + sin_pole_lat*CPART
    x /= numpy.cos(unrotated_lat)   # dch /0 or overflow here? surely
                                    # lat could be ~+-pi/2? if so,
                                    # does x ~ cos(lat)?
    x = numpy.clip(x, -1.0, 1.0)
    unrotated_lon = -numpy.arccos(x)
    
    unrotated_lon = numpy.where(rot_lon > 0.0, 
                                -unrotated_lon, unrotated_lon)
    if pole_lon >= ATOL():
        SOCK = pole_lon - _PI
    else:
        SOCK = 0
    unrotated_lon += SOCK

    # Convert unrotated latitudes and longitudes to degrees
    unrotated_lat = numpy.rad2deg(unrotated_lat)
    unrotated_lon = numpy.rad2deg(unrotated_lon)

    return unrotated_lat, unrotated_lon

    # Return unrotated latitudes and longitudes
#    return numpy.rollaxis(numpy.dstack((unrotated_lat, unrotated_lon)), 2)
#--- End: def

def _extra_data(ppfile, m=None):
    '''

Read the extra data (if any) of the PP field at the current position
in the given PP file, decode it and store it in the field's metadata
object.

:parameter m: A PP field's metadata object
:type m: PPMetadata

Excerpt from UMDP F3 vn7.8
--------------------------
The data record may include 'extra data' in addition to the usual
grid-point values. Cross-section fields must always be followed by
extra data giving the x- and y-coordinate values of each column and
row of the grid. (This is to allow irregular grid). It is also
possible to use extra data in conjunction with other types of field,
eg. a character string description of the field. The header variable
LBEXT shows the length of extra data associated with a field. Thus the
total length (in words) of the data record (LBLREC) equals the number
of grid-point values plus LBEXT. Extra data is arranged in vectors,
each of which is made up of an integer code followed by data
values. The integer code is (1000xIA + IB), where IA is the length of
the vector in words (i.e. usually the number of data values) and IB is
a standard code. A zero integer code indicates no (more) extra data.
For cross-sections the extra data MUST start with an x-vector followed
by a y-vector; optionally other extra data may follow.
'''
    if m.lhd[lbext] <= 0:
        # Skip the 4th block control word
        ppfile.seek(_word_size, os.SEEK_CUR) 
        m.extra = {}
        return m

    extra = {}

    domain_titles = []

    len = 0        
    while len < m.lhd[lbext]:

        intcode = numpy.fromfile(ppfile, dtype=m.file_int32, count=1)

        # Stop if we have run out of extra data
        if not (intcode or intcode[0]):
            break

        # Find ia = size of this extra data variable, 
        #      ib = type of this extra data variable
        ia, ib = divmod(intcode[0], 1000)
        
        if ib in (10, 11):
            # Read characters from the file
            title = numpy.fromfile(ppfile, dtype=m.file_str32, count=ia)
            title = ''.join(title)
            if ib == 10:
                extra[_extra_data_name[ib]] = numpy.array([title])
            else:
                domain_titles.append(title)
        else:
            # Don't read an array of number, but store it as file
            # pointer
            parameters = {'_file'     : ppfile, ### dch!!!!!!
                          '_start'    : ppfile.tell(),
                          '_dtype'    : m.file_float32,
                          '_shape'    : (ia,),
                          }        
            extra[_extra_data_name[ib]] = PPFileArray(**parameters)
            
            # Skip to the next extra data array
            ppfile.seek(ia*_word_size, os.SEEK_CUR)
        #--- End: if

        len += ia + 1
    #--- End: while    

    for bounds_type in ('', '_domain'):

        for dim in ('x', 'y'):
            dim_bounds_type = '%s%s' % (dim, bounds_type)
            lower  = '%s_lower_bound' % dim_bounds_type
            upper  = '%s_upper_bound' % dim_bounds_type
            bounds = '%s_bounds'      % dim_bounds_type
            
            if lower in extra and upper in extra:
                extra[bounds] = PPFileArrayBounds(extra[lower], extra[upper])         
                extra.pop(lower)
                extra.pop(upper)
    #--- End: for

    # If we have domain titles then add them to the extra data
    # dictionary as a numpy array
    if domain_titles:
        extra[_extra_data_name[11]] = numpy.array(domain_titles)

    # Skip the 4th block control word so that we're ready to read the
    # next PP field
    ppfile.seek(_word_size, os.SEEK_CUR)            

    # Add the extra data to the metadata object
    m.extra = extra

    # Return the updated metadata object
    return m
#--- End: def

def _field_data(ppfile, m=None):    
    '''

Return a file pointer to the field's 2-d data array.

:parameter ppfile: dch 

:type ppfile: file

:parameter m: A PP field's complete metadata
:type m: PPMetadata

:returns: A `.PPFileArray` object
'''
    size = m.lhd[lblrec] - m.lhd[lbext]

    if not size:
        return

    # Set the data type
    binary_mask = False
    if m.lhd[lbuser1] == 1:
        datatype = m.file_float32
    elif m.lhd[lbuser1] == 2:
        datatype = m.file_int32
    elif m.lhd[lbuser1] == 3:
        datatype = m.file_int32
        binary_mask = True
    else:
        # Unknown data type => assume real
        datatype = m.file_float32

    # Set the shape
    shape = (m.lhd[lbrow], m.lhd[lbnpt])

    file_offset = ppfile.tell()

    # Create the data object by setting up a parameters dictionary to
    # initialize a PPFileArray object
    parameters = {'_binary_mask': binary_mask,
                  '_file'       : ppfile.name,
                  '_file_offset': file_offset,
                  'dtype'       : datatype,
                  'shape'       : shape,
                  'size'        : size,
                  'ndim'        : 2,
                  }

    # Set the _FillValue from BMDI (unless it's -1.0e30, which is a
    # flag for the field containing no missing data). Note that the
    # _FillValue must be of the same type as the data values.
    if m.bhd[bmdi] != -1.0e30:
        parameters['_FillValue'] = m.bhd[bmdi]
        if datatype == m.file_int32:
            parameters['_FillValue'] = int(parameters['_FillValue'])

    # Treat BDATUM as an add_offset if it is not 0.0
    if abs(m.bhd[bdatum]) > ATOL():
        parameters['add_offset'] = m.bhd[bdatum]

    # Treat BMKS as a scale_factor if it is not 0.0 or 1.0
    if (abs(m.bhd[bmks]-1.0) > ATOL() + RTOL() and 
        abs(m.bhd[bmks])     > ATOL()):
        parameters['scale_factor'] = m.bhd[bmks]

    data = PPFileArray(**parameters)

    # Skip data and 4th block control word to get to the start of the
    # next PP field in the file
    ppfile.seek(file_offset + size*_word_size, 0) #os.SEEK_CUR)

    # Return the data object
    return data
#--- End: def

def _set_coordinate_data(c, m=None,
                         array=None, bounds_array=None, climatology=False, 
                         coord_type=None, extra_type=None,
                         parameters={}):
    '''
    
Set a coordinate's data in place.

:Parameters:

    c : Coordinate

    m : PPMetadata
        A PP field's complete metadata.

    array : array-like, optional

    bounds_array : array-like, optional

    climatology : bool, optional

:Returns:

    None

'''
    def _insert_data(c, array=None, bounds_array=None, 
                     climatology=False):
        '''

:Parameters:

    c : Coordinate

    array : array-like, optional

    bounds_array : array-like, optional

    climatology : bool, optional

:Returns:

    None

'''
        _FillValue = getattr(c, '_FillValue', None)

        # Set the coordinate's data array
        if array is not None:
            c.Data = Data(array,
                          units      = c.Units,
                          _FillValue = _FillValue,
                          chunk      = False)

        # Set the coordinate's bounds array
        if bounds_array is not None:
            c.bounds      = CoordinateBounds()        
            c.bounds.Data = Data(bounds_array,
                                 units      = c.Units,
                                 _FillValue = _FillValue,
                                 chunk      = False)
            
            if climatology:
                c.climatology = True

        # Return the updated coordinate
#        return c
    #--- End: def

    if c is None:
        return

    if array is not None or bounds_array is not None:
        _insert_data(c, array, bounds_array, climatology)
        return c
    #--- End: if

    origin        = None
    create_bounds = True

    #-----------------------------------------------------------------
    # Set time data from vtime and dtime
    #-----------------------------------------------------------------
    if coord_type == 't':
        # This PP field's data array does not have time as one of its
        # two axes, so create a size 1 dimension coordinate for time
        ib = m.lbtim_ib
        if ib <= 1 or m.vtime >= m.dtime:
            array = numpy.array(m.vtime)                       # 0-d
        else:
            array = numpy.array(0.5*(m.vtime + m.dtime))       # 0-d
            if ib <= 3:
                bounds_array = numpy.array([m.vtime, m.dtime]) # 1-d
                if ib == 3:
                    climatology = True
    #--- End: if

    # ----------------------------------------------------------------
    #
    # ----------------------------------------------------------------
    if coord_type == 'x':
        if abs(m.bhd[bdx]) <= ATOL() or m.timeseries: # DCH 
            # Create a coordinate data object from the field's extra
            # data
            try:
                array = numpy.array(m.extra[coord_type])
            except KeyError: 
                pass
            
            # Create a coordinate bounds data object from the field's
            # extra data
            try:
                 bounds_array = numpy.array(m.extra[coord_type+'_bounds'])
            except KeyError: 
                pass
                
        else:
            # Find, from the PP header, parameters for creating
            # coordinate data and coordinate bounds data objects
            delta  = m.bhd[bdx]
            origin = m.bhd[bzx] + delta
            size   = m.lhd[lbnpt]

            if (m.ix == 11 and origin + delta*(size-1) > 360.0):
                origin -= 360.0

            if m.ix in (13, 40, 99):
                create_bounds = False                    

    elif coord_type == 'y':
        if abs(m.bhd[bdy]) <= ATOL() or m.timeseries: # DCH Set a
            # Create a coordinate data object from the field's extra
            # data
            try:
                array  = numpy.array(m.extra[coord_type])
            except KeyError: 
                pass
            
            # Create a coordinate bounds data object from the field's
            # extra data
            try:
                bounds_array = numpy.array(m.extra[coord_type+'_bounds'])
            except KeyError: 
                pass
        else:
            # Find, from the PP header, parameters for creating
            # coordinate data and coordinate bounds data objects
            delta  = m.bhd[bdy]
            origin = m.bhd[bzy]
            size   = m.lhd[lbrow]

            if (m.iy == 11 and origin + delta*(size-1) > 360.0):
                origin -= 360.0

            if m.iy in (13, 40, 99):
                create_bounds = False         
    #--- End: if
        
    if extra_type is not None:
        try:
            m.extra[extra_type] 
        except KeyError: 
            pass
        else:
            if not extra_type.endswith('bounds'):
                # Create a coordinate data object from the field's extra
                # data
                array = numpy.array(m.extra[extra_type])
            else:
                # Create a coordinate bounds data object from the field's
                # extra data
                bounds_array = numpy.array(m.extra[extra_type])
    #--- End: if

    if origin is not None:
        # Create a 1-d coordinate data object from an origin, size and
        # delta
        array = numpy.arange(origin, origin+delta*size, delta, dtype='float32')

        # Create the 1-d coordinate's bounds array
        if create_bounds:
            bounds_array       = numpy.empty((size,2), dtype=array.dtype)          
            bounds_array[:, 0] = array - 0.5*delta
            bounds_array[:, 1] = array + 0.5*delta
    #--- End: if 

    # Set the coordinate's data array
    _insert_data(c, array, bounds_array, climatology)
   
    # Return coordinate
    return c
#--- End: def

def _create_Coordinate(f, key, axis_code=None, m=None, pubattr={},
                       data=None, bounds=None, climatology=False, 
                       coord_type=None, extra_type=None,
                       parameters=None, dimensions=[], cfunits=None):
    '''
    
Create a coordinate variable.

:parameter m: A PP field's complete metadata
:type m: PPMetadata

'''
    c = Coordinate()

    if axis_code:
        # Set attributes
        c.axis          = _coord_axis[axis_code]
        c.positive      = _coord_positive[axis_code]
        c.standard_name = _coord_standard_name[axis_code]
        c.units         = _coord_units[axis_code]

        if c.standard_name is None:
            c.long_name = _coord_long_name[axis_code]
            del c.standard_name
        #--- End: if

        # Special cases
        if axis_code in (20, 23):
            # Time
            c.units    = m.time_units
            c.calendar = m.calendar
        elif 100 < m.lhd[lbcode] < 10000:            
            # Rotated lat-lon
            if axis_code == 10:
                c.units         = 'degrees'
                c.standard_name = 'grid_latitude'
            elif axis_code == 11:
                c.units         = 'degrees'
                c.standard_name = 'grid_longitude'
        #--- End: if

        # Remove None attributes        
        [c.delattr(attr) for attr in ('axis', 'positive') 
         if c.getattr(attr) is None]
    #--- End: if

    # Apply extra public attributes, or overwrite existing ones
    for attr, value in pubattr.iteritems():
        if value is not None:
            c.setattr(attr, value)
        else:
            try:
                c.delattr(attr)
            except AttributeError: 
                pass
    #--- End: for

    # units
    if cfunits is not None:
        c.Units = cfunits.copy()

    # data
    _set_coordinate_data(c, m=m,
                         array        = data,
                         bounds_array = bounds, 
                         climatology  = climatology,
                         coord_type   = coord_type,
                         extra_type   = extra_type, 
                         parameters   = parameters)

    f.space.dimensions[key] = dimensions

    if key.startswith('d'): # d for dim
        # Find the dimension size
        try:
            dimension_size = c.size
        except AttributeError:
            if coord_type == 'x':
                dimension_size = m.lhd[lbnpt]
            elif coord_type == 'y':
                dimension_size = m.lhd[lbrow]
            else:
                raise ValueError,"Can't find dimension size for coord_type '%s'" % \
                    coord_type
        #--- End: try
        f.space.dimension_sizes[key] = dimension_size
    #--- End: if

    f.space[key] = c

    # Set the coordinate's netCDF variable name
    c.ncvar = getattr(c, 'standard_name', 
                      getattr(c, 'long_name', None))
        
    return c
#--- End: def

def _decode_exp(lbexp):
    '''
    
Decode the integer value of LBEXP in the PP header into a 5-character
runid.

If this value has already been decoded, then it will be returned from
the cache, otherwise the value will be decoded and then added to the
cache.

:parameter lbexp: The value of LBEXP in the PP header
:type lbexp: int

:returns: A 5-character string derived from LBEXP. If LBEXP is a
          negative integer then that number is returned as a string.
'''
    try:
        # Try to return a cached decoding of this LBEXP
        return _cache['runid'][lbexp]
    except KeyError:
        # This LBEXP hasn't been decoded yet
        pass

    if lbexp < 0:
        return str(lbexp)

    # Convert LBEXP to a binary string, filled out to 30 bits with
    # zeros
    bits = bin(lbexp)
    bits = bits.lstrip('0b').zfill(30)

    # Step through 6 bits at a time, converting each 6 bit chunk into
    # a decimal integer, which is used as an index to the characters
    # lookup list.
    runid = []
    for i in xrange(0,30,6):
        index = int(bits[i:i+6], 2) 
        try:
            runid.append(_characters[index])
        except IndexError:
            # Can't translate index >= 36
            pass
    #--- End: for
    runid = ''.join(runid)

    # Enter this runid into the cache
    _cache['runid'][lbexp] = runid

    # Return the runid
    return runid
#--- End: def 

def _set_axis_codes(m=None):
    '''
    
Find the axis codes for the dimensions occupying the T, Z, Y and X
dimension positions of the PP header.

:parameter m: A PP field's complete metadata
:type m: PPMetadata

:returns: The PP metadata object updated with the axis codes set in
          its `it`, `iz`, `iy` and `ix` attributes.
'''
    code = m.lhd[lbcode]
    if code in (1, 101, 2, 102):
        m.ix, m.iy = 11, 10
    elif code >= 10000:
        temp,   ix = divmod(code, 10000)
        m.ix, m.iy = divmod(ix, 100)
    else:
        m.ix, m.iy = None, None

    # Determine if we have a site-time cross section and, if so, if it
    # is a timeseries
    m.site_time_cross_section = False
    m.timeseries              = False
    if (m.ix, m.iy) == (13, 23):
        m.site_time_cross_section = True
        m.timeseries = m.lhd[lbuser3] == m.lhd[lbrow]

    # Set iv from LBVC
    
    try:
        m.iv = _lbvc_codes[m.lhd[lbvc]]
    except KeyError:
        m.iv = None

    # Set it
    if m.calendar == 'gregorian':
        m.it = 20
    else:
        m.it = 23

    return m
#--- End def

def _pp_header(ppfile):
    '''

Read the header of the PP field at the current position in the given
PP file and create some derived quantities.

:parameter ppfile: An open file object for a PP file.
:type ppfile: file

:returns: A :class:`~cf.pp.PPMetadata` object containing all of the metadata for the
          current PP field in the input file, including some derived
          metadata quantities, such as 'calendar', 'time_units',
          'timeseries', etc.
'''

    m = PPMetadata()

    # Read 1st block control word
    bcw1 = numpy.fromfile(ppfile, dtype=m.file_int32, count=1)
   
    # Stop now if we have reached the end of the file
    if not bcw1:
        return None

    # Change data types if we have a wrong endian PP file. ('Wrong'
    # simply means different to that assumed by class PPMetadata.)
    if bcw1[0] == 65536:        
        m.file_float32 = m.float32.newbyteorder('S')
        m.file_int32   = m.int32.newbyteorder('S')
        m.file_str32   = m.str32.newbyteorder('S')
    elif bcw1[0] != 256:
        raise RuntimeError,"Not a PP field: Magic number = %s" % bcw1[0]

    # Read the integer and real headers (and the 2nd and 3rd block
    # control words)
    m.lhd = numpy.fromfile(ppfile, dtype=m.file_int32  , count=45)
    m.bhd = numpy.fromfile(ppfile, dtype=m.file_float32, count=21)

    # Set some derived metadata quantities. Note that we need to this
    # before we get any extra data.
    m.lbtim_ia, ib         = divmod(m.lhd[lbtim], 100)
    m.lbtim_ib, m.lbtim_ic = divmod(ib, 10)

    # Set units to 'days since 1860-12-1' for now, since field
    # aggregation can't deal with equivalent but non-equal units, yet.
#    m.time_units = 'days since 1860-12-1'
    m.time_units = 'days since %d-%d-%d' % (m.lhd[lbyr], m.lhd[lbmon], m.lhd[lbdat])
#    print m.time_units
    if m.lbtim_ic == 1:
        m.calendar = 'gregorian'
    else:
        m.calendar = '360_day'

    # For speed, try to return a cached numeric time value, otherwise
    # calculate one.
    v = (m.lhd[lbyr] , m.lhd[lbmon] , m.lhd[lbdat] , m.lhd[lbhr] , m.lhd[lbmin])
    d = (m.lhd[lbyrd], m.lhd[lbmond], m.lhd[lbdatd], m.lhd[lbhrd], m.lhd[lbmind])
    
    vdatetime = datetime(v[0], v[1], v[2], v[3], v[4])
    ddatetime = datetime(d[0], d[1], d[2], d[3], d[4])

    for attr, t, mm in zip(('vtime', 'dtime'),
                           ( vdatetime ,  ddatetime),
                           ( v, d)):

#        print 'XX ','days since %d-%d-%d' % (v[0:3])
#        m.time_units = 'days since %d-%d-%d' % (v[0:3])
#        print mm

#        try:
#            value = _cache['time'][mm]
#            print value
#        except KeyError:
        value = netCDF4.date2num(t, m.time_units, m.calendar)
#            _cache['time'][mm] = value
        setattr(m, attr, value)
    #--- End: for
#    print m.vtime, m.dtime
    version, model = divmod(m.lhd[lbsrce], 10000)
    if model > 0:
        m.model = 'UM'
    if version > 0:
        m.version = str(version)
    
    m = _set_axis_codes(m=m)

    # Return the metadata
    return m
#--- End: def

def _test_pp_condition(pp_condition, m=None):
    '''

Test whether or not a field satisfies the condition specified for a PP
STASH code to standard name conversion.

:keyword m: A PP field's complete metadata
:type m: PPMetadata

:returns: True if a field satisfies the condition specified, False
          otherwise.
'''
    if pp_condition is '':
        # Return True if no condition is set
        return True
    
    # Still here?
    if pp_condition == 'true_latitude_longitude':
        if m.lhd[lbcode] in (1, 2):
            return True
        # Check pole location in case of incorrect LBCODE
        if (abs(m.bhd[bplat]-90.0) <= ATOL() + RTOL()*90.0 and 
            abs(m.bhd[bplon]) <= ATOL()):
            return True

    elif pp_condition == 'rotated_latitude_longitude':
        if m.lhd[lbcode] in (101, 102):
            return True
        # Check pole location in case of incorrect LBCODE
        if not (abs(m.bhd[bplat]-90.0) <= ATOL() + RTOL()*90.0 and 
                abs(m.bhd[bplon]) <= ATOL()):
            return True
        
    else:
        raise ValueError,"Unknown PP condition in STASH code conversion table: '%s'" % \
            pp_condition

    # Still here? Then the condition has not been satisfied.
    return False
#--- End: def

def _test_umversion(valid_from, valid_to, m=None):
    '''

Test whether or not the UM version applicable to this field is within
the given range.

If possible, the UM version is derived from the PP header and stored
in the metadata object. otherwise it is taken from the UMVERSION
environment variable.

:returns: True the UM version applicable to this field is within the
          range, False otherwise.
'''
    try:
        umversion = m.version
    except AttributeError:
        try:            
            umversion = _UMVERSION
        except NameError:
            raise RuntimeError,"UM version is not set in PP header nor the environment"
    #--- End: try

    if valid_from is '':
        valid_from = None

    if valid_to is '':
        if valid_from <= umversion:
            return True 
    elif valid_from <= umversion <= valid_to:
        return True 

    return False
#--- End: def

def _field_attributes(f, m=None):
    '''

Set public and private attributes on the space (including the standard
name).
'''
    f.pp   = m
    f.file = m.file

    f._FillValue = m.bhd[bmdi]
    f.setattr('Conventions', __Conventions__)    
    f.setattr('history'    , m.history)
    f.setattr('runid'      , _decode_exp(m.lhd[lbexp]))

#    if abs(m.bhd[bdatum]) > ATOL():
#        f.setattr('add_offset', m.bhd[bdatum])
#
#    if not abs(m.bhd[bmks]-1.0) <= ATOL() + RTOL():
#        f.setattr('scale_factor', m.bhd[bmks]) # dch: check how this interacts with f.units       

    try:
        f.setattr('source', m.model)
    except AttributeError:
        pass
    else:
        try:
            f.setattr('source', '%s vn%s' % (f.getattr('source'), m.version))
        except AttributeError:
            pass        

    try:
        f.setattr('title', m.extra['title'])
    except KeyError:
        pass

    # ----------------------------------------------------------------
    # Set the field's standard_name, long_name and units from its
    # STASH code and possibly other properties
    # ----------------------------------------------------------------
    stash    = m.lhd[lbuser4]
    submodel = m.lhd[lbuser7]

    f.setattr('stash_code', stash)
    f.setattr('submodel'  , submodel)

    m.cf_info = {}

    # The STASH code has been set in the PP header, so try to find
    # its standard_name from the conversion table
    try:
        lines = _stash2standard_name[(submodel, stash)]
    except KeyError:
        if stash:
            f.long_name = 'PP_%d_%d' % (submodel, stash)
        else:
            f.long_name = 'PP_%d_fc%d' % (submodel, m.lhd[lbfc])
    else:
        for (long_name, 
             units,
             valid_from,
             valid_to, 
             standard_name,
             cf_info,
             pp_condition) in lines:

            # Check that conditions are met             
            version_ok      = _test_umversion(valid_from, valid_to, m=m)
            pp_condition_ok = _test_pp_condition(pp_condition, m=m)

            if not (version_ok and pp_condition_ok):                    
                continue

            # Still here? Then we have our standard_name, etc.
            f.long_name = long_name.rstrip()
            if units:
                f.units = units
            if standard_name:
                f.standard_name = standard_name
                                    
            m.cf_info = cf_info
                
            break
        #--- End: for            
    #--- End: try

    # If there is no standard name then set a private aggregating name
    # based on the submodel and STASHcode (and maybe the field code)
    try:
        f.standard_name
    except AttributeError:
        if stash:
#            f.setpriv('aggregating_name', 'PP_%d_%d' % (submodel, stash))
            f.id = 'PP_%d_%d' % (submodel, stash)
        else:
#            f.setpriv('aggregating_name', 'PP_%d_fc%d' % (submodel, m.lhd[lbfc]))
            f.id = 'PP_%d_fc%d' % (submodel, m.lhd[lbfc])
    #--- End: try

    # If there is no long name then set a default
    try:
        f.long_name
    except AttributeError:
        if stash:
            f.long_name = 'PP_%d_%d' % (submodel, stash)
        else:
            f.long_name = 'PP_%d_fc%d' % (submodel, m.lhd[lbfc])

    # Set the field's coordinate's netCDF variable name
#    f.ncvar = getattr(f, 'standard_name', f.getpriv('aggregating_name', None))
    f.ncvar = getattr(f, 'standard_name', getattr(f, 'id', None))
        
#
#        try:
#            f.getattr('long_name')
#            try:   # dch this is a nightmare
#                if not (f.getattr('units') == '%' and f.getattr('scale_factor') == 0.01):
#                    f.setattr('units', '%s %s' % (f.getattr('scale_factor'), f.getattr('units')))
#                    f.delattr('scale_factor')
#            except AttributeError:
#                pass
#        except AttributeError:
#            # Default long name in the absence of a conversion
#            f.setattr('long_name', 'STASHcode=%s' % stash)
#    #--- End: if

    return f
#--- End: def

def read_stash2standard_name(table=None, delimiter='!'):
    ''' 

Read the STASH to standard_name conversion table, as found at
http://puma.nerc.ac.uk/STASH_to_CF/STASH_to_CF.txt, into a dictionary.

:Parameters:

    table : str, optional
        Use the conversion table at this file location. By default the
        table will be looked for at
        ``os.path.join(os.path.dirname(cf.__file__),'etc/STASH_to_CF.txt')``

    delimiter : str, optional
        The table's delimiter.

:Returns:

    out: dict
        A dictionary of tuples of lists

*Examples*

>>> stash2sn = read_stash2standard_name()
>>> print stash2sn[(1,2)]
(['U COMPNT OF WIND AFTER TIMESTEP     ',
  'm s-1',
  '',
  '',
  'eastward_wind',
  '',
  'true_latitude_longitude'],
 ['U COMPNT OF WIND AFTER TIMESTEP     ',
  'm s-1',
  '',
  '',
  'x_wind',
  '',
  'rotated_latitude_longitude'])
'''

    # 0  Model           
    # 1  STASH code      
    # 2  STASH name      
    # 3  units           
    # 4  valid from UM vn
    # 5  valid to   UM vn
    # 6  standard_name   
    # 7  CF extra info   
    # 8  PP extra info

    if table is None:
        # Use default conversion table
        package_path = os.path.dirname(__file__)
        table = os.path.join(package_path, 'etc/STASH_to_CF.txt')

    lines = csv.reader(open(table, 'r'), 
                       delimiter=delimiter, skipinitialspace=True)

    raw_list = []
    [raw_list.append(line) for line in lines]

    # Get rid of comments
    for line in raw_list[:]:
        if line[0].startswith('#'):
            raw_list.pop(0)
            continue
        break
    #--- End: for

    # Convert to a dictionary which is keyed by (submodel, STASHcode)
    # tuples

    (model, stash, name,
     units, 
     valid_from, valid_to,
     standard_name, cf, pp) = range(9)
        
    stash2sn = {}
    for x in raw_list:
        key = (int(x[model]), int(x[stash]))
        try:            
            cf_info = {}
            if x[cf]:
                for d in x[7].split():
                    if d.startswith('height='): 
                        cf_info['height'] = re.split(_number_regex, d,
                                                     re.IGNORECASE)[1:4:2]
                        if cf_info['height'] == '':
                            cf_info['height'][1] = '1'

                    if d.startswith('below_'):
                        cf_info['below'] = re.split(_number_regex, d,
                                                     re.IGNORECASE)[1:4:2]
                        if cf_info['below'] == '':
                            cf_info['below'][1] = '1'

                    if d.startswith('where_'):         
                        cf_info['where'] = d.replace('where_', 'where ', 1)
                    if d.startswith('over_'):         
                        cf_info['over'] = d.replace('over_', 'over ', 1)

            x[cf] = cf_info                    
        except IndexError:
            pass

        x[pp] = x[pp].rstrip()

        line = (x[name:],)

        try:
            stash2sn[key] += line
        except KeyError:
            stash2sn[key] = line

    #--- End: for

    return stash2sn
#--- End: def

# Create the STASH code to standard_name conversion dictionary at import time
_stash2standard_name = read_stash2standard_name()


def read(infile, close=False):
    ''' 

Read fields from an input PP file on disk.

All data arrays from all field components are not read from disk but
are stored in the returned fields as file pointers.

:Parameters:

    filename : file or str
        An open file object, or a string giving the file name, from
        which to read.

    close : bool, optional
       If True then close the file after it has been read.

:Returns:

    out : FieldList
        The fields in the file.

**Examples**

>>> f = cf.pp.read('file.pp')
>>> f = cf.pp.read('file[0-9].pp', close=True)
'''

    # coord_type: A short string which acts as .... Valid values are
    # 't'  : time 
    #
    # 'v'  : an axis other than time which has the the same value for
    #        all data points, as described by LBVC
    #
    # 'p'  : pseudolevel
    #
    # 'rwl': radiation wavelength
    #
    # 'x'  : The X axis of the data array
    #
    # 'y'  : The Y axis of the data array

    # ------------------------------------------------------------
    # Preliminaries
    # ------------------------------------------------------------
    # Open the file, if it is not already open
    try:
        ppfile = open(infile, 'rb') 
    except TypeError:
        ppfile = infile

    # Initialize the output list of fields
    fields_in_file = FieldList()

    history = '%s Converted from PP by cf-python v%s' % \
        (time.strftime('%Y-%m-%d %H:%M:%S UTC', time.gmtime()),
         __version__)

    # ****************************************************************
    # Loop round each PP field in the file, reading it into a field
    # ****************************************************************
    while True:

        dimN   = 0  # Counter for dimension coordinates
        auxN   = 0  # Counter for auxiliary coordinates
        transN = 0  # Counter for transforms

        # ============================================================
        # Get the next PP field's metadata and data and find some
        # derived metadata quantities, such as 'calendar',
        # 'time_units', 'timeseries', etc.
        # ============================================================
        # Get the header (lhd and bhd)
        m = _pp_header(ppfile)

        # Stop now if we have reached the end of the file
        if m is None:
            break

        # Get the data
        field_data = _field_data(ppfile, m=m)

        # Get any extra data
        m = _extra_data(ppfile, m=m)

        m.history = history
        m.file    = ppfile.name

        # ============================================================
        # Initialize the field. The involves setting the field's
        # attributes and data and giving it an empty space, for now.
        # ============================================================
        f       = Field()
        f.space = Space()
        space   = f.space

        # Set the field's attributes
        f = _field_attributes(f, m=m)

        axis_name = {}
        axis_dim  = {}
        
        # ============================================================
        # Create the field's space
        # ============================================================
       
        # ------------------------------------------------------------
        # Create a TIME dimension coordinate
        # ------------------------------------------------------------
        dim  = 'dim%(dimN)d' % locals()
        dimN += 1         # Increment dimension number
        tdim = dim      

        # Set coordinate type and axis code
        if m.ix in (20, 23):
            coord_type = 'x'
            axis_code  = m.ix
            # Flag that we have already created an 'X' coordinate
            m.ix = None
        elif m.iy in (20, 23):
            coord_type = 'y'
            axis_code  = m.iy
            # Flag that we have already created a 'Y' coordinate
            m.iy = None
        else:
            coord_type = 't'
            axis_code  = m.it
            
        # Set the coordinate's data
        if m.timeseries:
            # This PP field is a timeseries. The validity time is
            # taken to be the time for the first sample, the data time
            # for the last sample, with the others evenly between.
            size   = m.lhd[lbrow]
            delta  = (m.dtime - m.vtime)/(size - 1)
            origin = m.vtime
            array = numpy.linspace(origin, origin + delta*(size-1), 
                                   num=size)
        

            coord = _create_Coordinate(f, tdim, axis_code, m=m, 
                                       array = array,
                                       dimensions = [tdim])       
        else:
            coord = _create_Coordinate(f, tdim, axis_code, m=m,
                                       coord_type = coord_type, 
                                       dimensions = [tdim])
       
        # Save the dimension name
        if coord_type == 'x':
            data_dim0 = tdim
        elif coord_type == 'y' or m.timeseries:
            data_dim1 = tdim

        # Set the T axis name for cell methods
        axis_name['t'] = coord.standard_name
        axis_dim['t']  = tdim
        
        # ------------------------------------------------------------
        # Create a 'VERTICAL' dimension coordinate. This size 1
        # coordinate will often by a vertical coordinate such as
        # pressure, but could be anything, such as longitude.
        # ------------------------------------------------------------
        if m.iv is not None: 
            dim  = 'dim%(dimN)d' % locals()
            dimN += 1         # Increment dimension number
            vdim = dim      

            if 'height' in m.cf_info:
                # Create the height coordinate from the information
                # given in the STASH to standard_name conversion table
                height, units = m.cf_info['height']

                coord = _create_Coordinate(f, vdim, axis_code=15, m=m,
                                           cfunits    = Units(units),
                                           data       = numpy.array([height],
                                                                    dtype=float),
                                           dimensions = [vdim])
            else:
                # Create the height coordinate from the PP header
                axis_code = m.iv
                data      = None

                # Create the height coordinate from the PP header
                if m.lhd[lbvc] == 9:
                    # atmosphere_hybrid_sigma_pressure_coordinate
                    pstar  = 1.0e5
                    data   = numpy.array([m.bhd[blev] + m.bhd[bhlev]/pstar])
                    bounds = numpy.array([[m.bhd[brlev]  + m.bhd[bhrlev]/pstar, 
                                           m.bhd[brsvd1] + m.bhd[brsvd2]/pstar]])
                elif m.lhd[lbvc] == 65:
                    # atmosphere_hybrid_height_coordinate: Can't
                    # calculate eta (because we don't know the height
                    # of the top level), so we'll create dimension
                    # coordinate for model_level_number instead
                    lev = m.lhd[lblev]
                    if lev >= 0:
                        data      = numpy.array([lev])
                        bounds    = None
                        axis_code = 5
                else:
                    data   = numpy.array([m.bhd[blev]])
                    bounds = numpy.array([[m.bhd[brlev], m.bhd[brsvd1]]])
                    if equals(bounds[0,0], bounds[0,1]):
                        bounds = None
                #--- End: if

                if data is not None:
                    # We have found some data, so create a dimension
                    # coordinate
                    coord = _create_Coordinate(f, vdim, axis_code=axis_code,
                                               m=m,
                                               data       = data,
                                               bounds     = bounds,
                                               dimensions = [vdim])
            #--- End: if

            # Set the V axis name for cell methods
            axis_name['v'] = coord.standard_name
            axis_dim['v']  = vdim
        #--- End: if

        # ------------------------------------------------------------
        # Create a PSEUDOLEVEL dimension coordinate
        # ------------------------------------------------------------
        if m.lhd[lbuser5] > 0: 

            dim  = 'dim%(dimN)d' % locals()
            dimN += 1         # Increment dimension number
            pdim = dim      
            
            coord = _create_Coordinate(f, pdim, axis_code=40, m=m, 
                                       data       = numpy.array([m.lhd[lbuser5]]),
                                       dimensions = [pdim])
        
            # Set the pseudolevel axis name for cell methods
            axis_name['p'] = coord.standard_name  ## dch: should this be something else?
            axis_dim['p']  = pdim
        #--- End: if

        # ------------------------------------------------------------
        # Create a RADIATION WAVELENGTH dimension coordinate
        # ------------------------------------------------------------
            #dch : do this instead of VERTICAL, I wonder
        try:
            rwl, units = m.cf_info['below']
        except (KeyError, TypeError):
            pass
        else:
            
            dim   = 'dim%(dimN)d' % locals()
            dimN += 1         # Increment dimension number
            rwldim = dim      
            
            # Create the radiation wavelength coordinate
            pubattr = {'standard_name' : 'radiation_wavelength'}

            coord = _create_Coordinate(f, rwldim, axis_code=None, m=m,
                                       pubattr    = pubattr,
                                       cfunits    = Units(units),
                                       data       = numpy.array([rwl], 
                                                                dtype=float),
                                       bounds     = numpy.array([[0.0, rwl]]),
                                       dimensions = [rwldim])
            
            # Set the radiation wavelength axis name for cell methods
            axis_name['rwl'] = coord.standard_name
        #--- End: try

        # ------------------------------------------------------------
        # Create X and Y dimension coordinates, if they haven't
        # already been created. These coordinates represent the 'x'
        # and 'y' dimensions of the data array. In the case of cross
        # sections, one of them may have already been created.
        # ------------------------------------------------------------
        # Set coordinate type and axis code
        for coord_type, axis_code in zip(('y' , 'x' ),
                                         (m.iy, m.ix)):

            # Skip this coordinate if it has already been created
            if axis_code is None:
                continue    

            # Still here? Then create a coordinate
            dim   = 'dim%(dimN)d' % locals()           
            dimN += 1         # Increment dimension number       

            coord = _create_Coordinate(f, dim, axis_code, m=m,
                                       coord_type=coord_type,
                                       dimensions=[dim])
        
            if coord_type == 'y':
                ydim      = dim
                data_dim0 = ydim
            else:
                xdim      = dim
                data_dim1 = xdim
 
            # Set the X or Y axis name for cell methods
            axis_name[coord_type] = coord.standard_name
            axis_dim[coord_type]  = dim
        #--- End: for

        # ============================================================
        # Create 1-d, size 1 auxiliary coordinates
        # ============================================================

        vc = m.lhd[lbvc]
        if vc == 9:
            # ------------------------------------------------------------
            # Atmosphere hybrid sigma pressure coordinate components
            # (see Appendix D of the CF conventions)
            # ------------------------------------------------------------
            for long_name, value, units, bounds in zip(
                ('atmosphere_hybrid_sigma_pressure_coordinate_ak',      
                 'atmosphere_hybrid_sigma_pressure_coordinate_bk'),
                (m.bhd[bhlev],                     # value
                 m.bhd[blev]),
                ('Pa',                             # units
                 '1'),
                ([m.bhd[bhrlev], m.bhd[brsvd2]],   # bounds
                 [m.bhd[brlev] , m.bhd[brsvd1]]),
                ):
                aux   = 'aux%(auxN)d' % locals()           
                auxN += 1         # Increment auxiliary number
                
                pubattr = {'standard_name': long_name}

                coord = _create_Coordinate(f, aux, axis_code=None,
                                           m=m,
                                           pubattr    = pubattr, 
                                           cfunits    = Units(units),
                                           data       = numpy.array([value]),
                                           bounds     = numpy.array([bounds]),
                                           dimensions = [vdim])
            #--- End: for
        elif vc == 65:
            # ------------------------------------------------------------
            # Atmosphere hybrid height coordinate components (see
            # Appendix D of the CF conventions)
            # ------------------------------------------------------------
            for long_name, value, units, bounds in zip(
                ('atmosphere_hybrid_height_coordinate_ak',      
                 'atmosphere_hybrid_height_coordinate_bk'),
                (m.bhd[blev],                      # value
                 m.bhd[bhlev]),
                ('m',                              # units
                 '1'),
                ([m.bhd[brlev] , m.bhd[brsvd1]],   # bounds
                 [m.bhd[bhrlev], m.bhd[brsvd2]])
                 
                ):
                aux   = 'aux%(auxN)d' % locals()           
                auxN += 1         # Increment auxiliary number
                
                pubattr = {'standard_name': long_name}

                coord = _create_Coordinate(f, aux, axis_code=None,
                                           m=m,
                                           pubattr    = pubattr, 
                                           cfunits    = Units(units),
                                           data       = numpy.array([value]),
                                           bounds     = numpy.array([bounds]),
                                           dimensions = [vdim])
            #--- End: for
  
        #--- End: if

        # ------------------------------------------------------------
        # forecast_reference_time auxiliary coordinate
        # ------------------------------------------------------------
        if m.lbtim_ib == 1: 
            aux   = 'aux%(auxN)d' % locals()           
            auxN += 1         # Increment auxiliary number
       
            pubattr = {'standard_name': 'forecast_reference_time'}

            coord = _create_Coordinate(f, aux, axis_code=None, m=m,
                                       pubattr    = pubattr,
                                       cfunits    = Units(m.time_units,
                                                          m.calendar),
                                       data       = numpy.array([m.dtime], 
                                                                dtype=float),
                                       dimensions = [tdim])           
        #--- End: if

        # -------------------------------------------------------
        # model_level_number auxiliary coordinate
        # -------------------------------------------------------
        lev = m.lhd[lblev]
        if (m.iv is not None and 
            m.lhd[lbvc] in (2, 9) and 
            lev not in (7777, 8888, 9999)): # dch check 7777
            aux   = 'aux%(auxN)d' % locals()           
            auxN += 1         # Increment auxiliary number

            coord = _create_Coordinate(f, aux, axis_code=5, m=m, 
                                       pubattr    = {'axis': None}, 
                                       data       = numpy.array([lev]), 
                                       dimensions = [vdim])
        #--- End: if

        # ------------------------------------------------------------
        # Domain title auxiliary coordinate
        # ------------------------------------------------------------
        try:            
            domain_titles = m.extra['domain_title']
        except KeyError:
            pass
        else:
            dim = None
            if m.ix == 13:
                dim = xdim
            elif m.iy == 13:
                dim = ydim
            if dim:
                aux   = 'aux%(auxN)d' % locals()           
                auxN += 1                        # Increment auxiliary number

                pubattr = {'long_name': 'domain title'}

                coord = _create_Coordinate(f, aux, axis_code=None, m=m, 
                                           data       = numpy.array(domain_titles), # dch array, or not?
                                           pubattr    = pubattr,
                                           dimensions = [dim])
        #--- End: try

        # ============================================================
        # Create N-d (N>1) auxiliary coordinates
        # ============================================================

        # ------------------------------------------------------------
        # 2-d unrotated latitude and longitude auxiliary coordinates
        # ------------------------------------------------------------
        if 'degrees' == space[ydim].units == space[xdim].units:

            if space[ydim].standard_name == 'grid_longitude':
                dim_lat, dim_lon = xdim, ydim
                rotated_lon = space[ydim]
                rotated_lat = space[xdim]
            else:
                dim_lat, dim_lon = ydim, xdim
                rotated_lon = space[xdim]
                rotated_lat = space[ydim]

            # Try to find the appropriate unrotated latitude and
            # longitude arrays in the cache, to avoid having to create
            # them.
            found_cached_latlon = False            
            for rot_lat, rot_lon in _cache['latlon']:
                if (rot_lon.Data.equals(rotated_lon.Data) and 
                    rot_lat.Data.equals(rotated_lat.Data)):
                    
                    found_cached_latlon = True

                    lat, lon = _cache['latlon'][(rot_lat, rot_lon)] 

                    break
            #--- End: for

            # Create the unrotated latitude and longitude arrays if we
            # couldn't find them in the cache
            if not found_cached_latlon:
#                print  'CALC'
                lat, lon = unrotated_latlon(rotated_lat.varray, 
                                            rotated_lon.varray,
                                            m.bhd[bplat], m.bhd[bplon]) 

                # Add the newly created unrotated latitude and
                # longitude auxiliary coordinates to the cache
                _cache['latlon'][(rotated_lat, rotated_lon)] = (lat, lon)
            #--- End: if
             
            # Create the transform
            transform_id = 'trans%(transN)d' % locals()           
            transN += 1     # Increment transform number

            grid_mapping_name = 'rotated_latitude_longitude'

            transform = Transform()
            transform.name = grid_mapping_name
            transform['grid_mapping_name']         = grid_mapping_name
            transform['grid_north_pole_latitude']  = m.bhd[bplat]
            transform['grid_north_pole_longitude'] = m.bhd[bplon]

            space.transforms[transform_id] = transform

            # Add the transform pointer to the dimension coordinates
            space[ydim].transform = transform_id
            space[xdim].transform = transform_id

            # Create the unrotated latitude and longitude coordinates        
            for (i, axis_code, standard_name, data,
                 units) in zip((  0            ,   1           ),
                               ( 10            ,  11           ), # axis code
                               ('latitude'     , 'longitude'   ), # standard name
                               ( lat           ,  lon          ), # numpy array
                               ('degrees_north', 'degrees_east'), # units
                               ):
                aux   = 'aux%(auxN)d' % locals()           
                auxN += 1                        # Increment auxiliary number

                pubattr = {'standard_name': standard_name,
                           'axis'         : None,
                          }

                coord = _create_Coordinate(f, aux, axis_code, m=m,
                                           pubattr    = pubattr,
                                           cfunits    = Units(units),
                                           data       = data,
                                           dimensions = [dim_lat, dim_lon]) 

                # Add the transform pointer to the auxiliary
                # coordinates
                coord.transform = transform_id
             #--- End: for

        #--- End: if

        # ------------------------------------------------------------
        # Create auxiliary coordinates for the latitudes and
        # longitudes of the sites in site-time cross sections. (Don't
        # consider 'z' extra data because we don't know what it refers
        # to.)
        # ------------------------------------------------------------
        if m.site_time_cross_section:

            for axis_code, extra_type in zip((11 , 10 ),
                                             ('x', 'y')):
                try:
                    coord_type = '%s_domain_bounds' % extra_type
                    m.extra[coord_type]
                except KeyError:
                    try:
                        coord_type = '%s_domain_lower_bound' % extra_type
                        m.extra[coord_type]
                    except KeyError:
                        pass
                    else:           
                        # Create, from extra data, an auxiliary
                        # coordinate with data but no bounds, if the
                        # data noes not contain any missing values
                        file_position = ppfile.tell()
                        data = m.extra[coord_type][...]
                        # Reset the file pointer after reading the
                        # extra data into a numpy array
                        ppfile.seek(file_position, os.SEEK_SET)
                        if not numpy.any(data == _pp_rmdi): # dch also test in bmdi?   
                            aux   = 'aux%(auxN)d' % locals()           
                            auxN += 1                        # Increment auxiliary number
                            coord = _create_Coordinate(f, aux, axis_code, m=m, 
                                                       data       = numpy.array(data),
                                                       pubattr    = {'axis': None}, 
                                                       dimensions = [xdim])# DCH      
                else:                                                                         # xdim?    
                    # Create, from extra data, an auxiliary coordinate                        # should   
                    # with 1) data and bounds, if the upper and lower                         # be       
                    # bounds have no missing values; or 2) data but no                        # the      
                    # bounds, if the upper bound has missing values                           # axis     
                    # but the lower bound does not.                                           # which    
                    file_position = ppfile.tell()                                             # has      
                    bounds = m.extra[coord_type][...]                                         # axis_code
                    # Reset the file pointer after reading the extra                          # 13       
                    # data into a numpy array
                    ppfile.seek(file_position, os.SEEK_SET)
                    data = None
                    if numpy.any(bounds[...,1] == _pp_rmdi): # dch also test in bmdi?
                        if not numpy.any(bounds[...,0] == _pp_rmdi): # dch also test in bmdi?
                            data = bounds[...,0]
                        bounds = None
                    else:
                        data = numpy.mean(bounds, axis=1)
                    if (data, bounds) != (None, None):
                        aux   = 'aux%(auxN)d' % locals()           
                        auxN += 1                        # Increment auxiliary number
                        
                        coord = _create_Coordinate(f, aux, axis_code, m=m,
                                                   data       = data,
                                                   bounds     = bounds, 
                                                   pubattr    = {'axis': None},
                                                   dimensions = [xdim]) # DCH      
                                                                        # xdim?    
                                                                        # should   
                                                                        # be       
                                                                        # the      
                                                                        # axis     
                                                                        # which    
                                                                        # has      
                                                                        # axis_code
                                                                        # 13       
               #--- End: try
           #--- End: for
        #--- End: if
        
        # ------------------------------------------------------------
        # Set the field's data dimension names and the netCDF
        # dimension names for each of the space's dimensions
        # ------------------------------------------------------------
        space.dimensions['data'] = [data_dim0, data_dim1]

        space.nc_dimensions = {}
        for k, ncdim in axis_name.iteritems():
            space.nc_dimensions[axis_dim[k]] = ncdim

        # ------------------------------------------------------------
        # Set the field's data
        # ------------------------------------------------------------
        if field_data is not None:
            f.Data = Data(field_data, 
                          units      = f.Units,
                          _FillValue = getattr(f, '_FillValue', None),
                          chunk      = False)
        #--- End: if

        # ============================================================
        # Create the field's cell methods
        # ============================================================
        cell_methods = []

        proc       = m.lhd[lbproc]
        tmean_proc = 0
        if m.lbtim_ib in (2, 3):
            if proc-128 in (0, 64, 2048, 4096, 8192):
                tmean_proc = 128
                proc -= 128
                
        # ------------------------------------------------------------
        # Area cell methods
        # ------------------------------------------------------------
        if m.ix in (10, 11, 12) and m.iy in (10, 11, 12):
            if 'where' in m.cf_info:
                cell_methods.append('area: mean')
                axis_dim['area']  = None
                axis_name['area'] = 'area'

                cell_methods.append(m.cf_info['where'])
                if 'over' in m.cf_info:
                    cell_methods.append(m.cf_info['over'])
            #--- End: if

            if proc == 64:
                cell_methods.append('x: mean')

#            if proc == 64 or 'where' in m.cf_info:
#                cell_methods.append('area: mean')
#                axis_dim['area']  = None
#                axis_name['area'] = 'area'
#
#            try:
#                cell_methods.append(m.cf_info['where'])
#            except (KeyError, TypeError):
#                pass
#            else:
#                try:
#                    cell_methods.append(m.cf_info['over'])
#                except (KeyError, TypeError):
#                    pass

#            for info in m.cf_info:                
#                if info.startswith('area: mean where_'):                    
#                    cell_methods.append(info.replace('where_', 'where ', 1))
#                elif info.startswith('over_'):
#                    cell_methods.append(info.replace('over_', 'over ', 1))
        # dch : do special zonal mean as as in pp_cfwrite
        
        # ------------------------------------------------------------
        # Vertical cell methods
        # ------------------------------------------------------------
        if proc == 2048:
            cell_methods.append('v: mean')

        # ------------------------------------------------------------
        # Time cell methods
        # ------------------------------------------------------------
        if m.lbtim_ib in (0, 1):
            cell_methods.append('t point')
        elif proc == 4096:
            cell_methods.append('t: minimum')
        elif proc == 8192:
            cell_methods.append('t: maximum')
        if tmean_proc == 128:
            if m.lbtim_ib == 2:
                cell_methods.append('t: mean')
            elif m.lbtim_ib == 3:
                cell_methods.append('t: mean within years')
                cell_methods.append('t: mean over years')

#        print m.lhd[lbproc], proc, tmean_proc, cell_methods, m.lbtim_ib
        # ------------------------------------------------------------
        # Add the cell methods to the field
        # ------------------------------------------------------------
        if cell_methods:
            cell_methods   = CellMethods(' '.join(cell_methods))
            f.cell_methods = _parse_cell_methods(cell_methods, 
                                                 axis_name, axis_dim)

        f.setattr('lbproc', m.lhd[lbproc])
            
        f.finalize()

        # ============================================================
        # The field is complete(!). So add it to the list of fields to
        # be returned
        # ============================================================
        fields_in_file.append(f)

#        print  f

        # ============================================================
        # Now read the next PP field in the file back at the top of
        # this while loop
        # ============================================================
    # --- End: while

    # Close the PP file
#    if close:
    ppfile.close()

    # Return the fields
    return fields_in_file
#--- End: def

def _parse_cell_methods(cell_methods, axis_name, axis_dim):
    '''
'''
    for cell_method in cell_methods:
        for i in xrange(len(cell_method['name'])):
            name = cell_method['name'][i]
            cell_method['dim'][i]  = axis_dim[name]
            cell_method['name'][i] = axis_name[name]
    #--- End: for

    return cell_methods
#--- End: def
