from .utils import RTOL, ATOL, CfDict
from .utils import equals as utils_equals
from .units import Units

# ====================================================================
#
# Space object
#
# ====================================================================

class Space(CfDict):
    '''

Completely describe a field's coordinate system (space).

It contains the dimension constructs, auxiliary coordinate constructs,
cell measure constructs and transform constructs defined by the CF
data model.

The space is a dictionary-like object whose key/value pairs identify
and store the coordinate and cell measure constructs which describe
it.

The dimensionality of the space's components and its transforms are
stored as attributes.

'''

    def __init__(self, *args, **kwargs):
        '''

**Initialization**

The attributes `dimension_sizes`, `dimensions` and `transforms` are
automatically initialized.

:Parameters:

    args, kwargs
        Keys and values are initialized exactly as for a built-in
        dict. Keys are coordinate and cell measure construct
        identifiers (such as dim1, aux0, and cm2) and values are
        coordinate and cell measure instances.

'''
        super(Space, self).__init__(*args, **kwargs)

        self.dimension_sizes = {}
        '''

The sizes of the space's dimensions.

**Examples**

>>> s.dimension_sizes
{'dim2': 96,
 'dim1': 73,
 'dim0': 1}

'''
        self.dimensions = {}
        '''

The dimensions of each of the space's components and of the its
field's data array.

>>> s.dimensions
{'data': ['dim0', 'dim1', 'dim2'],
 'aux0': ['dim1', 'dim2'],
 'aux1': ['dim2', 'dim1'],
 'dim0': ['dim0'],
 'dim1': ['dim1'],
 'dim2': ['dim2'],
 'cm0' : ['dim1', 'dim2']}

'''
 
        self.transforms = CfDict()
        '''

The space's transform constructs.

>>> s.transforms
{'trans0': <CF Transform: ocean_sigma_z_coordinate>
>>> type(s.transforms)
'cf.utils.CfDict'

'''
    #--- End: def

    def __repr__(self):
        '''
x.__repr__() <==> repr(x)

'''
        return '<CF %s: %s>' % (self.__class__.__name__, 
                                tuple(self.dimension_sizes.values()))
    #--- End: def

    def __str__(self):
        '''
x.__str__() <==> str(x)

'''
        def _print_coord(space, key, dimension_coord):
            '''Private function called by __str__'''
            try: 
                variable = space[key]
            except KeyError: 
                size = space.dimension_sizes[key]
                try:
                    name = space.nc_dimensions[key]
                except (KeyError, AttributeError):
                    name = key
                return '%s(%d)' % (name, size)
            #--- End: try

            # Still here?
            if dimension_coord:
                # Dimension coordinate
                shape = variable.shape
                if shape:
                    shape = '(%d)' % variable.size
                else:
                    shape = ''
                name = variable.name(ncvar=True, default=key)
            else:
                # Auxiliary coordinate
                try:
                    name = variable.name(ncvar=True, default=key)
                except TypeError:
                    name = key
                shape = space.dimensions[key][:]
                for i in xrange(len(shape)):
                    try:
                        shape[i] = space[shape[i]].name(ncvar=True, default=shape[i])
                    except KeyError:
                        shape[i] = space.nc_dimensions[shape[i]]
                shape = str(tuple(shape)).replace("'", "")
                shape = shape.replace(',)', ')')
            #--- End: if
            x = [name]
            x.append(shape)

            try:
                variable.compress
            except AttributeError:
                if variable.Units and variable.Units != Units('1'):
                    units = variable.units
                else:
                    units = ''

                ndim = variable.ndim
                size = variable.size

                o_brackets = '[' * ndim
                c_brackets = ']' * ndim

                first = variable.first_datum
                last  = variable.last_datum

                if variable.dtype.kind == 'S':
                    # String valued coordinate's first and last values
                    if size == 1:
                        x.append(" = %s'%s'%s %s" % (o_brackets, first,
                                                      c_brackets, units))
                    elif size >= 3:
                        x.append(" = %s'%s', ..., '%s'%s %s" %  (o_brackets, first,
                                                                 last, c_brackets,
                                                                 units))
                    else:
                        x.append(" = %s'%s', '%s'%s %s" %  (o_brackets, first,
                                                            last, c_brackets,
                                                            units))
                else:
                    # Numeric valued coordinate's first and last values
                    if size == 1:
                        x.append(' = %s%.7g%s %s' % (o_brackets, first,
                                                     c_brackets, units))
                    elif size >= 3:
                        x.append(' = %s%.7g, ..., %.7g%s %s' % (o_brackets, first,
                                                                last, c_brackets,
                                                                units))
                    else:
                        x.append(' = %s%.7g, %.7g%s %s' % (o_brackets, first,
                                                           last, c_brackets,
                                                           units))
                        
            else:
                x.append(' -> compressed ')
                compressed = []
                for unc in space[key].compress:
                    shape = str(unc.size)
                    compressed.append(unc.name(ncvar=True,
                                               default='unc')+'('+shape+')')
                x.append(', '.join(compressed))

            return ''.join(x)
        #--- End: def

        string = ['Dimensions      : ']
        x = []
        for dim in sorted(self.dimension_sizes):
            x.append(_print_coord(self, dim, True))
        string.append('\n                : '.join(x))
        string.append('\n')

        string.append('Auxiliary coords: ')
        x = []
        for aux in sorted(self.get_keys('^aux')):
            x.append(_print_coord(self, aux, False))
        string.append('\n                : '.join(x))
        string.append('\n')

        # Cell measures
        cell_measures = self.get_keys('^cm')
        if cell_measures:
            string.append('Cell measures   : ')
            x = []
            for cm in sorted(cell_measures):
                x.append(_print_coord(self, cm, False))
            string.append('\n                : '.join(x))
            string.append('\n')

        # Virtual auxiliary coordinates
        if self.transforms:
            transforms = self.transforms.values()
            string.append('Virtual coords  : ')
            x = []
            for transform in transforms:
                try:
                    transform['grid_mapping_name']
                    continue
                except KeyError:
                    x.append(repr(transform))
            #-- End: for
            string.append('\n                : '.join(x))
        #--- End: if

        return ''.join(string)
    #--- End: def

    def direction(self, dim):
        '''

Return True if a dimension is increasing, otherwise return False.

A dimension is considered to be increasing if its dimension coordinate
values are increasing in index space or if it has no dimension
coordinate.

The direction is taken directly from the appropriate coordinate's Data
object, if available. (This is because we can assume that the space
has been finalized.)

:Parameters:

    dim : str
        The identifier of the dimension (such as 'dim0').

:Returns:

    out : bool
        Whether or not the dimension is increasing.
        
**Examples**

>>> s.dimension_sizes
{'dim0': 3, 'dim1': 1, 'dim2': 2, 'dim3': 2, 'dim4': 99}
>>> s.dimensions
{'dim0': ['dim0'],
 'dim1': ['dim1'],
 'aux0': ['dim0'],
 'aux1': ['dim2'],
 'aux2': ['dim3'],
}
>>> s['dim0'].array
array([  0  30  60])
>>> s.direction('dim0')
True
>>> s['dim1'].array
array([15])
>>> s['dim1'].bounds.array
array([  30  0])
>>> s.direction('dim1')
False
>>> s['aux1'].array
array([0, -1])
>>> s.direction('dim2')
True
>>> s['aux2'].array
array(['z' 'a'])
>>> s.direction('dim3')
True
>>> s.direction('dim4')
True

'''
        if dim not in self:
            return True
        
        coord = self[dim]

        if not hasattr(coord, 'Data'):
            return True

        data = coord.Data

        if data.is_scalar:
            return data.direction

        return data.direction.values()[0]
    #--- End: def

    def new_dimension(self):
        '''
Return a new

'''
        dimension_sizes = set(self.dimension_sizes)

        N       = len(dimension_sizes)
        new_dim = 'dim%d' % N

        while new_dim not in dimension_sizes:
            N      += 1
            new_dim = 'dim%d' % N
        #--- End: while

        return new_dim
    #--- End: def

    def new_auxiliary(self):
        '''
Return a new 

'''
        aux_keys = self.get_keys('^aux')

        N       = len(aux_keys)
        new_aux = 'aux%d' % N

        while new_aux not in aux_keys:
            N      += 1
            new_aux = 'aux%d' % N
        #--- End: while

        return new_aux
    #--- End: def

    def new_transform(self):
        '''
Return a new 

'''
#        if not hasattr(self, 'transform'):
#            return 'trans0'

        if not self.transforms:
            return 'trans0'
        
        transforms = set(self.transforms)

        N             = len(transforms)
        new_transform = 'trans%d' % N

        while new_transform not in transforms:
            N += 1
            new_transform = 'trans%d' % N
        #--- End: while

        return new_transform
    #--- End: def

    def insert_coordinate(self, coord, dim_name_map=None, 
                          dim=False, aux=False, dimensions=None, 
                          space=None, key=None):
        '''

Insert a new dimension or auxiliary coordinate to the space in place.

:Parameters:

    coord : Coordinate
        The new coordinate.

    dim_name_map : dict, optional

    dim : bool, optional
        True if the new coordinate is to be a dimension coordinate.

    aux : bool, optional
        True if the new coordinate is to be an auxiliary coordinate.

    dimensions : list
        The ordered dimensions of the new coordinate. Ignored if the
        coordinate is a dimension coordinate. Required if the
        coordinate is an auxiliary coordinate.

    space : Space, optional        
        Provide a space which contains both the new coordinate and its
        transforms, thus enabling transforms of the new coordinate to
        be included. By default, transforms of the new coordinate are
        not included.

    key : str, optional
        The identifier for the new coordinate. By default a unique
        identifier will be generated.

:Returns:

    None

**Examples**

>>>

'''
        if dim:
            if not key:
                key = self.new_dimension()

            dimensions                = [key]
            self.dimension_sizes[key] = coord.size
        elif aux:
            if not key:
                key = self.new_auxiliary()
            if not dimensions:
                raise ValueError(
"Must specify the dimensions for a new auxiliary coordinate")
        else:
            raise ValueError("Must specify the role of the new coordinate")
        #--- End: if

        self.dimensions[key] = dimensions[:]
            
        if dim_name_map is None:
            pass

        coord = coord.copy()
        coord.Data.change_dimension_names(dim_name_map)

        if hasattr(coord, 'transform'):
            if not space:
                del coord.transform
            else:
                transform_id = coord.transform
                if space.transforms[transform_id].is_grid_mapping:
                    # Keep grid mappings
#                    if not hasattr(self, 'transforms'):
#                        self.transforms = CfDict()
                        
                    transN = self.new_transform()
                    if space.transforms[transform_id] not in self.transforms:
                        self.transforms[transN] = \
                            space.transforms[transform_id].copy()
                        
                    coord.transform = transN
                else:
                    # Throw away formula_terms (for now ...?)
                    del coord.transform
        #--- End: if

        self[key] = coord

        return coord
    #--- End: def

    def remove_coordinate(self, key):
        '''

Remove a coordinate from the space in place.

:Parameters:

    key : str
        The coordinate's identifier.

:Returns:

    None

**Examples**

>>> s.remove_coordinate('dim0')
>>> s.remove_coordinate('aux1')

'''        
        del self[key]
        del self.dimensions[key]
    #--- End: def

    def squeeze(self, dim):
        '''

Remove a size 1 dimension from the space in place.

If the dimension has a dimension coordinate then it is removed, as are
1-d auxiliary coordinates and cell measures which span the
dimension. The dimension is squeezed from multidimensional auxiliary
coordinates and cell measures if they span it.

The dimension is not squeezed from the field's data array if it spans
it, therefore the field's data array may need to be squeezed
concurrently.

:Parameters:

    dim : str
        The identifier of the dimension to remove.
    
:Returns:

    None

**Examples**

>>> s.dimension_sizes
{'dim0': 12, 'dim1': 73, 'dim2': 1}
>>> s.dimensions
{'data': ['dim0', 'dim1', 'dim2'],
 'aux0': ['dim1', 'dim2'],
 'aux1': ['dim2', 'dim1'],
 'dim0': ['dim0'],
 'dim1': ['dim1'],
 'dim2': ['dim2'],
 'cm0' : ['dim1', 'dim2']}
>>> s.squeeze('dim2')
>>> s.dimension_sizes
{'dim0': 12, 'dim1': 73}
>>> s.dimensions
{'data': ['dim0', 'dim1', 'dim2'],
 'aux0': ['dim1'],
 'aux1': ['dim1'],
 'dim0': ['dim0'],
 'dim1': ['dim1'],
 'cm0' : ['dim1']}

'''
        if dim in self.dimension_sizes:
            if self.dimension_sizes[dim] > 1:
                raise ValueError("Can't squeeze a dimension with size > 1 from %s" %
                                 self.__class__.__name__)

            # Remove the dimension
            del self.dimension_sizes[dim]
        else:
            raise ValueError("Can't remove non-existent dimension from %s" %
                             self.__class__.__name__)

        if dim in self:
            self.remove_coordinate(dim)

        dimensions = self.dimensions.copy()
        del dimensions['data']

        for key, dims in dimensions.iteritems():
            if dim not in dims:
                continue

            if len(dims) == 1:
                # Remove the dimension's 1-d auxiliary coordinates
                self.remove_coordinate(key)
            else:
                # Squeeze the dimension out of N-d auxiliary coordinates
                axis = dims.index(dim)
                dims.pop(axis)
                self[key].squeeze(axis)
        #--- End: for
    #--- End: def

    def aux_coords(self, dim=None):
        '''

Return a dictionary whose values are the auxiliary coordinates which
span the given dimension and keys of the space's auxiliary coordinate
identifiers.

:Parameters:

    dim : str, optional
        The identifier of the dimension to be spanned. By default all
        dimensions are considered (so all auxiliary coordinates are
        returned).

:Returns:

    out : dict
        The auxiliary coordinates and their identifiers.

**Examples**

>>> s.dimensions
{'data': ['dim0', 'dim1', 'dim2'],
 'dim0': ['dim0'],
 'dim1': ['dim1'],
 'dim2': ['dim2'],
 'aux0': ['dim1', 'dim2'],
 'aux1': ['dim0'],
 'aux2': ['dim2', 'dim1']}
>>> s.aux_coords()
{'aux0': <CF Coordinate: ...>,
 'aux1': <CF Coordinate: ...>,
 'aux2': <CF Coordinate: ...>}
>>> s.aux_coords('dim2')
{'aux0': <CF Coordinate: ...>,
 'aux2': <CF Coordinate: ...>}

'''       
        out = {}

        if dim is None:            
            for key in self.get_keys('^aux'):
                out[key] = self[key]

        else:
            dimensions = self.dimensions.copy()
            del dimensions['data']
            for key in self.get_keys('^aux'):
                if dim in dimensions[key]:
                    out[key] = self[key]
        #--- End: if

        return out
    #--- End: def

    def cell_measures(self, dim=None):
        '''

Return a dictionary whose values are the cell measures which span the
given dimension and keys of the space's cell measure identifiers.

:Parameters:

    dim : str, optional
        The identifier of the dimension to be spanned. By default all
        dimensions are considered (so all cell measures are returned).

:Returns:

    out : dict
        The cell measures and their identifiers.

**Examples**

>>> s.dimensions
{'data': ['dim0', 'dim1', 'dim2'],
 'dim0': ['dim0'],
 'dim1': ['dim1'],
 'dim2': ['dim2'],
 'cm0' : ['dim1', 'dim2'],
 'cm1' : ['dim1', 'dim2', 'dim3']}
>>> s.cell_measures()
{'cm0': <CF CellMeasures: ...>,
 'cm1': <CF CellMeasures: ...>}
>>> s.cell_measures('dim3')
{'cm1': <CF CellMeasures: ...>}

'''       
        out = {}

        if dim is None:            
            for key in self.get_keys('^cm'):
                out[key] = self[key]

        else:
            dimensions = self.dimensions.copy()
            del dimensions['data']
            for key in self.get_keys('^cm'):
                if dim in dimensions[key]:
                    out[key] = self[key]
        #--- End: if

        return out
    #--- End: def

    def expand_dims(self, coord=None, size=1):
        '''

Expand the space with a new dimension in place.

The new dimension may by of any size greater then 0.

:Parameters:

    coord : Coordinate, optional
        A dimension coordinate for the new dimension. The new
        dimension's size is set to the size of the coordinate's array.

    size : int, optional
        The size of the new dimension. By default a dimension of size
        1 is introduced. Ignored if `coord` is set.

:Returns:

    None

**Examples**

>>> s.expand_dims()
>>> s.expand_dims(size=12)
>>> c
<CF Coordinate: >
>>> s.expand_dims(coord=c)

'''
        dim = self.new_dimension()

        if coord:
            self[dim] = coord.copy()
            if hasattr(coord, 'size'):
                size = coord.size
        #--- End: if

        self.dimension_sizes[dim] = size
    #--- End: def

    def dump(self, id=None):
        '''
        
Return a string containing a full description of the space.

:Parameters:

    id : str, optional
        Set the common prefix of component names. By default the
        instance's class name is used.

:Returns:

    out : str
        A string containing the description.

**Examples**

>>> x = s.dump()
>>> print s.dump()
>>> print s.dump(id='space1')

'''
        if id is None:
            id = self.__class__.__name__

        try:
            string = ['%s space' % self.name]
        except AttributeError:
            string = ['space']
        string.append(''.ljust(len(string[0]), '-'))

        string.append('%s.dimension_sizes = %s\n' % (id, self.dimension_sizes))
        
        for key in sorted(self.dimensions):
            dims = self.dimensions[key]
            string.append("%(id)s.dimensions['%(key)s'] = %(dims)s" % locals())

        string.append('')

        for key in (sorted(self.get_keys('^dim')) + 
                    sorted(self.get_keys('^aux')) + 
                    sorted(self.get_keys('^cm'))):
            string.append(self[key].dump(id="%(id)s['%(key)s']" % locals()))
           
        for key, transform in self.transforms.iteritems():
            string.append(transform.dump(
                    id="%(id)s.transforms['%(key)s']" % locals()))
            string.append('')
        #--- End: for

        return '\n'.join(string)
    #--- End: def

    def equals(self, other, rtol=None, atol=None, traceback=False):
        '''

True if two spaces are logically equal, False otherwise.

Equality is defined as follows:

    * There is one-to-one correspondence between dimensions and
      dimension sizes between the two spaces.

    * For each space component type (dimension coordinate, auxiliary
      coordinate and cell measures), the set of constructs in one
      space equals that of the other space. The component identifiers
      need not be the same.

    * The set of transforms in one space equals that of the other
      space. The transform identifiers need not be the same.

Equality of numbers is to within a tolerance.

:Parameters:

    other :
        The object to compare for equality.

    atol : float, optional
        The absolute tolerance for all numerical comparisons, By
        default the value returned by the `ATOL` function is used.

    rtol : float, optional
        The relative tolerance for all numerical comparisons, By
        default the value returned by the `RTOL` function is used.

    traceback : bool, optional
        If True then print a traceback highlighting where the two
        instances differ.

:Returns: 

    out : bool
        Whether or not the two instances are equal.

**Examples**

>>> s.equals(t)
True

'''
        if self is other:
            return True
        
        # Check that each instance is the same type
        if type(self) != type(other):
            print \
"%s: Different types: %s, %s" % (self.__class__.__name__,
                                 self.__class__.__name__,
                                 other.__class__.__name__)
            return
        #--- End: if

        if (sorted(self.dimension_sizes.values()) != 
            sorted(other.dimension_sizes.values())):
            # There is not a 1-1 correspondence between dimensions and
            # dimension sizes between the two spaces.           
            if traceback:
                print \
"%s: Different ranks: %s, %s" % (self.__class__.__name__,
                                 sorted(self.dimension_sizes.values()),
                                 sorted(other.dimension_sizes.values()))
            return
        #--- End: if

        if rtol is None:
            rtol = RTOL()
        if atol is None:
            atol = ATOL()              

        # ------------------------------------------------------------
        # Test the transforms
        # ------------------------------------------------------------
        if not self.transforms:
            if other.transforms:
                # Self doesn't have any transforms but other does
                if traceback:
                    print \
"%s: Different numbers of transforms" % self.__class__.__name__
                return
        else:
            if not other.transforms:
                # Other doesn't have any transforms but self does
                if traceback:
                    print \
"%s: Different numbers of transforms" % self.__class__.__name__
                return
            #--- End: if

            transforms0 = self.transforms
            transforms1 = other.transforms

            self_keys  = sorted(transforms0.keys())
            other_keys = sorted(transforms1.keys())
            for key0 in self_keys:
                found_match = False
                for key1 in other_keys: 
                    if utils_equals(transforms0[key0], transforms1[key1],
                                    rtol=rtol, atol=atol, traceback=traceback):
                        found_match = True
                        other_keys.remove(key1)
                        break
                #--- End: for
                if not found_match: 
                    if traceback:
                        print \
"%s: Different transform: %s" % (self.__class__.__name__,
                                 repr(transforms0[key0]))
                    return
            #--- End: for
        #--- End: if

        # ------------------------------------------------------------
        # Test the coordinates and cell measures
        # ------------------------------------------------------------
        for coord_type in ('^dim', '^aux', '^cm'):
            self_keys  = sorted(self.get_keys(coord_type))
            other_keys = sorted(other.get_keys(coord_type))
            for key0 in self_keys:
                found_match = False
                for key1 in other_keys: 
                    if self[key0].equals(other[key1], rtol=rtol, atol=atol,
                                         traceback=traceback):
                        found_match = True
                        other_keys.remove(key1)
                        break
                #--- End: for
                if not found_match:
                    if traceback:
                        print \
"%s: Different coordinate: %s" % (self.__class__.__name__, repr(self[key0]))
                    return
            #--- End: for
        #--- End: for

        # ------------------------------------------------------------
        # Still here? Then the two spaces are equal
        # ------------------------------------------------------------
        return True
    #--- End: def

    def coord(self, arg, role=None, key=False, exact=False, maximal_match=True):
        '''

Find a coordinate of the space by name. Refer to
:meth:`cf.Field.coord` for details.

'''       
        if arg in self:
            if key:
                return arg
            
            return self[arg]

        # Still here?
        if isinstance(arg, (int, long)):
            try:
                result_key = self.dimensions['data'][arg]
            except (IndexError, KeyError):
                return None
            
            if result_key not in self:
                return None
            
            if key:                        
                return result_key            
            
            return self[result_key]
        #--- End: if

        # Still here?
        if isinstance(arg, str):
            arg = {'standard_name': arg}

        if role is None:
            space_keys = self.get_keys('^dim|^aux')
        elif role == 'dim':
            space_keys = self.get_keys('^dim')
        elif role == 'aux':
            space_keys = self.get_keys('^aux')       

        result_key = None

        for space_key in space_keys:

            found_coordinate = False

            for name, value in arg.iteritems():
           
                x = self[space_key].getattr(name, None)

                match = (x is not None and 
                         (exact and x == value) or
                         (not exact and 
                          ((isinstance(x, str)   and x.startswith(value)) or
                           (isinstance(x, Units) and x.equivalent(value)) or
                           x == value))
                         )

                if match:
                    found_coordinate = True
                    if not maximal_match:
                        break
                
                elif maximal_match:
                    found_coordinate = False
                    break
            #--- End: for

            if found_coordinate:
                if result_key is not None:
                    # Found two coordinates which match, so return
                    # None
                    return None

                result_key = space_key
            #--- End: if
        #--- End: for

        if result_key is None:
            # Found no coordinates which match, so return None
            return None

        # Still here? Then found exactly one coordinate which matches,
        # so return it or its key
        if key:
            return result_key
        
        return self[result_key]
    #--- End: def

#--- End: class
