from .utils import CfDict

# ====================================================================
#
# Transform object
#
# ====================================================================

class Transform(CfDict):
    '''

A CF transform construct.

The named parameters and their values of the transformation (i.e. the
transformation's mappings) comprise the object's key-value pairs.

A transformation is equivalent to either a netCDF(CF) 'formula_terms'
or 'grid_mapping' property. The latter is identified by the presence
of the 'grid_mapping_name' key which contains the mapping's name.

In the 'formula_terms' case, a mapping to a coordinate (as opposed to
another field) uses the coordinate's space key name as a pointer
rather than a copy of the coordinate itself.

**Examples**

>>> t
<CF Transform: atmosphere_sigma_coordinate>
>>> print t.dump()
atmosphere_sigma_coordinate transform
-------------------------------------
Transform['ps'] = <CF Field: surface_air_pressure(73, 96)>
Transform['ptop'] = 0.05
Transform['sigma'] = 'dim0'

>>> t
<CF Transform: rotated_latitude_longitude>
>>> print t.dump()
rotated_latitude_longitude transform
------------------------------------
Transform['grid_mapping_name'] = 'rotated_latitude_longitude'
Transform['grid_north_pole_latitude'] = 33.67
Transform['grid_north_pole_longitude'] = 190.0

'''
    def __init__(self, *args, **kwargs):
        '''

**Initialization**

:Parameters:

    args, kwargs
        Keys and values are initialized exactly as for a built-in
        dict. Keys are transform parameter names (such as
        grid_north_pole_latitude or sigma) with appropriate values of
        string, numeric, Coordinate or Field types.

'''
        super(Transform, self).__init__(*args, **kwargs)

        self.name = None
        '''

The identifying name of the transformation.

**Examples**

>>> t.name
'atmosphere_hybrid_sigma_pressure_coordinate'
>>> t.name = 'rotated_latitude_longitude'

'''
    #--- End: def

    def __repr__(self):
        '''
x.__repr__() <==> repr(x)

''' 
        try:
            return '<CF %s: %s>' % (self.__class__.__name__, self.name)
        except AttributeError:
            return '<CF %s: >' % self.__class__.__name__
    #--- End: def

    def __str__(self):
        '''
x.__str__() <==> str(x)

'''    
        return 'Transform       : %s' % repr(self)
    #--- End: def

    @property
    def is_formula_terms(self):
        '''

True if the transform is a formula_terms.

**Examples**

>>> t
<CF Transform: rotated_latitude_longitude>
>>> t.is_formula_terms
False

'''
        return not self.is_grid_mapping
    #--- End: def

    @property
    def is_grid_mapping(self):
        '''

True if the transform is a grid_mapping.

**Examples**

>>> t
<CF Transform: rotated_latitude_longitude>
>>> t.is_grid_mapping
True

'''
        return 'grid_mapping_name' in self
    #--- End: def

    def dump(self, id=None):
        '''

Return a string containing a full description of the transform.

:Parameters:

    id : str, optional
        Set the common prefix of variable component names. By default
        the instance's class name is used.

:Returns:

    out : str
        A string containing the description.

**Examples**

>>> x = t.dump()
>>> print t.dump()
>>> print t.dump(id='transform1')

'''
        if id is None:
            id = self.__class__.__name__
            
        try:
            string = ['%s transform' % self.name]
        except AttributeError:
            string = [' transform']

        string.append(''.ljust(len(string[0]), '-'))

        for key in sorted(self.keys()):
            string.append("%s['%s'] = %s" % (id, key, repr(self[key])))

        return '\n'.join(string)
    #--- End: def

    def hash(self):
        '''

Return a hash value for the transform.

:Returns:

    out : int
        The hash value.

:Raises:

    ValueError:
        If the transform is not a grid mapping.

**Examples**

>>> t.hash()
5768338254506892753

'''
        if not self.is_grid_mapping:
            raise ValueError("Can't hash a formula_terms transform")

        x = []

        for key in sorted(self.keys()):
            x.append(self[key])

        return hash(tuple(x))
    #--- End: def

#--- End: class
