from copy     import deepcopy
from re       import match  as re_match
from textwrap import fill as textwrap_fill
from .utils import RTOL, ATOL
from .utils import equals as utils_equals
from .units import Units


# ====================================================================
#
# Variable object
#
# ====================================================================

class Variable(object):
    '''

Base class for storing a data array with metadata.

A variable contains a data array and metadata comprising properties to
describe the physical nature of the data.

All components of a variable are optional.

'''

    _reserved_attrs = ('_reserved_attrs',
                       '_insert_data'
                       '_getter',
                       '_setter',
                       '_deleter',
                       '_set_Data_attributes',
                       'attributes',
                       'copy',
                       'delattr',
                       'dump',
                       'equals',
                       'extract',
                       'first_datum',
                       'getattr',
                       'hasattr',
                       'is_time',
                       'last_datum',
                       'match',
                       'name',
                       'override_Units',
                       'setattr',
                       )

    _reserved_public_attrs = set(('units',
                                  'calendar',
                                  '_FillValue',
                                  ))

    _special_attributes = set(('units',
                               'calendar',
                               '_FillValue',
                               ))

    def __init__(self, attributes={}):
        '''

**Initialization**

:Parameters:

    attributes : dict, optional
        Initialize a new instance with attributes from the
        dictionary's key/value pairs. Values are deep copied.

'''
        # Initialize the _private dictionary with an empty Units
        # object
        self._private = {'special_attributes': {'Units'     : Units(),
                                                '_FillValue': None},
                         'simple_attributes' : {},
                         }

        for attr, value in attributes.iteritems():
            self.setattr(attr, deepcopy(value))
    #--- End: def

    def __setattr__(self, attr, value):
        '''
x.__setattr__(attr, value) <==> x.attr=value

'''
        super(Variable, self).__setattr__(attr, value)
        return

        raise AttributeError(
            "Can't set %s attribute '%s' (consider the 'setattr()' method instead)" %
            (self.__class__.__name__, attr))
    #--- End: def

    def __delattr__(self, attr):
        '''
x.__delattr__(attr) <==> del x.attr

'''
        if attr in self._reserved_attrs:
            raise AttributeError("Can't delete reserved attribute '%s'" % attr)

        super(Variable, self).__delattr__(attr)
    #--- End: def

    def __getitem__(self, index):
        '''
x.__getitem__(index) <==> x[index]

'''
        if isinstance(index, slice):
            if not (index == slice(None)                                       or
                    (index.start ==  0 and index.stop == None)                 or
                    (index.start == -1 and index.stop == None)                 or
                    (index.start ==  0 and index.stop == 1 and index.step > 0)):
                raise IndexError("%s index out of range: %s" %
                                 (self.__class__.__name__, index))
        else:
            if index not in (0, -1):
                raise IndexError("%s index out of range: %s" %
                                 (self.__class__.__name__, index))
        #--- End: if

        return self
    #--- End: def

    def __setitem__(self, index, value):
        '''
x.__setitem__(index, value) <==> x[index]=value

'''
        pass # for now! DCH
    #--- End: def

    def __len__(self):
        '''
x.__len__() <==> len(x)

'''
        return 1
    #--- End: def

    def __deepcopy__(self, memo):
        '''

Used if copy.deepcopy is called on the variable.

'''
        return self.copy()
    #--- End: def

    def _binary_operation(self, other, method):
        '''

Implement binary arithmetic and comparison operations on the
Variable's Data with the numpy broadcasting rules. It is intended to
be called by the binary arithmetic and comparison methods, such as
__sub__() and __lt__().

:Parameters:

    operation : str
        The binary arithmetic or comparison method name (such as
        "__imul__" or "__ge__").

:Returns:

    new : Variable
        A new Variable, or the same Variable if the operation was
        in-place.

**Examples**

>>> u = cf.Data([0, 1, 2, 3])
>>> v = cf.Data([1, 1, 3, 4])

>>> w = u._binary_operation(u, '__add__')
>>> w.array
array([1 2 5 7])

>>> w = u._binary_operation(v, '__lt__')
>>> w.array
array([ True False  True  True])

>>> u._binary_operation(2, '__imul__')
>>> u.array
array([0 2 4 6])

'''
        if not hasattr(self, 'Data'):
            raise ValueError(
                "Can't use a %s with no Data in %s operation" % \
                    (self.__class__.__name__, method))

        inplace = method[2] == 'i'

        if hasattr(other, 'Data'):
            other = other.Data

        if not inplace:
            new      = self.copy()
            new.Data = getattr(self.Data, method)(other)

            #if not new.Data.Units.equivalent(original_units):
            #    # this is coarse!
            #    new.delattr('standard_name')
            #
            #    if hasattr(new, 'history'):
            #        history = [new.getattr('history')]
            #    else:
            #        history = []
            #
            #    history.append(new.getattr('standard_name'))
            #    history.append(method)
            #
            #    new.setattr('history', ' '.join(history))
            ##--- End: if

            return new

        else:
            getattr(self.Data, method)(other)
            return self
    #--- End: def

    def _unary_operation(self, method):
        '''

Implement unary arithmetic operations on the Variable's Data. It is
intended to be called by the unary arithmetic methods, such as
__abs__().

:Parameters:

    operation : str
        The unary arithmetic method name (such as "__abs__").

:Returns:

    new : Variable
        A new Variable.

**Examples**

>>> type(v)
<CF cf.Variable>
>>> v.array
array([1 2 -3 -4 -5])

>>> w = v._unary_operation('__abs__')
>>> w.array
array([1 2 3 4 5])

>>> w = v.__abs__()
>>> w.array
array([1 2 3 4 5])

>>> w = abs(v)
>>> w.array
array([1 2 3 4 5])

'''
        if not hasattr(self, 'Data'):
            raise TypeError(
                "Can't apply unary operator %s to a %s with no Data" %                
                (method, self.__class__.__name__))

        new      = self.copy()
        new.Data = getattr(new.Data, method)()

        return new
    #--- End: def

    def __add__(self, other):
        '''
x.__add__(y) <==> x+y

'''
        return self._binary_operation(other, '__add__')
    #--- End: def

    def __iadd__(self, other):
        '''
x.__iadd__(y) <==> x+=y

'''
        return self._binary_operation(other, '__iadd__')
    #--- End: def

    def __radd__(self, other):
        '''
x.__radd__(y) <==> y+x

'''
        return self._binary_operation(other, '__radd__')
    #--- End: def

    def __sub__(self, other):
        '''
x.__sub__(y) <==> x-y

'''
        return self._binary_operation(other, '__sub__')
    #--- End: def

    def __isub__(self, other):
        '''
x.__isub__(y) <==> x-=y

'''
        return self._binary_operation(other, '__isub__')
    #--- End: def

    def __rsub__(self, other):
        '''
x.__rsub__(y) <==> y-x

'''
        return self._binary_operation(other, '__rsub__')
    #--- End: def

    def __mul__(self, other):
        '''
x.__mul__(y) <==> x*y

'''
        return self._binary_operation(other, '__mul__')
    #--- End: def

    def __imul__(self, other):
        '''
x.__imul__(y) <==> x*=y

'''
        return self._binary_operation(other, '__imul__')
    #--- End: def

    def __rmul__(self, other):
        '''
x.__rmul__(y) <==> y*x

'''
        return self._binary_operation(other, '__rmul__')
    #--- End: def

    def __div__(self, other):
        '''
x.__div__(y) <==> x/y

'''
        return self._binary_operation(other, '__div__')
    #--- End: def

    def __idiv__(self, other):
        '''
x.__idiv__(y) <==> x/=y

'''
        return self._binary_operation(other, '__idiv__')
    #--- End: def

    def __rdiv__(self, other):
        '''
x.__rdiv__(y) <==> y/x

'''
        return self._binary_operation(other, '__rdiv__')
    #--- End: def

    def __floordiv__(self, other):
        '''
x.__floordiv__(y) <==> x//y

'''
        return self._binary_operation(other, '__floordiv__')
    #--- End: def

    def __ifloordiv__(self, other):
        '''
x.__ifloordiv__(y) <==> x//=y

'''
        return self._binary_operation(other, '__ifloordiv__')
    #--- End: def

    def __rfloordiv__(self, other):
        '''
x.__rfloordiv__(y) <==> y//x

'''
        return self._binary_operation(other, '__rfloordiv__')
    #--- End: def

    def __truediv__(self, other):
        '''
x.__truediv__(y) <==> x/y

'''
        return self._binary_operation(other, '__truediv__')
    #--- End: def

    def __itruediv__(self, other):
        '''
x.__itruediv__(y) <==> x/=y

'''
        return self._binary_operation(other, '__itruediv__')
   #--- End: def

    def __rtruediv__(self, other):
        '''
x.__rtruediv__(y) <==> y/x

'''
        return self._binary_operation(other, '__rtruediv__')
    #--- End: def

    def __pow__(self, other, modulo=None):
        '''
x.__pow__(y) <==> x**y

'''
        if modulo is not None:
            raise NotImplementedError(
                "3-argument power not supported for '%s'" %
                self.__class__.__name__)
        
        return self._binary_operation(other, '__pow__')
    #--- End: def

    def __ipow__(self, other, modulo=None):
        '''
x.__ipow__(y) <==> x**=y

'''
        if modulo is not None:
            raise NotImplementedError(
                "3-argument power not supported for '%s'" %
                self.__class__.__name__)

        return self._binary_operation(other, '__ipow__')
    #--- End: def

    def __rpow__(self, other, modulo=None):
        '''
x.__ipow__(y) <==> x**=y

'''
        if modulo is not None:
            raise NotImplementedError(
                "3-argument power not supported for '%s'" % \
                self.__class__.__name__)

        return self._binary_operation(other, '__rpow__')
    #--- End: def

    def __eq__(self, other):
        '''
x.__eq__(y) <==> x==y

'''
        return self._binary_operation(other, '__eq__')
    #--- End: def

    def __ne__(self, other):
        '''
x.__ne__(y) <==> x!=y

'''
        return self._binary_operation(other, '__ne__')
    #--- End: def

    def __ge__(self, other):
        '''
x.__ge__(y) <==> x>=y

'''
        return self._binary_operation(other, '__ge__')
    #--- End: def

    def __gt__(self, other):
        '''
x.__gt__(y) <==> x>y

'''
        return self._binary_operation(other, '__gt__')
    #--- End: def

    def __le__(self, other):
        '''
x.__le__(y) <==> x<=y

'''
        return self._binary_operation(other, '__le__')
    #--- End: def

    def __lt__(self, other):
        '''
x.__lt__(y) <==> x<y

'''
        return self._binary_operation(other, '__lt__')
    #--- End: def

    def __and__(self, other):
        '''
x.__and__(y) <==> x&y

'''
        return self._binary_operation(other, '__and__')
    #--- End: def

    def __or__(self, other):
        '''
x.__or__(y) <==> x|y

'''
        return self._binary_operation(other, '__or__')
    #--- End: def

    def __abs__(self):
        '''
x.__abs__() <==> abs(x)

'''
        return self._unary_operation('__abs__')
    #--- End: def

    def __neg__(self):
        '''
x.__neg__() <==> -x

'''
        return self._unary_operation('__neg__')
    #--- End: def

    def __invert__(self):
        '''
x.__invert__() <==> ~x

'''
        return self._unary_operation('__invert__')
    #--- End: def

    def __pos__(self):
        '''
x.__pos__() <==> +x

'''
        return self._unary_operation('__pos__')
    #--- End: def

    def __repr__(self):
        '''
x.__repr__() <==> repr(x)

'''
        try:
            name = self.name(default='', long_name=True, ncvar=True)
        except TypeError:
            # DCH This is so, very bad. The whol cfatts things need rethinking
            name = self.name

        try:
            shape = str(self.shape).replace(',)', ')', 1)
            return '<CF %s: %s%s>' % (self.__class__.__name__, name, shape)
        except AttributeError:
            return '<CF %s: %s>' % (self.__class__.__name__, name)
    #--- End: def

    def __str__(self):
        '''

x.__str__() <==> str(x)

'''
        return self.__repr__()
    #--- End: def

    def _set_Data_attributes(self, dimensions, directions):
        '''

Set the data metadata.

:Parameters:

    dimensions : list

    directions : dict

:Returns:

    None

'''
        data = self.Data

        # ----------------------------------------------------------------
        # Dimension names
        # ----------------------------------------------------------------
        dim_name_map = dict((dim, new_dim)
                            for dim, new_dim in zip(data.order, dimensions))
        data.change_dimension_names(dim_name_map)

        # ----------------------------------------------------------------
        # Dimension directions
        # ----------------------------------------------------------------
        if data.is_scalar:
            # Scalar data
            if dimensions:
                data.direction = directions[dimensions[0]]
            else:
                data.direction = True

            for partition in data.partitions.flat():
                partition.direction = data.direction
        else:
            # 1+ dimensional data
            direction = data.direction
            for dim in direction:
                direction[dim] = directions[dim]

            for partition in data.partitions.flat():
                partition.direction = direction.copy()
        #--- End: if
    #--- End: def

    def _insert_data(self, other, indices,
                     ADim, ADim_direction, dim_name_map):
        '''

'''
        self.Data._insert_data(other.Data, indices,
                               ADim, ADim_direction, dim_name_map)
    #--- End: def

    def _get_special_attr(self, attr):
        '''

'''
        try:
            return self._private['special_attributes'][attr]
        except KeyError:
            raise AttributeError(
                "'%s' object doesn't have attribute '%s'" %
                (self.__class__.__name__, attr))
    #--- End: def

    def _set_special_attr(self, attr, value):
        '''

'''
        self._private['special_attributes'][attr] = value
    #--- End: def

    def _del_special_attr(self, attr):
        try:
            del self._private['special_attributes'][attr]
        except KeyError:
            raise AttributeError(
                "'%s' object doesn't have attribute '%s'" %
                (self.__class__.__name__, attr))
    #--- End: def

    def _getter(self, attr):
        '''

Private function for getting attributes from the _private dictionary.

'''
        try:
            return self._private[attr]
        except (KeyError, AttributeError):
            raise AttributeError("Can't get '%s' attribute '%s'" %
                                 (self.__class__.__name__, attr))
    #--- End: def

    def _setter(self, attr, value):
        '''

Private function for setting attributes in the _private dictionary.

'''
        self._private[attr] = value
    #--- End: def

    def _deleter(self, attr):
        '''

Private function for deleting attributes from the _private dictionary.

'''
        try:
            del self._private[attr]
        except KeyError:
            raise AttributeError("Can't delete '%s' attribute '%s'" %
                                 (self.__class__.__name__, attr))
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: Data
    # ----------------------------------------------------------------
    @property
    def Data(self):
        '''

The `Data` object containing the data array.

**Examples**

>>> f.Data
<CF Data: >

'''
        if 'Data' in self._private:
            return self._private['Data']

        raise AttributeError("%s doesn't have attribute 'Data'" %
                             self.__class__.__name__)
    #--- End: def
    @Data.setter
    def Data(self, value):
        self._private['Data'] = value

        # Delete from the variable, if they had been set
        special_attributes = self._private['special_attributes']

        if 'Units' in special_attributes:
            special_attributes.pop('Units')

        if '_FillValue' in special_attributes:
            special_attributes.pop('_FillValue')
    #--- End: def

    @Data.deleter
    def Data(self):
        if hasattr(self, 'Data'):
            # Save the Units and _FillValue attributes before deletion
            special_attributes = self._private['special_attributes']
            special_attributes['Units']      = self.Data.Units
            special_attributes['_FillValue'] = self.Data._FillValue

            del self._private['Data']
        else:
            raise AttributeError(
                "'%s' object doesn't have attribute 'Data'" %
                self.__class__.__name__)
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: Units
    # ----------------------------------------------------------------
    @property
    def Units(self):
        '''

The `Units` object containing the units of the data array.

Stores the units and calendar CF attributes in an internally
consistent manner. These attributes are mirrored by the `units` and
`calendar` attributes respectively.

**Examples**

>>> f.Units
<CF Units: K>

'''
        if hasattr(self, 'Data'):
            return self.Data.Units
        else:
            return self._get_special_attr('Units')
    #--- End: def

    @Units.setter
    def Units(self, value):
        if not isinstance(value, Units):
            raise TypeError(
                "'%s' attribute 'Units' must be a 'Units' object (got '%s')" %
                (self.__class__.__name__, value.__class__.__name__))

        try:
            self.Data.Units = value
        except AttributeError:
            self._set_special_attr('Units', value)
     #--- End: def

    @Units.deleter
    def Units(self):

        raise AttributeError("Can't delete '%s' attribute 'Units'" %
                             self.__class__.__name__)
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: add_offset
    # ----------------------------------------------------------------
    @property
    def add_offset(self):
        '''

The add_offset CF attribute.

This attribute is only used when writing to a file on disk.

**Examples**

>>> f.add_offset = -4.0
>>> f.add_offset
-4.0
>>> del f.add_offset

'''
        return self.getattr('add_offset')
    #--- End: def
    @add_offset.setter
    def add_offset(self, value): self.setattr('add_offset', value)
    @add_offset.deleter
    def add_offset(self):        self.delattr('add_offset')

    # ----------------------------------------------------------------
    # Property attribute: calendar
    # ----------------------------------------------------------------
    @property
    def calendar(self):
        '''

The calendar CF attribute.

This attribute is a mirror of the calendar stored in the `Units`
attribute.

**Examples**

>>> f.calendar = 'noleap'
>>> f.calendar
'noleap'
>>> del f.calendar

'''
        if hasattr(self.Units, 'calendar'):
            return self.Units.calendar

        raise AttributeError("Can't get '%s' attribute 'calendar'" %
                             self.__class__.__name__)
    #--- End: def

    @calendar.setter
    def calendar(self, value):
        self.Units.calendar = value
    #--- End: def

    @calendar.deleter
    def calendar(self):
        try:
            del self.Units.calendar
        except AttributeError:
            raise AttributeError("Can't delete '%s' attribute 'calendar'" %
                                 self.__class__.__name__)
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: comment (a simple attribute)
    # ----------------------------------------------------------------
    @property
    def comment(self):
        '''

The comment CF attribute.

**Examples**

>>> f.comment = 'a comment'
>>> f.comment
'a comment'
>>> del f.comment

'''
        return self.getattr('comment')
    #--- End: def
    @comment.setter
    def comment(self, value): self.setattr('comment', value)
    @comment.deleter
    def comment(self):        self.delattr('comment')

    # ----------------------------------------------------------------
    # Property attribute: _FillValue
    # ----------------------------------------------------------------
    @property
    def _FillValue(self):
        '''

The _FillValue CF attribute.

This attribute is forced to be consistent with the `missing_value`
attribute as follows:

    * Assigning a value to the `missing_value` attribute also assigns
      the same value to the `_FillValue` attribute whether the latter
      has been previously set or not.

    * Assigning a value to the `_FillValue` attribute also assigns the
      same value to the `missing_value` attribute, but only if the
      latter has previously been defined.

**Examples**

>>> f._FillValue = -1e30

'''
        if hasattr(self, 'Data'):
            return self.Data._FillValue
        else:
           return self._get_special_attr('_FillValue')
    #--- End: def

    @_FillValue.setter
    def _FillValue(self, value):
        if hasattr(self, 'Data'):
            self.Data._FillValue = value
        else:
            self._set_special_attr('_FillValue', value)

        if hasattr(self, 'missing_value'):
            if value is None:
                del self._private['simple_attributes']['missing_value']
            else:
                self._private['simple_attributes']['missing_value'] = value
    #--- End: def

    @_FillValue.deleter
    def _FillValue(self):
        if hasattr(self, 'Data'):
            del self.Data._FillValue
        else:
            self._del_special_attr('_FillValue')

        try:
            del self.missing_value
        except AttributeError:
            pass
    #--- End: def
    # ----------------------------------------------------------------
    # Property attribute: history	
    # ----------------------------------------------------------------
    @property
    def history(self):
        '''

The history CF attribute.

**Examples**

>>> f.history = 'created on 2012/10/01'
>>> f.history
'created on 2012/10/01'
>>> del f.history

'''
        return self.getattr('history')
    #--- End: def

    @history.setter
    def history(self, value): self.setattr('history', value)
    @history.deleter
    def history(self):        self.delattr('history')

    # ----------------------------------------------------------------
    # Property attribute: leap_month (a simple attribute)
    # ----------------------------------------------------------------
    @property
    def leap_month(self):
        '''

The leap_month CF attribute.

**Examples**

>>> f.leap_month = 2
>>> f.leap_month
2
>>> del f.leap_month

'''
        return self.getattr('leap_month')
    #--- End: def
    @leap_month.setter
    def leap_month(self, value): self.setattr('leap_month', value)
    @leap_month.deleter
    def leap_month(self):        self.delattr('leap_month')

    # ----------------------------------------------------------------
    # Property attribute: leap_year (a simple attribute)
    # ----------------------------------------------------------------
    @property
    def leap_year(self):
        '''

The leap_year CF attribute.

**Examples**

>>> f.leap_year = 1984
>>> f.leap_year
1984
>>> del f.leap_year

'''
        return self.getattr('leap_year')
    #--- End: def
    @leap_year.setter
    def leap_year(self, value): self.setattr('leap_year', value)
    @leap_year.deleter
    def leap_year(self):        self.delattr('leap_year')

    # ----------------------------------------------------------------
    # Property attribute: long_name
    # ----------------------------------------------------------------
    @property
    def long_name(self):
        '''

The long_name CF attribute.

**Examples**

>>> f.long_name = 'zonal_wind'
>>> f.long_name
'zonal_wind'
>>> del f.long_name

'''
        return self.getattr('long_name')
    #--- End: def
    @long_name.setter
    def long_name(self, value): self.setattr('long_name', value)
    @long_name.deleter
    def long_name(self):        self.delattr('long_name')

    # ----------------------------------------------------------------
    # Property attribute: missing_value
    # ----------------------------------------------------------------
    @property
    def missing_value(self):
        '''

The missing_value CF attribute.

This attribute is forced to be consistent with the `_FillValue`
attribute as follows:

    * Assigning a value to the `missing_value` attribute also assigns
      the same value to the `_FillValue` attribute whether the latter
      has been previously set or not.

    * Assigning a value to the `_FillValue` attribute also assigns the
      same value to the `missing_value` attribute, but only if the
      latter has previously been defined.

**Examples**

>>> f.missing_value = 1e30
>>> f.missing_value
1e30
>>> f._FillValue
1e30
>>> del f.missing_value
>>> f._FillValue
1e30

'''        
        return self.getattr('missing_value')
    #--- End: def
    @missing_value.setter
    def missing_value(self, value): self.setattr('missing_value', value)
    #--- End: def
    @missing_value.deleter
    def missing_value(self):        self.delattr('missing_value')
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: month_lengths (a simple attribute)
    # ----------------------------------------------------------------
    @property
    def month_lengths(self):
        '''

The month_lengths CF attribute.

'''
        return self.getattr('month_lengths')
    #--- End: def

    @month_lengths.setter
    def month_lengths(self, value): self.setattr('month_lengths', value)
    @month_lengths.deleter
    def month_lengths(self):        self.delattr('month_lengths')

    # ----------------------------------------------------------------
    # Property attribute: scale_factor
    # ----------------------------------------------------------------
    @property
    def scale_factor(self):
        '''

The scale_factor CF attribute.

This attribute is only used when writing to a file on disk.

**Examples**

>>> f.scale_factor = 10.0
>>> f.scale_factor
10.0
>>> del f.scale_factor

'''
        return self.getattr('scale_factor')
    #--- End: def
    @scale_factor.setter
    def scale_factor(self, value): self.setattr('scale_factor', value)
    @scale_factor.deleter
    def scale_factor(self):        self.delattr('scale_factor')

    # ----------------------------------------------------------------
    # Property attribute: standard_name
    # ----------------------------------------------------------------
    @property
    def standard_name(self):
        '''

The standard_name CF attribute.

**Examples**

>>> f.standard_name = 'time'
>>> f.standard_name
'time'
>>> del f.standard_name

'''
        return self.getattr('standard_name')
    #--- End: def
    @standard_name.setter
    def standard_name(self, value): self.setattr('standard_name', value)
    @standard_name.deleter
    def standard_name(self):        self.delattr('standard_name')

    # ----------------------------------------------------------------
    # Property attribute: units
    # ----------------------------------------------------------------
    @property
    def units(self):
        '''

The units CF attribute.

This attribute is a mirror of the units stored in the `Units`
attribute.

**Examples**

>>> f.units = 'K'
>>> f.units
'K'
>>> del f.units

'''
        if hasattr(self.Units, 'units'):
            return self.Units.units

        raise AttributeError("Can't get '%s' attribute 'units'" %
                             self.__class__.__name__)
    #--- End: def

    @units.setter
    def units(self, value):
        self.Units.units = value
    #--- End: def
    @units.deleter
    def units(self):
        try:
            del self.Units.units
        except AttributeError:
            raise AttributeError("Can't delete '%s' attribute 'units'" %
                                 self.__class__.__name__)
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: valid_max
    # ----------------------------------------------------------------
    @property
    def valid_max(self):
        '''

The valid_max CF attribute.

**Examples**

>>> f.valid_max = 100.0
>>> f.valid_max
100.0
>>> del f.valid_max

'''
        return self.getattr('valid_max')
    #--- End: def
    @valid_max.setter
    def valid_max(self, value): self.setattr('valid_max', value)
    @valid_max.deleter
    def valid_max(self):        self.delattr('valid_max')

    # ----------------------------------------------------------------
    # Property attribute: valid_min
    # ----------------------------------------------------------------
    @property
    def valid_min(self):
        '''

The valid_min CF attribute.

**Examples**

>>> f.valid_min = 100.0
>>> f.valid_min
100.0
>>> del f.valid_min

'''
        return self.getattr('valid_min')
    #--- End: def
    @valid_min.setter
    def valid_min(self, value): self.setattr('valid_min', value)
    @valid_min.deleter
    def valid_min(self):        self.delattr('valid_min')

    # ----------------------------------------------------------------
    # Property attribute: valid_range
    # ----------------------------------------------------------------
    @property
    def valid_range(self):
        '''

The valid_range CF attribute.

May be set as a numpy array, a list or a tuple. Always returned as a
tuple.

**Examples**

>>> f.valid_range = array([100., 400.])
>>> f.valid_range
(100.0, 400.0)
>>> del f.valid_range
>>> f.valid_range = (50., 450.)

'''
        return tuple(self.getattr('valid_range'))
    #--- End: def
    @valid_range.setter
    def valid_range(self, value): self.setattr('valid_range', tuple(value))
    @valid_range.deleter
    def valid_range(self):        self.delattr('valid_range')

    @property
    def subset(self):
        '''

Return a new variable whose data is subsetted.

This attribute may be indexed to select a subset from dimension index
values.

**Subsetting by indexing**

Subsetting by dimension indices uses an extended Python slicing
syntax, which is similar numpy array indexing. There are two
extensions to the numpy indexing functionality:

* Size 1 dimensions are never removed.

  An integer index i takes the i-th element but does not reduce the
  rank of the output array by one.

* When advanced indexing is used on more than one dimension, the
  advanced indices work independently.

  When more than one dimension's slice is a 1-d boolean array or 1-d
  sequence of integers, then these indices work independently along
  each dimension (similar to the way vector subscripts work in
  Fortran), rather than by their elements.

**Examples**

'''
        return SubsetVariable(self)
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: shape (can't set or delete)
    # ----------------------------------------------------------------
    @property
    def shape(self):
        '''

Tuple of the data array's dimension sizes.

**Examples**

>>> f.shape
(73, 96)

'''
        try:
            return tuple(self.Data.shape)
        except AttributeError:
            raise AttributeError("Can't get '%s' attribute 'shape'" %
                                 self.__class__.__name__)
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: ndim (can't set or delete)
    # ----------------------------------------------------------------
    @property
    def ndim(self):
        '''

Number of dimensions in the data array.

**Examples**

>>> f.shape
(73, 96)
>>> f.ndim
2

'''
        try:
            return self.Data.ndim
        except AttributeError:
            raise AttributeError("Can't get '%s' attribute 'ndim'" %
                                 self.__class__.__name__)
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: size (can't set or delete)
    # ----------------------------------------------------------------
    @property
    def size(self):
        '''

Number of elements in the data array.

**Examples**

>>> f.shape
(73, 96)
>>> f.size
7008

'''
        try:
            return self.Data.size
        except AttributeError:
            raise AttributeError("Can't get '%s' attribute 'size'" %
                                 self.__class__.__name__)
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: dtype
    # ----------------------------------------------------------------
    @property
    def dtype(self):
        '''

Numpy data-type of the data array.

**Examples**

>>> f.dtype
dtype('float64')

'''
        try:
            return self.Data.dtype
        except AttributeError:
            raise AttributeError("Can't get '%s' attribute 'dtype'" %
                                 self.__class__.__name__)
    #--- End: def

    @property
    def array(self):
        '''

A numpy array deep copy of the data array.

Changing the returned numpy array does not change the data array.

**Examples**

>>> a = f.array
>>> type(a)
<type 'numpy.ndarray'>
>>> a
array([0, 1, 2, 3, 4])
>>> a[0] = 999
>>> f.array
array([0, 1, 2, 3, 4])

'''
        if hasattr(self, 'Data'):
            return self.Data.array

        raise AttributeError("%s has no Data'" % self.__class__.__name__)
    #--- End: def

    @property
    def varray(self):
        '''

A numpy array view of the data array.

Changing the elements of the returned view changes the data array.

**Examples**

>>> a = f.varray
>>> type(a)
<type 'numpy.ndarray'>
>>> a
array([0, 1, 2, 3, 4])
>>> a[0] = 999
>>> f.varray
array([999, 1, 2, 3, 4])

'''
        if hasattr(self, 'Data'):
            return self.Data.varray

        raise AttributeError("%s has no Data'" % self.__class__.__name__)
    #--- End: def

#    @property
#    def tarray(self):
#        if not self.is_time():
#            raise TypeError("Can't create a 'Time' array from a non-time variable")
#
#        try:
#            calendar = self.calendar
#        except AttributeError:
#            calendar = 'standard'
#
#        return cf.Time(self.varray, units=self.units, calendar=calendar)
#    #--- End: def

#    def set_time(self, cftime):
#        if not self.is_time():
#            raise TypeError("Can't set time for a non-time variable")
#
#        # Check that the Time array has the same size and shape as the
#        # Variable array
#
#        self.units    = cftime.units
#        self.calendar = cftime.calendar
#
#        self[...] = cftime.array
#    #--- End: def

#    def is_time(self):
#        '''
#'''
#        try:
#            if ' since ' in self.units:
#                return True
#            return False
#        except AttributeError:
#            return False
#    #--- End: def

#    def change_calendar(self, calendar, fix=False):
#        t = self.tarray
#        t = t.change_calendar(calendar, fix=fix)
#        self.set_time(t)
#    #--- End: def

    @property
    def is_scalar(self):
        '''

True if and only if the data array is a scalar array.

**Examples**

>>> f.array
array(2)
>>> f.is_scalar
True

>>> f.array
array([2])
>>> f.is_scalar
False

'''
        return self.Data.is_scalar
    #--- End: if

    @property
    def first_datum(self):
        '''

The first element of the data array.

Equivalent to ``f.subset[(0,)*f.ndim].array.item()``

**Examples**

>>> f.array
array([[1, 2],
       [3, 4]])
>>> f.first_datum
1

'''
        return self.Data[(slice(0,1),) * self.ndim].varray.item()
    #--- End: def

    @property
    def last_datum(self):
        '''

The last element of the data array.

Equivalent to ``f.subset[(-1,)*f.ndim].array.item()``

**Examples**

>>> f.array
array([[1, 2],
       [3, 4]])
>>> f.first_datum
4

'''
        return self.Data[(slice(-1,None),) * self.ndim].varray.item()
    #--- End: def

    def chunk(self, chunksize=None, extra_boundaries=[], adim=[]):
        '''

Partition the data array using Large Amounts of Massive Arrays (LAMA)
functionality.

:Parameters:

    chunksize : int

    extra_boundaries : list

    adim : list

:Returns:

     extra_boundaries, adim : {list, list}

'''
        return self.Data.chunk(chunksize, extra_boundaries, adim)
    #--- End: def

    def copy(self, _omit_special=()):
        '''

Return a deep copy.

Equivalent to ``copy.deepcopy(f)``.

:Returns:

    out :
        The deep copy.

**Examples**

>>> g = f.copy()

'''
        new = type(self)()

        if hasattr(self, 'Data'):
            new.Data = self.Data.copy()

        attrs = set((self._private)) - set(('Data','special_attributes','simple_attributes'))
        for attr in attrs:
            new._private[attr] = deepcopy(self._private[attr])

        attrs = set((self._private['special_attributes'])) - set(_omit_special)
        for attr in attrs:
            new._private['special_attributes'][attr] = \
                deepcopy(self._private['special_attributes'][attr])

        attrs = set((self._private['simple_attributes']))
        for attr in attrs:
            new._private['simple_attributes'][attr] = \
                deepcopy(self._private['simple_attributes'][attr])

        # Attributes
        attrs = set(self.__dict__) - set(('_private',))
        for attr in attrs:
            setattr(new, attr, deepcopy(self.__dict__[attr]))

        return new
    #--- End: def

    def dump(self, id=None, omit=()):
        '''

Return a string containing a full description of the instance.

:Parameters:

    id : str, optional
       Set the common prefix of component names. By default the
       instance's class name is used.

    omit : sequence of strs
        Omit the given attributes from the description.

:Returns:

    out : str
        A string containing the description.

**Examples**

>>> x = v.dump()
>>> print v.dump()
>>> print v.dump(id='variable1')
>>> print v.dump(omit=('long_name',))

'''
        if id is None:
            id = self.__class__.__name__

        string = []

        if hasattr(self, 'Data'):
            string.append(self.Data.dump(id=id))

        # Special attributes
        special = self._private['special_attributes'].copy()
        for attr in omit:
            if attr in special:
                special.pop(attr)
        #--- End: for
        for attr, value in special.iteritems():
            string.append('%s.%s = %s' % (id, attr, repr(value)))

        string.append('')

        # Simple attributes
        simple = self._simple_attributes()
        attrs  = sorted(set(simple) - set(omit))
        if attrs:
            for attr in attrs:
                name   = '%(id)s.%(attr)s = ' % locals()
                value  = repr(simple[attr])
                indent = ' ' * (len(name))
                if value.startswith("'") or value.startswith('"'):
                    indent = '%(indent)s ' % locals()
                string.append(textwrap_fill(name+value, 79,
                                            subsequent_indent=indent))
            #--- End: for
            string.append('')
        #--- End: if

        return '\n'.join(string)
    #--- End: def

    def equals(self, other, rtol=None, atol=None, traceback=False,
               ignore=set()):
        '''

True if two variables are logically equal, False otherwise.

:Parameters:

    other : 
        The object to compare for equality.

    atol : float, optional
        The absolute tolerance for all numerical comparisons, By
        default the value returned by the `ATOL` function is used.

    rtol : float, optional
        The relative tolerance for all numerical comparisons, By
        default the value returned by the `RTOL` function is used.

    traceback : bool, optional
        If True then print a traceback highlighting where the two
        instances differ.

    ignore : set, optional
        Omit these attributes from the comparison.

:Returns: 

    out : bool
        Whether or not the two instances are equal.

**Examples**

>>> f.equals(f)
True
>>> g = f + 1
>>> f.equals(g)
False
>>> g -= 1
>>> f.equals(g)
True
>>> f.setattr('name', 'name0')
>>> g.setattr('name', 'name1')
>>> f.equals(g)
False

'''
        # Check each instance's id
        if self is other:
            return True

        # Check that each instance is the same type
        if self.__class__ != other.__class__:
            if traceback:
                print \
"%s: Different type: %s, %s" % (self.__class__.__name__,
                                self.__class__.__name__,
                                other.__class__.__name__)
            return
        #--- End: if

        self_simple  = self._private['simple_attributes']
        other_simple = other._private['simple_attributes']
        if (set(self_simple) - ignore) != (set(other_simple) - ignore):
            if traceback:
                print \
"%s: Different attributes: %s" % (self.__class__.__name__,
                                  set(self_simple).symmetric_difference(set(other_simple)))
            return
        #--- End: if

        if rtol is None:
            rtol = RTOL()
        if atol is None:
            atol = ATOL()

        # Check that each instance has the same CF property values
        for attr, x in self_simple.iteritems():
            if attr in ignore:
                continue
            y = other_simple[attr]
            if not utils_equals(x, y, rtol=rtol, atol=atol, traceback=traceback):
                if traceback:
                    print \
"%s: Different '%s': %s, %s" % (self.__class__.__name__, attr, repr(x), repr(y))
                return
        #--- End: for

        # Check the special attributes which have 'equals' methods,
        # such as Units, etc.
        self_special  = self._private['special_attributes']
        other_special = other._private['special_attributes']
        if set(self_special) != set(other_special):
            if traceback:
                print \
"%s: Different attributes: %s" % (self.__class__.__name__,
                                  set(self_special).symmetric_difference(set(other_special)))
            return
        #--- End: if

        for attr, x in self_special.iteritems():
            y = other_special[attr]
            if not utils_equals(x, y, rtol=rtol, atol=atol, traceback=traceback): 
                if traceback:
                    print \
"%s: Different %s: %s, %s" % (self.__class__.__name__,
                              attr, repr(x), repr(y))
                return
        #--- End: for

        # Check the data
        if hasattr(self, 'Data'):
            if hasattr(other, 'Data'):
                if not self.Data.equals(other.Data, rtol=rtol, atol=atol,
                                        traceback=traceback):
                    if traceback:
                        print \
"%s: Different data" % self.__class__.__name__
                    return
            else:
                return
        elif hasattr(other, 'Data'):
            return
        #--- End: if

        return True
    #--- End: def

    def match(self, attr={}, priv={}):
        '''

Determine whether or not a variable matches conditions on its
attributes.

:Parameters:

    attr : dict
        Dictionary for which each key/value pair is a standard or
        non-standard CF attribute name and a condition for the
        attribute to be tested against. If the value is a sequence of
        conditions then the attribute matches if at least one of the
        conditions is passed.

        In general, a condition is any object and it is passed if the
        attribute is equal to the object, with the following
        exceptions:

            * If the attribute is string-valued, then the condition
              may be a regular expression pattern recognised by the
              `re` module and the condition is passed if the attribute
              matches the regular expression. Special characters for
              the start and end of the string are assumed and need not
              be included. For example, '.*wind' is equivalent to
              '^.*wind$'.

            * For the 'Units' attribute, the condition is passed if
              the attribute is equivalent (rather than equal) to the
              object. (Note that this behaviour does not apply to the
              'units' attribute.)

    priv : dict, optional
        Dictionary for which each key/value pair is an attribute name
        other than their standard and non-standard CF attribute and a
        condition for the attribute to be tested against. If the value
        is a sequence of conditions then the attribute matches if at
        least one of the conditions is passed.

        In general, a condition behaves as for `attr`, with the
        following exception:

            * For the `Units` attribute, the condition is passed if
              the attribute is equivalent (rather than equal) to the
              object. (Note that this behaviour does not apply to the
              `units` attribute.)

:Returns:

    out : bool
        Whether or not the variable matches the given criteria.

**Examples**

'''
        priv = priv.copy()

        units = priv.pop('Units', None)
        if units is not None:
            if not isinstance(units, (tuple, list)):
                units = (units,)

            match = False
            for v in units:
                if isinstance(v, basestring):
                    v = Units(v)

                if v.equivalent(self.Units):
                    match = True
                    break
                else:
                    continue
            #--- End: for
            if not match:
                return False
        #--- End: if

        for x, getfunc in zip(( attr     , priv             ),
                              ('getattr', '__getattribute__')):

            # Loop round criteria
            for key, value in x.iteritems():

                if value is None:
                    continue

                try:
                    v_value = getattr(self, getfunc)(key)
                except AttributeError:
                    # Variable doesn't have this attribute
                    return False

                if not isinstance(value, (tuple, list)):
                    value = (value,)

                match = False

                for v in value:
                    if isinstance(v, basestring):
                        # Add beginning and end markers
                        v = '^%s$' % v
                        if re_match(v, v_value):
                            match = True
                            break
                        else:
                            continue
                    #--- End: if

                    if v == v_value:
                        match = True
                        break
                    else:
                        continue
                #--- End: for

                if not match:
                    return False
            #--- End: for
        #--- End: for

        return True
    #--- End: def

    def binary_mask(self):
        '''
return new variable containing binary mask

'''
        binary_mask = self.copy()

        for attr in self.attributes:
            delattr(self, attr)

        binary_mask.long_name = 'binary_mask'

        binary_mask.Data = self.Data.binary_mask

        return binary_mask
    #--- End: def

    @property
    def mask(self):
        '''

The boolean missing data mask of the data array.

Returned as a `Data` object. The mask may be set to 'no missing data'
by deleting the attribute.

**Examples**

>>> f.shape
(12, 73, 96)
>>> m = f.mask
>>> type(m)
<cf.data.Data>
>>> m.dtype
dtype('bool')
>>> m.shape
[12, 73, 96]
>>> m.array.shape
(12, 73, 96)

>>> del f.mask
>>> f.array.mask
False
>>> import numpy
>>> f.array.mask is numpy.ma.nomask
True

'''
        return self.Data.mask
    #--- End: def
    @mask.setter
    def mask(self, new_mask): self.Data.mask = new_mask
    @mask.deleter
    def mask(self):           del self.Data.mask

    def extract(self, *args, **kwargs):
        '''

Return the instance if it matches the given conditions.

Equivalent to:

.. code-block:: python

   def extract(f, *args, **kwargs):
       if f.match(*args, **kwargs):
           return f
       raise ValueError('')

:Parameters:

    args, kwargs :
        As for the `match` method.

:Returns:

    out :
        The variable as an object identity, if it matches the given
        conditions.

:Raises:

    ValueError :
        If the variable does not match the conditions.

'''
        if self.match(*args, **kwargs):
            return self
        else:
            raise ValueError("'%s' does not match criteria" %
                             self.__class__.__name__)
    #--- End: def

    def reverse_dims(self, axes=None):
        '''

Reverse the directions of the data array dimensions in place.

Equivalent to indexing the specified dimensions with ``::-1``.

:Parameters:

    axes : int or sequence of ints
         By default all dimensions are reversed.

:Returns:

    out : list of ints
        The axes which were reversed, in arbitrary order.

**Examples**

>>> f.reverse_dims()
>>> f.reverse_dims(1)
>>> f.reverse_dims([0, 1])

'''
        return self.Data.reverse_dims(axes)
    #--- End: def

    def _equivalent(self, other, dim_name_map,
                    rtol=None, atol=None,
                    _check_units=True):
        '''

'''
        standard_name0 = self.getattr('standard_name', None)
        standard_name1 = other.getattr('standard_name', None)

        if not (standard_name0 and standard_name0 == standard_name1):
            return

        return self._equivalent_data(other,
                                     dim_name_map,
                                     rtol=rtol, atol=atol,
                                     _check_units=_check_units)
    #--- End: def

    def _equivalent_data(self, other, dim_name_map,
                         rtol=None, atol=None,
                         _check_units=True):
        '''

'''
        self_data  = getattr(self , 'Data', None)
        other_data = getattr(other, 'Data', None)

        if self_data is None or other_data is None:
            return False

        return self_data._equivalent(other_data,
                                     dim_name_map,
                                     rtol=rtol, atol=atol,
                                     _check_units=_check_units,
                                     use_directions=True)
    #--- End: def

    def expand_dims(self, axis, dim, direction):
        '''
axis is an integer
dim is a string
direction is a boolean

'''
        self.Data.expand_dims(axis, dim, direction)
    #--- End: def

    def squeeze(self, axes=None):
        '''
axes are integers

'''
        return self.Data.squeeze(axes=axes)
    #--- End: def

    def transpose(self, axes=None):
        '''
axes are integers

'''
        self.Data.transpose(self, axes=axes)
    #--- End: def

    @property
    def attributes(self):
        '''

A dictionary of the standard and non-standard CF attributes.

Note that modifying the returned dictionary will not change the
attributes.

**Examples**

>>> f.attributes
{'Conventions': 'CF-1.0',
 '_FillValue': 1e+20,
 'cell_methods': <CF CellMethods: time: mean>,
 'experiment_id': 'stabilization experiment (SRES A1B)',
 'long_name': 'Surface Air Temperature',
 'standard_name': 'AIR_TEMP',
 'title': 'SRES A1B',
 'units': 'K'}

'''
        out = {}

        # Start with special attributes
        for attr in self._special_attributes:
            if hasattr(self, attr):
                out[attr] = getattr(self, attr)

        # Add simple attributes
        out.update(self._simple_attributes())

        return out
    #--- End: def

    def _simple_attributes(self):
        '''

'''
        return self._private['simple_attributes'].copy()
    #--- End: def

    def setattr(self, attr, value):
        '''

Set a standard or non-standard CF attribute.

:Parameters:

    attr : str
        The name of the attribute to set.

    value :
        The value for the attribute.

:Returns:

     None

**Examples**

>>> f.setattr('standard_name', 'time')
>>> f.setattr('foo', 12.5)

'''
        # Set a special attribute
        if attr in self._special_attributes:
            setattr(self, attr, value)
            return None
        #--- End: if

        # Still here? Then set a simple attribute
        self._private['simple_attributes'][attr] = value
        
        if attr == 'missing_value':
            self._FillValue = value
    #--- End: def

    def hasattr(self, attr):
        '''

Return True if the variable has standard or non-standard CF attribute.

:Parameters:

    attr : str
        The name of the attribute.

:Returns:

     out : bool
         True if the instance has the attribute.

**Examples**

>>> f.hasattr('standard_name')
>>> f.hasattr('foo')

'''
        # Has a special attribute?
        if attr in self._special_attributes:
            return hasattr(self, attr)

        # Still here? Then has a simple attribute?
        return attr in self._private['simple_attributes']
    #--- End: def

    def getattr(self, attr, *default):
        '''

Get a standard or non-standard CF attribute.

:Parameters:

    attr : str
        The name of the attribute to get.

    default : optional
        Return `default` if and only if the variable does not have the
        named attribute.

:Returns:

    out :
        The value of the named attribute, or the default value.

:Raises:

    AttributeError :
        If the variable does not have the named attribute a default
        value has not been set.

**Examples**

>>> f.getattr('standard_name')
>>> f.getattr('standard_name', None)
>>> f.getattr('foo')

'''
        # Get a special attribute
        if attr in self._special_attributes:
            return getattr(self, attr, *default)

        # Still here? Then get a simple attribute
        try:
            return self._private['simple_attributes'][attr]
        except KeyError:
            if default:
                return default[0]

        # Still here?
        raise AttributeError("%s doesn't have attribute '%s'" %
                             (self.__class__.__name__, attr))
    #--- End: def

    def delattr(self, attr):
        '''

Delete a standard or non-standard CF attribute.

:Parameters:

    attr : str
        The name of the attribute to delete.

:Returns:

     None

:Raises:

    AttributeError :
        If the variable does not have the named attribute.

**Examples**

>>> f.delattr('standard_name')
>>> f.delattr('foo')

'''
        # Delete a special attribute
        if attr in self._special_attributes:
            delattr(self, attr)
            return None

        # Still here? Then delete a simple attribute
        try:
            del self._private['simple_attributes'][attr]
        except KeyError:
            raise AttributeError(
                "'%s' object doesn't have CF attribute '%s'" %
                (self.__class__.__name__, attr))
    #--- End: def

    def name(self, long_name=False, ncvar=False, default=None):
        '''

Return a name for the variable.

Returns the standard_name, long_name (if requested) or netCDF variable
name (if requested), whichever it finds first, otherwise returns a
default name.

:Parameters:

    long_name : bool, optional

        If True, return the long_name if standard_name does not
        exist.

    ncvar : bool, optional

        If True, return ncvar if neither the standard_name nor
        long_name has already been returned.

    default : str, optional

        Return default if neither standard_name, long_name nor ncvar
        has already been returned.

:Returns:

    name : str
        The name of the variable.

'''
        try:
            return self.standard_name
        except AttributeError:
            pass

        if long_name:
            try:
                return self.long_name
            except AttributeError:
                pass

        if ncvar:
            try:
                return self.ncvar
            except AttributeError:
                pass

        return default
    #--- End: def

    def override_units(self, new_units):
        '''

Override the data array units in place.

Not to be confused with setting the `Units` attribute to units which
are equivalent to the original units.

This is different to setting the `Units` attribute, as the new units
need not be equivalent to the original ones and the data array
elements will not be changed to reflect the new units.

:Parameters:

    new_units : str or Units
        The new units for the data array.

:Returns:

    None

**Examples**

>>> f.Units
<CF Units: hPa>
>>> f.first_datum
100000.0
>>> f.override_units('km')
>>> f.Units
<CF Units: km>
>>> f.first_datum
100000.0
>>> f.override_units(cf.Units('watts'))
>>> f.Units
<CF Units: watts>
>>> f.first_datum
100000.0

'''
        self.Data.override_units(new_units)
    #--- End: def

#--- End: class


# ====================================================================
#
# Variable slice object
#
# ====================================================================

class SubsetVariable(object):

    __slots__ = ('variable',)

    def __init__(self, variable):
        '''

Set the contained variable.

'''
        self.variable = variable
    #--- End: def

    def __getitem__(self, indices):
        '''
x.__getitem__(indices) <==> x[indices]

'''
        variable = self.variable
        new      = variable.copy()

        if hasattr(variable, 'Data'):
            new.Data = variable.Data[indices]

        return new
   #--- End: def

    def __setitem__(self, indices, value):
        '''
x.__setitem__(indices, value) <==> x[indices] = value

'''
        self.variable.Data[indices] = value
    #--- End: def

#--- End: class
