from .utils import CfList

# ====================================================================
#
# VariableList object
#
# ====================================================================

class VariableList(CfList):
    '''

An ordered sequence of variables stored in a list-like object.

In some contexts, whether an object is a variable or a variable list
is not known and does not matter. So to avoid ungainly type testing,
some aspects of the VariableList interface are shared by a variable
and vice versa.

Any attribute or method belonging to a variable may be used on a
variable list and will be applied independently to each element.

Just as it is straight forward to iterate over the variables in a
variable list, a variable will behave like a single element variable
list in iterative and indexing contexts.

'''

#    _reserved_attrs = Variable._reserved_attrs[:]

    def __delattr__(self, attr):
        '''
x.__delattr__(attr) <==> del x.attr

'''
        for v in self.__dict__['_list']:
            delattr(v, attr)
    #--- End: def

    def __getattribute__(self, attr, *default):
        '''
x.__getattribute__(attr) <==> x.attr

'''
        if (attr.startswith('__') or
            attr in super(VariableList, self).__getattribute__('__dict__') or
            attr in dir(self)):
            # Return an attribute of the list-like object
            return super(VariableList, self).__getattribute__(attr, *default)

        # Return a built-in list of the attributes from each element
        # of the cf list
        return [getattr(v, attr, *default)
                for v in super(VariableList, self).__getattribute__('_list')]
    #--- End: def

    def __setattr__(self, attr, value):
        '''
x.__setattr__(attr, value) <==> x.attr=value

'''
        if (attr.startswith('__') or
            attr in super(VariableList, self).__getattribute__('__dict__') or
            attr in dir(self)):
            # Set an attribute on the list-like object
            super(VariableList, self).__setattr__(attr, value)
            return
        #--- End: if

        # Still here? Then do a built-in setattr on each of the list's
        # elements
        for v in self.__dict__['_list']:
            setattr(v, attr, value)
    #--- End: def

    def __repr__(self):
        '''
x.__repr__() <==> repr(x)

'''
        return repr(self._list).replace('>, ', '>,\n ')
    #--- End: def

    def __str__(self):
        '''
x.__str__() <==> str(x)

'''
        return '\n\n'.join([v.__str__() for v in self._list])
    #--- End: def

    def __eq__(self, other):
        '''
x.__eq__(y) <==> x==y

'''
        return type(self)([v.__eq__(other) for v in self._list])
    #--- End: def

    def __ne__(self, other):
        '''
x.__ne__(y) <==> x!=y

'''
        return type(self)([v.__ne__(other) for v in self._list])
    #--- End: def

    def __neg__(self):
        '''
x.__neg__() <==> -x

'''
        return type(self)([v.__neg__() for v in self._list])
    #--- End: def

    def __pos__(self):
        '''
x.__pos__() <==> +x

'''
        return type(self)([v.__pos__() for v in self._list])
    #--- End: def

    def __abs__(self):
         '''
x.__abs__() <==> abs(x)

'''
         return type(self)([v.__abs__() for v in self._list])
    #--- End: def

    def __invert__(self):
         '''
x.__invert__() <==> ~x

'''
         return type(self)([v.__invert__() for v in self._list])
    #--- End: def

    def __add__(self, other):
        '''
x.__add__(y) <==> x+y

        '''
        return type(self)([v.__add__(other) for v in self._list])
    #--- End: def

    def __sub__(self, other):
        '''
x.__sub__(y) <==> x-y
        '''
        return type(self)([v.__sub__(other) for v in self._list])
    #--- End: def

    def __mul__(self, other):
        '''
x.__mul__(y) <==> x*y
        '''
        return type(self)([v.__mul__(other) for v in self._list])
    #--- End: def

    def __div__(self, other):
        '''
x.__div__(y) <==> x/y
'''
        return type(self)([v.__div__(other) for v in self._list])
    #--- End: def

    def __floordiv__(self, other):
        '''
x.__floordiv__(y) <==> x//y
'''
        return type(self)([v.__floordiv__(other) for v in self._list])
    #--- End: def


    def __truediv__(self, other):
        '''
x.__truediv__(y) <==> x/y
'''
        return type(self)([v.__truediv__(other) for v in self._list])
    #--- End: def

    def __radd__(self, other):
        '''
x.__radd__(y) <==> y+x
'''
        return type(self)([v.__radd__(other) for v in self._list])
    #--- End: def

    def __rmul__(self, other):
        '''
x.__rmul__(y) <==> y*x
'''
        return type(self)([v.__rmul__(other) for v in self._list])
    #--- End: def

    def __iadd__(self, other):
        '''
x.__iadd__(y) <==> x+=y
'''
        for v in self._list:
            v.__iadd__(other)
        return self
    #--- End def

    def __isub__(self, other):
        '''
x.__isub__(y) <==> x-=y
'''
        for v in self._list:
            v.__isub__(other)
        return self
    #--- End def

    def __imul__(self, other):
        '''
x.__imul__(y) <==> x*=y
'''
        for v in self._list:
            v.__imul__(other)
        return self
    #--- End def

    def __idiv__(self, other):
        '''
x.__idiv__(y) <==> x/=y
'''
        for v in self._list:
            v.__idiv__(other)
        return self
    #--- End def

    def __ifloordiv__(self, other):
        '''
x.__ifloordiv__(y) <==> x//=y
'''
        for v in self._list:
            v.__ifloordiv__(other)
        return self
    #--- End def

    def __itruediv__(self, other):
        '''
x.__itruediv__(y) <==> x/=y
'''
        for v in self._list:
            v.__itruediv__(other)
        return self
    #--- End def

    @property
    def subset(self):
        '''

Subset each variable in the list, returning a new list of variables.

**Examples**

>>> vl
[<CF Variable: air_temperature(73, 96)>,
 <CF Variable: air_temperature(73, 96)>]
>>> vl.subset[0,0]
[<CF Variable: air_temperature(1,1)>,
 <CF Variable: air_temperature(1,1)>]

'''
        return SubsetVariableList(self)
    #--- End: def

    def dump(self, *arg, **kwargs):
        '''

Return a string containing the full descriptions of each variable in
the list.

'''
        return '\n\n\n'.join([v.dump(*arg, **kwargs) for v in self._list])
    #--- End: def

    def setattr(self, attr, value):
        '''

Set a public attribute from each element of the list of variables.

'''
        for v in self._list:
            v.setattr(attr, value)
    #--- End: def

    def delattr(self, attr):
        '''

Delete a public attribute from each element of the list of variables.

'''
        for v in self._list:
            v.delattr(attr)
    #--- End: def

    def getattr(self, *args):
        '''

Return a built-in list of the public attributes of each element of the
list of variables.

'''
        return [v.getattr(*args) for v in self._list]
    #--- End: def

    def hasattr(self, *args):
        '''

Return a built-in list of the public attributes of each element of the
list of variables.

'''
        return [v.hasattr(*args) for v in self._list]
    #--- End: def

    def name(self, *arg, **kwargs):
        '''

Return a built-in list of the names of each element of the list of
variables.

'''
        return [v.name(*arg, **kwargs) for v in self._list]
    #--- End: def

    def match(self, *args, **kwargs):
        '''

Return a list of booleans showing which elements match the given
conditions.

The match conditions are passed to each element's `match` method in
turn.

:Parameters:

    args, kwargs :
        As for the variable's `match` method.

:Returns:

    out : list
        A built-in list of booleans showing which elements match the
        conditions.

**Examples**

>>> f
[<>
 <>]
>>> f.match(attr={'standard_name': 'air_temperature'})
[True, False]

'''
        return [v.match(*args, **kwargs) for v in self._list]
    #--- End: def

    def extract(self, *args, **kwargs):
        '''

Return the elements which match the given conditions.

The match conditions are passed to each element's `match` method in
turn.

:Parameters:

    args, kwargs :
        As for the variable's `match` method.

:Returns:

    out :
        A new list containing the matching elements.

**Examples**

>>> f
[<>
 <>]
>>> f.extract(attr={'standard_name': 'air_temperature'})
[<>]

'''
        return type(self)([v
                           for v in self._list
                           if v.match(*args, **kwargs)]
                          )
    #--- End: def

#    def override_Units(self, *args, **kwargs):
#        '''
#
#Change the variables' units in place without changing the variables'
#data.
#'''
#        for v in self._list:
#            v.override_Units(*args, **kwargs)
#    #--- End: def

#--- End: class


# ====================================================================
#
# SubsetVariableList object
#
# ====================================================================

class SubsetVariableList(object):
    '''

'''
    __slots__ = ['variablelist']

    def __init__(self, variablelist):
        '''

Set the contained list of variables.

'''
        self.variablelist = variablelist
    #--- End: def

    def __getitem__(self, indices):
        '''
x.__getitem__(indices) <==> x[indices]

'''
        variablelist = self.variablelist
        return type(variablelist)([v.subset[indices] for v in variablelist])
    #--- End: def

    def __setitem__(self, index, value):
        '''
x.__setitem__(index, value) <==> x[index]=value

'''
        for v in self.variablelist:
            v.subset[index] = value
    #--- End: def

#--- End: class
